/*****************************************************************************
 *                                                                           *
 * File:                                                                     *
 *  chtoe.c                                                                  *
 *                                                                           *
 * Description:                                                              *
 *  Chelsio 10Gb Ethernet TOE driver                                         *
 *                                                                           *
 * This program is free software; you can redistribute it and/or modify      *
 * it under the terms of the GNU General Public License, version 2, as       *
 * published by the Free Software Foundation.                                *
 *                                                                           *
 * You should have received a copy of the GNU General Public License along   *
 * with this program; if not, write to the Free Software Foundation, Inc.,   *
 * 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.                 *
 *                                                                           *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR IMPLIED    *
 * WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF      *
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.                     *
 *                                                                           *
 * http://www.chelsio.com                                                    *
 *                                                                           *
 * Copyright (c) 2003 - 2005 Chelsio Communications, Inc.                    *
 * All rights reserved.                                                      *
 *                                                                           *
 * Maintainers: maintainers@chelsio.com                                      *
 *                                                                           *
 * Authors: Dimitrios Michailidis   <dm@chelsio.com>                         *
 *          Tina Yang               <tainay@chelsio.com>                     *
 *          Felix Marti             <felix@chelsio.com>                      *
 *          Scott Bardone           <sbardone@chelsio.com>                   *
 *          Kurt Ottaway            <kottaway@chelsio.com>                   *
 *          Frank DiMambro          <frank@chelsio.com>                      *
 *          Karen Xie               <kxie@chelsio.com>                       *
 *          Chait Tumuluri          <chait@chelsio.com>                      *
 *                                                                           *
 * History:                                                                  *
 *                                                                           *
 ****************************************************************************/
/* $Date: 2005/11/12 00:37:30 $ $RCSfile: chtoe.c,v $ $Revision: 1.120 $ */

#include "common.h"
#include "version.h"
#include <linux/config.h>
#include <linux/module.h>
#include <linux/init.h>
#include <linux/pci.h>
#include <linux/netdevice.h>
#ifdef CONFIG_CHELSIO_T1_OFFLOAD
#include <linux/toedev.h>
#endif
#include <linux/etherdevice.h>
#include <linux/if_vlan.h>
#include <linux/mii.h>
#include <linux/sockios.h>
#include <linux/proc_fs.h>
#include <linux/version.h>
#include <asm/uaccess.h>

#include "chtoetool.h"
#include "cpl5_cmd.h"
#include "regs.h"
#include "gmac.h"
#include "cphy.h"
#include "sge.h"
#include "tp.h"
#include "espi.h"
#include "elmer0.h"
#ifdef CONFIG_CHELSIO_T1_OFFLOAD
#include "mc3.h"
#include "mc4.h"
#include "ulp.h"
#include "mc5.h"
#include "toe_ctl_defs.h"
#include "tom/tom.h"
#include "tom/tcb.h"
#endif

#ifdef work_struct
#include <linux/tqueue.h>
#define INIT_WORK INIT_TQUEUE
#define schedule_work schedule_task
#define flush_scheduled_work flush_scheduled_tasks

FILE_IDENT("@(#) $Id: chtoe.c,v 1.120 2005/11/12 00:37:30 sbardone Exp $");

static inline void schedule_mac_stats_update(struct adapter *ap, int secs)
{
	mod_timer(&ap->stats_update_timer, jiffies + secs * HZ);
}

static inline void cancel_mac_stats_update(struct adapter *ap)
{
	del_timer_sync(&ap->stats_update_timer);
	flush_scheduled_tasks();
}

/*
 * Stats update timer for 2.4.  It schedules a task to do the actual update as
 * we need to access MAC statistics in process context.
 */
static void mac_stats_timer(unsigned long data)
{
	struct adapter *ap = (struct adapter *)data;

	schedule_task(&ap->stats_update_task);
}
#else
#include <linux/workqueue.h>

static inline void schedule_mac_stats_update(struct adapter *ap, int secs)
{
	schedule_delayed_work(&ap->stats_update_task, secs * HZ);
}

static inline void cancel_mac_stats_update(struct adapter *ap)
{
	cancel_delayed_work(&ap->stats_update_task);
	//flush_scheduled_work();
}
#endif

#if BITS_PER_LONG == 64 && !defined(CONFIG_X86_64)
# define FMT64 "l"
#else
# define FMT64 "ll"
#endif

#ifndef HAVE_FREE_NETDEV
#define free_netdev(dev) kfree(dev)
#endif

static char driver_name[]    = DRV_NAME;
static char driver_string[]  = "Chelsio " DRV_TYPE "Network Driver";
static char driver_version[] = DRIVER_VERSION;

#define PCI_DMA_64BIT ~0ULL
#define PCI_DMA_32BIT 0xffffffffULL

#define MAX_CMDQ_ENTRIES 16384
#define MAX_CMDQ1_ENTRIES 1024
#define MAX_RX_BUFFERS 16384
#define MAX_RX_JUMBO_BUFFERS 16384
#define MAX_TX_BUFFERS_HIGH	16384U
#define MAX_TX_BUFFERS_LOW	1536U
#define MAX_TX_BUFFERS		1460U
#define MIN_FL_ENTRIES 32

#define DFLT_MSG_ENABLE (NETIF_MSG_DRV | NETIF_MSG_PROBE | NETIF_MSG_LINK | \
			 NETIF_MSG_TIMER | NETIF_MSG_IFDOWN | NETIF_MSG_IFUP |\
			 NETIF_MSG_RX_ERR | NETIF_MSG_TX_ERR)

/*
 * The EEPROM is actually bigger but only the first few bytes are used so we
 * only report those.
 */
#define EEPROM_SIZE 32

MODULE_DESCRIPTION(MODULE_DESC);
MODULE_AUTHOR("Chelsio Communications");
MODULE_LICENSE("GPL");
MODULE_DEVICE_TABLE(pci, t1_pci_tbl);

static int dflt_msg_enable = DFLT_MSG_ENABLE;

MODULE_PARM(dflt_msg_enable, "i");
MODULE_PARM_DESC(dflt_msg_enable, "Chelsio T1 message enable bitmap");

#ifdef CONFIG_CHELSIO_T1_OFFLOAD
static int dflt_toe_enable = 1;

static int __init toe_disable(char *str)
{
	dflt_toe_enable = 0;
	return 0;
}

__setup("toe_disable", toe_disable);

MODULE_PARM(dflt_toe_enable, "i");
MODULE_PARM_DESC(dflt_toe_enable, "Enable/Disable of Default Toe capability");
#endif

#define HCLOCK 0x0
#define LCLOCK 0x1

/* T1 cards powersave mode */
static int t1_clock(struct adapter *adapter, int mode);
static int dflt_t1_powersave_mode = HCLOCK;
static int __init t1_powersave(char *str)
{
	dflt_t1_powersave_mode = LCLOCK;
	return 0;
}

__setup("t1powersave", t1_powersave);

MODULE_PARM(dflt_t1_powersave, "i");
MODULE_PARM_DESC(dflt_t1_powersave, "Enable/Disable T1 powersaving mode");

static const char pci_speed[][4] = {
	"33", "66", "100", "133"
};

/*
 * Setup MAC to receive the types of packets we want.
 */
static void t1_set_rxmode(struct net_device *dev)
{
	struct adapter *adapter = dev->priv;
	struct cmac *mac = adapter->port[dev->if_port].mac;
	struct t1_rx_mode rm;

	rm.dev = dev;
	rm.idx = 0;
	rm.list = dev->mc_list;
	mac->ops->set_rx_mode(mac, &rm);
}

static void link_report(struct port_info *p)
{
	if (!netif_carrier_ok(p->dev))
		printk(KERN_INFO "%s: link is down\n", p->dev->name);
	else {
		const char *s = "10 Mbps";

		switch (p->link_config.speed) {
			case SPEED_10000: s = "10 Gbps"; break;
			case SPEED_1000:  s = "1000 Mbps"; break;
			case SPEED_100:   s = "100 Mbps"; break;
			case SPEED_10:    s = "10 Mbps"; break;
		}

		printk(KERN_INFO "%s: link is up at %s, %s duplex\n",
		       p->dev->name, s,
		       p->link_config.duplex == DUPLEX_FULL ? "full" : "half");
	}
}

extern void t1_sge_link_speed_change(struct adapter *adapter, int port, int speed);

void t1_os_link_changed(struct adapter *adapter, int port_id, int link_stat,
			int speed, int duplex, int pause)
{
	struct port_info *p = &adapter->port[port_id];

	if (link_stat != netif_carrier_ok(p->dev)) {
		if (link_stat)
			netif_carrier_on(p->dev);
		else
			netif_carrier_off(p->dev);
		link_report(p);

		/* multi-ports: inform toe */
		if ( (speed > 0) && (adapter->params.nports > 1)) {
			unsigned int sched_speed = 10;
			switch(speed) {
			case SPEED_1000:  sched_speed = 1000; break;
			case SPEED_100:   sched_speed = 100; break;
			case SPEED_10:    sched_speed = 10; break;
			}
			(void) t1_sched_update_parms(adapter->sge, port_id, 0, 
						     sched_speed);	
		}
	}
}

static void link_start(struct port_info *p)
{
	struct cmac *mac = p->mac;

	mac->ops->reset(mac);
	if (mac->ops->macaddress_set)
		mac->ops->macaddress_set(mac, p->dev->dev_addr);
	t1_set_rxmode(p->dev);
	t1_link_start(p->phy, mac, &p->link_config);
	mac->ops->enable(mac, MAC_DIRECTION_RX | MAC_DIRECTION_TX);
}

static void enable_hw_csum(struct adapter *adapter)
{
	if (adapter->flags & TSO_CAPABLE)
		t1_tp_set_ip_checksum_offload(adapter->tp, 1); // for TSO only
	if (adapter->flags & UDP_CSUM_CAPABLE)
		t1_tp_set_udp_checksum_offload(adapter->tp, 1);
	t1_tp_set_tcp_checksum_offload(adapter->tp, 1);
}

/*
 * Things to do upon first use of a card.
 * This must run with the rtnl lock held.
 */
static int chtoe_up(struct adapter *adapter)
{
	int err = 0;

	if (!(adapter->flags & FULL_INIT_DONE)) {
		err = t1_init_hw_modules(adapter);
		if (err)
			goto out_err;

		enable_hw_csum(adapter);
		adapter->flags |= FULL_INIT_DONE;
	}

	t1_interrupts_clear(adapter);
	if ((err = request_irq(adapter->pdev->irq,
			       t1_select_intr_handler(adapter), SA_SHIRQ,
			       adapter->name, adapter))) {
		goto out_err;
	}
	t1_sge_start(adapter->sge);
	t1_interrupts_enable(adapter);
 out_err:
	return err;
}

/*
 * Release resources when all the ports and the TOE have been stopped.
 */
static void chtoe_down(struct adapter *adapter)
{
	t1_sge_stop(adapter->sge);
	t1_interrupts_disable(adapter);
	free_irq(adapter->pdev->irq, adapter);
	//flush_scheduled_work();              // wait for external IRQ handler
}

static int chtoe_open(struct net_device *dev)
{
	int err;
	struct adapter *adapter = dev->priv;
	int other_ports = adapter->open_device_map & PORT_MASK;

	if (!adapter->open_device_map && (err = chtoe_up(adapter)) < 0)
		return err;

	__set_bit(dev->if_port, &adapter->open_device_map);
	link_start(&adapter->port[dev->if_port]);
	netif_start_queue(dev);
	if (!other_ports && adapter->params.stats_update_period)
		schedule_mac_stats_update(adapter,
					  adapter->params.stats_update_period);
#ifdef CONFIG_CHELSIO_T1_OFFLOAD
	if ((adapter->tdev != NULL) &&
		(dflt_toe_enable == 1)) {
		err = activate_toedev(adapter->tdev);
		if (err != 0) {
			printk(KERN_WARNING
				"Activation of TOE capability failed, \n");
			printk(KERN_WARNING
				"Regular NIC capability only is available.\n");
			err = 0;
		}
	}	
#endif

	return 0;
}

static int chtoe_close(struct net_device *dev)
{
	struct adapter *adapter = dev->priv;
	struct port_info *p = &adapter->port[dev->if_port];
	struct cmac *mac = p->mac;

	netif_stop_queue(dev);
	mac->ops->disable(mac, MAC_DIRECTION_TX | MAC_DIRECTION_RX);
	netif_carrier_off(dev);

	clear_bit(dev->if_port, &adapter->open_device_map);
	if (adapter->params.stats_update_period &&
	    !(adapter->open_device_map & PORT_MASK)) {
		/* Stop statistics accumulation. */
		smp_mb__after_clear_bit();
		spin_lock(&adapter->work_lock);   /* sync with update task */
		spin_unlock(&adapter->work_lock);
		cancel_mac_stats_update(adapter);
	}

	if (!adapter->open_device_map)
		chtoe_down(adapter);
	return 0;
}

static struct net_device_stats *t1_get_stats(struct net_device *dev)
{
	struct adapter *adapter = dev->priv;
	struct port_info *p = &adapter->port[dev->if_port];
	struct net_device_stats *ns = &p->netstats;
	const struct cmac_statistics *pstats;

	/* Do a full update of the MAC stats */
	pstats = p->mac->ops->statistics_update(p->mac,
						      MAC_STATS_UPDATE_FULL);

	ns->tx_packets = pstats->TxUnicastFramesOK +
		pstats->TxMulticastFramesOK + pstats->TxBroadcastFramesOK;

	ns->rx_packets = pstats->RxUnicastFramesOK +
		pstats->RxMulticastFramesOK + pstats->RxBroadcastFramesOK;

	ns->tx_bytes = pstats->TxOctetsOK;
	ns->rx_bytes = pstats->RxOctetsOK;

	ns->tx_errors = pstats->TxLateCollisions + pstats->TxLengthErrors +
		pstats->TxUnderrun + pstats->TxFramesAbortedDueToXSCollisions;
	ns->rx_errors = pstats->RxDataErrors + pstats->RxJabberErrors +
		pstats->RxFCSErrors + pstats->RxAlignErrors +
		pstats->RxSequenceErrors + pstats->RxFrameTooLongErrors +
		pstats->RxSymbolErrors + pstats->RxRuntErrors;

	ns->multicast  = pstats->RxMulticastFramesOK;
	ns->collisions = pstats->TxTotalCollisions;

	/* detailed rx_errors */
	ns->rx_length_errors = pstats->RxFrameTooLongErrors +
		pstats->RxJabberErrors;
	ns->rx_over_errors   = 0;
	ns->rx_crc_errors    = pstats->RxFCSErrors;
	ns->rx_frame_errors  = pstats->RxAlignErrors;
	ns->rx_fifo_errors   = 0;
	ns->rx_missed_errors = 0;

	/* detailed tx_errors */
	ns->tx_aborted_errors   = pstats->TxFramesAbortedDueToXSCollisions;
	ns->tx_carrier_errors   = 0;
	ns->tx_fifo_errors      = pstats->TxUnderrun;
	ns->tx_heartbeat_errors = 0;
	ns->tx_window_errors    = pstats->TxLateCollisions;
	return ns;
}

#ifdef SET_ETHTOOL_OPS
static u32 get_msglevel(struct net_device *dev)
{
	struct adapter *adapter = dev->priv;

	return adapter->msg_enable;
}

static void set_msglevel(struct net_device *dev, u32 val)
{
	struct adapter *adapter = dev->priv;

	adapter->msg_enable = val;
}

static char stats_strings[][ETH_GSTRING_LEN] = {
        "TxOctetsOK",
        "TxOctetsBad",
        "TxUnicastFramesOK",
        "TxMulticastFramesOK",
        "TxBroadcastFramesOK",
        "TxPauseFrames",
        "TxFramesWithDeferredXmissions",
        "TxLateCollisions",
        "TxTotalCollisions",
        "TxFramesAbortedDueToXSCollisions",
        "TxUnderrun",
        "TxLengthErrors",
        "TxInternalMACXmitError",
        "TxFramesWithExcessiveDeferral",
        "TxFCSErrors",
        "TxJumboFramesOK",
        "TxJumboOctetsOK",

        "RxOctetsOK",
        "RxOctetsBad",
        "RxUnicastFramesOK",
        "RxMulticastFramesOK",
        "RxBroadcastFramesOK",
        "RxPauseFrames",
        "RxFCSErrors",
        "RxAlignErrors",
        "RxSymbolErrors",
        "RxDataErrors",
        "RxSequenceErrors",
        "RxRuntErrors",
        "RxJabberErrors",
        "RxInternalMACRcvError",
        "RxInRangeLengthErrors",
        "RxOutOfRangeLengthField",
        "RxFrameTooLongErrors",
        "RxJumboFramesOK",
        "RxJumboOctetsOK",

	"TSO",
	"VLANextractions",
	"VLANinsertions",
	"RxCsumGood",
	"TxCsumOffload",
	"RxDrops",

	"respQ_empty",
	"respQ_overflow",
	"freelistQ_empty",
	"pkt_too_big",
	"pkt_mismatch",
	"cmdQ_full0",
	"cmdQ_full1",
	"tx_ipfrags",
	"tx_reg_pkts",
	"tx_lso_pkts",
	"tx_do_cksum",
	
	"espi_DIP2ParityErr",
	"espi_DIP4Err",
	"espi_RxDrops",
	"espi_TxDrops",
	"espi_RxOvfl",
	"espi_ParityErr"
};
 
#define T2_REGMAP_SIZE (3 * 1024)

static int get_regs_len(struct net_device *dev)
{
	return T2_REGMAP_SIZE;
}

static void get_drvinfo(struct net_device *dev, struct ethtool_drvinfo *info)
{
	struct adapter *adapter = dev->priv;

	strcpy(info->driver, driver_name);
	strcpy(info->version, driver_version);
	strcpy(info->fw_version, "N/A");
	strcpy(info->bus_info, pci_name(adapter->pdev));
}

static int get_stats_count(struct net_device *dev)
{
	return ARRAY_SIZE(stats_strings);
}

static void get_strings(struct net_device *dev, u32 stringset, u8 *data)
{
	if (stringset == ETH_SS_STATS)
		memcpy(data, stats_strings, sizeof(stats_strings));
}

static void get_stats(struct net_device *dev, struct ethtool_stats *stats,
		      u64 *data)
{
	struct adapter *adapter = dev->priv;
	struct cmac *mac = adapter->port[dev->if_port].mac;
	const struct cmac_statistics *s;
	const struct sge_port_stats *ss;
	const struct sge_intr_counts *t;

	s = mac->ops->statistics_update(mac, MAC_STATS_UPDATE_FULL);
	ss = t1_sge_get_port_stats(adapter->sge, dev->if_port);
	t = t1_sge_get_intr_counts(adapter->sge);

        *data++ = s->TxOctetsOK;
        *data++ = s->TxOctetsBad;
        *data++ = s->TxUnicastFramesOK;
        *data++ = s->TxMulticastFramesOK;
        *data++ = s->TxBroadcastFramesOK;
        *data++ = s->TxPauseFrames;
        *data++ = s->TxFramesWithDeferredXmissions;
        *data++ = s->TxLateCollisions;
        *data++ = s->TxTotalCollisions;
        *data++ = s->TxFramesAbortedDueToXSCollisions;
        *data++ = s->TxUnderrun;
        *data++ = s->TxLengthErrors;
        *data++ = s->TxInternalMACXmitError;
        *data++ = s->TxFramesWithExcessiveDeferral;
        *data++ = s->TxFCSErrors;
        *data++ = s->TxJumboFramesOK;
        *data++ = s->TxJumboOctetsOK;

        *data++ = s->RxOctetsOK;
        *data++ = s->RxOctetsBad;
        *data++ = s->RxUnicastFramesOK;
        *data++ = s->RxMulticastFramesOK;
        *data++ = s->RxBroadcastFramesOK;
        *data++ = s->RxPauseFrames;
        *data++ = s->RxFCSErrors;
        *data++ = s->RxAlignErrors;
        *data++ = s->RxSymbolErrors;
        *data++ = s->RxDataErrors;
        *data++ = s->RxSequenceErrors;
        *data++ = s->RxRuntErrors;
        *data++ = s->RxJabberErrors;
        *data++ = s->RxInternalMACRcvError;
        *data++ = s->RxInRangeLengthErrors;
        *data++ = s->RxOutOfRangeLengthField;
        *data++ = s->RxFrameTooLongErrors;
        *data++ = s->RxJumboFramesOK;
        *data++ = s->RxJumboOctetsOK;

	*data++ = ss->tso;
	*data++ = ss->vlan_xtract;
	*data++ = ss->vlan_insert;
	*data++ = ss->rx_cso_good;
	*data++ = ss->tx_cso;
	*data++ = ss->rx_drops;

	*data++ = (u64)t->respQ_empty;
	*data++ = (u64)t->respQ_overflow;
	*data++ = (u64)t->freelistQ_empty;
	*data++ = (u64)t->pkt_too_big;
	*data++ = (u64)t->pkt_mismatch;
	*data++ = (u64)t->cmdQ_full[0];
	*data++ = (u64)t->cmdQ_full[1];
	*data++ = (u64)t->tx_ipfrags;
	*data++ = (u64)t->tx_reg_pkts;
	*data++ = (u64)t->tx_lso_pkts;
	*data++ = (u64)t->tx_do_cksum;
	
	if (adapter->espi) {
		const struct espi_intr_counts *e;

		e = t1_espi_get_intr_counts(adapter->espi);
		*data++ = (u64)e->DIP2_parity_err;
		*data++ = (u64)e->DIP4_err;
		*data++ = (u64)e->rx_drops;
		*data++ = (u64)e->tx_drops;
		*data++ = (u64)e->rx_ovflw;
		*data++ = (u64)e->parity_err;
	}
}

static inline void reg_block_dump(struct adapter *ap, void *buf,
				  unsigned int start, unsigned int end)
{
	u32 *p = buf + start;

	for ( ; start <= end; start += sizeof(u32))
		*p++ = t1_read_reg_4(ap, start);
}

static void get_regs(struct net_device *dev, struct ethtool_regs *regs,
		     void *buf)
{
	struct adapter *ap = dev->priv;

	/*
	 * Version scheme: bits 0..9: chip version, bits 10..15: chip revision
	 */
	regs->version = 2;

	memset(buf, 0, T2_REGMAP_SIZE);
	reg_block_dump(ap, buf, 0, A_SG_RESPACCUTIMER);
	reg_block_dump(ap, buf, A_MC3_CFG, A_MC4_INT_CAUSE);
	reg_block_dump(ap, buf, A_TPI_ADDR, A_TPI_PAR);
	reg_block_dump(ap, buf, A_TP_IN_CONFIG, A_TP_TX_DROP_COUNT);
	reg_block_dump(ap, buf, A_RAT_ROUTE_CONTROL, A_RAT_INTR_CAUSE);
	reg_block_dump(ap, buf, A_CSPI_RX_AE_WM, A_CSPI_INTR_ENABLE);
	reg_block_dump(ap, buf, A_ESPI_SCH_TOKEN0, A_ESPI_GOSTAT);
	reg_block_dump(ap, buf, A_ULP_ULIMIT, A_ULP_PIO_CTRL);
	reg_block_dump(ap, buf, A_PL_ENABLE, A_PL_CAUSE);
	reg_block_dump(ap, buf, A_MC5_CONFIG, A_MC5_MASK_WRITE_CMD);
}

static int get_settings(struct net_device *dev, struct ethtool_cmd *cmd)
{
	struct adapter *adapter = dev->priv;
	struct port_info *p = &adapter->port[dev->if_port];

	cmd->supported = p->link_config.supported;
	cmd->advertising = p->link_config.advertising;

	if (netif_carrier_ok(dev)) {
		cmd->speed = p->link_config.speed;
		cmd->duplex = p->link_config.duplex;
	} else {
		cmd->speed = -1;
		cmd->duplex = -1;
	}

        cmd->port = (cmd->supported & SUPPORTED_TP) ? PORT_TP : PORT_FIBRE;
        cmd->phy_address = p->phy->addr;
        cmd->transceiver = XCVR_EXTERNAL;
        cmd->autoneg = p->link_config.autoneg;
        cmd->maxtxpkt = 0;
        cmd->maxrxpkt = 0;
	return 0;
}

static int speed_duplex_to_caps(int speed, int duplex)
{
	int cap = 0;

	switch (speed) {
	case SPEED_10:
		if (duplex == DUPLEX_FULL)
			cap = SUPPORTED_10baseT_Full;
		else
			cap = SUPPORTED_10baseT_Half;
		break;
	case SPEED_100:
		if (duplex == DUPLEX_FULL)
			cap = SUPPORTED_100baseT_Full;
		else
			cap = SUPPORTED_100baseT_Half;
		break;
	case SPEED_1000:
		if (duplex == DUPLEX_FULL)
			cap = SUPPORTED_1000baseT_Full;
		else
			cap = SUPPORTED_1000baseT_Half;
		break;
	case SPEED_10000:
		if (duplex == DUPLEX_FULL)
			cap = SUPPORTED_10000baseT_Full;
	}
	return cap;
}

#define ADVERTISED_MASK (ADVERTISED_10baseT_Half | ADVERTISED_10baseT_Full | \
		      ADVERTISED_100baseT_Half | ADVERTISED_100baseT_Full | \
		      ADVERTISED_1000baseT_Half | ADVERTISED_1000baseT_Full | \
		      ADVERTISED_10000baseT_Full)

static int set_settings(struct net_device *dev, struct ethtool_cmd *cmd)
{
	struct adapter *adapter = dev->priv;
	struct port_info *p = &adapter->port[dev->if_port];
	struct link_config *lc = &p->link_config;

	if (!(lc->supported & SUPPORTED_Autoneg))
		return -EOPNOTSUPP;             /* can't change speed/duplex */

	if (cmd->autoneg == AUTONEG_DISABLE) {
		int cap = speed_duplex_to_caps(cmd->speed, cmd->duplex);

		if (!(lc->supported & cap) || cmd->speed == SPEED_1000)
			return -EINVAL;
		lc->requested_speed = cmd->speed;
		lc->requested_duplex = cmd->duplex;
		lc->advertising = 0;
	} else {
		cmd->advertising &= ADVERTISED_MASK;
		if (cmd->advertising & (cmd->advertising - 1))
			cmd->advertising = lc->supported;
		cmd->advertising &= lc->supported;
		if (!cmd->advertising)
			return -EINVAL;
		lc->requested_speed = SPEED_INVALID;
		lc->requested_duplex = DUPLEX_INVALID;
		lc->advertising = cmd->advertising | ADVERTISED_Autoneg;
	}
	lc->autoneg = cmd->autoneg;
	if (netif_running(dev))
		t1_link_start(p->phy, p->mac, lc);
	return 0;
}

static void get_pauseparam(struct net_device *dev,
			   struct ethtool_pauseparam *epause)
{
	struct adapter *adapter = dev->priv;
	struct port_info *p = &adapter->port[dev->if_port];

	epause->autoneg = (p->link_config.requested_fc & PAUSE_AUTONEG) != 0;
	epause->rx_pause = (p->link_config.fc & PAUSE_RX) != 0;
	epause->tx_pause = (p->link_config.fc & PAUSE_TX) != 0;
}

static int set_pauseparam(struct net_device *dev,
			  struct ethtool_pauseparam *epause)
{
	struct adapter *adapter = dev->priv;
	struct port_info *p = &adapter->port[dev->if_port];
	struct link_config *lc = &p->link_config;

	if (epause->autoneg == AUTONEG_DISABLE)
		lc->requested_fc = 0;
	else if (lc->supported & SUPPORTED_Autoneg)
		lc->requested_fc = PAUSE_AUTONEG;
	else
		return -EINVAL;

	if (epause->rx_pause)
		lc->requested_fc |= PAUSE_RX;
	if (epause->tx_pause)
		lc->requested_fc |= PAUSE_TX;
	if (lc->autoneg == AUTONEG_ENABLE) {
		if (netif_running(dev))
			t1_link_start(p->phy, p->mac, lc);
	} else {
		lc->fc = lc->requested_fc & (PAUSE_RX | PAUSE_TX);
		if (netif_running(dev))
			p->mac->ops->set_speed_duplex_fc(p->mac, -1, -1,
							 lc->fc);
	}
	return 0;
}

static u32 get_rx_csum(struct net_device *dev)
{
	struct adapter *adapter = dev->priv;

	return (adapter->flags & RX_CSUM_ENABLED) != 0;
}

static int set_rx_csum(struct net_device *dev, u32 data)
{
	struct adapter *adapter = dev->priv;

	if (data)
		adapter->flags |= RX_CSUM_ENABLED;
	else
		adapter->flags &= ~RX_CSUM_ENABLED;
	return 0;
}

#ifdef NETIF_F_TSO
static int set_tso(struct net_device *dev, u32 value)
{
	struct adapter *adapter = dev->priv;

	if (!(adapter->flags & TSO_CAPABLE))
		return value ? -EOPNOTSUPP : 0;
	return ethtool_op_set_tso(dev, value);
}
#endif

static void get_sge_param(struct net_device *dev, struct ethtool_ringparam *e)
{
	struct adapter *adapter = dev->priv;
	int jumbo_fl = t1_is_T1B(adapter) ? 1 : 0;

	e->rx_max_pending = MAX_RX_BUFFERS;
	e->rx_mini_max_pending = 0;
	e->rx_jumbo_max_pending = MAX_RX_JUMBO_BUFFERS;
	e->tx_max_pending = MAX_CMDQ_ENTRIES;

	e->rx_pending = adapter->params.sge.freelQ_size[!jumbo_fl];
	e->rx_mini_pending = 0;
	e->rx_jumbo_pending = adapter->params.sge.freelQ_size[jumbo_fl];
	e->tx_pending = adapter->params.sge.cmdQ_size[0];
}

static int set_sge_param(struct net_device *dev, struct ethtool_ringparam *e)
{
	struct adapter *adapter = dev->priv;
	int jumbo_fl = t1_is_T1B(adapter) ? 1 : 0;

	if (e->rx_pending > MAX_RX_BUFFERS || e->rx_mini_pending ||
	    e->rx_jumbo_pending > MAX_RX_JUMBO_BUFFERS ||
	    e->tx_pending > MAX_CMDQ_ENTRIES ||
	    e->rx_pending < MIN_FL_ENTRIES ||
	    e->rx_jumbo_pending < MIN_FL_ENTRIES ||
	    e->tx_pending < (adapter->params.nports + 1) * (MAX_SKB_FRAGS + 1))
		return -EINVAL;

	if (adapter->flags & FULL_INIT_DONE) return -EBUSY;

	adapter->params.sge.freelQ_size[!jumbo_fl] = e->rx_pending;
	adapter->params.sge.freelQ_size[jumbo_fl] = e->rx_jumbo_pending;
	adapter->params.sge.cmdQ_size[0] = e->tx_pending;
	adapter->params.sge.cmdQ_size[1] = e->tx_pending > MAX_CMDQ1_ENTRIES ?
		MAX_CMDQ1_ENTRIES : e->tx_pending;
	return 0;
}

static int set_coalesce(struct net_device *dev, struct ethtool_coalesce *c)
{
	struct adapter *adapter = dev->priv;

	/*
	 * If RX coalescing is requested we use NAPI, otherwise interrupts.
	 * This choice can be made only when all ports and the TOE are off.
	 */
	if (adapter->open_device_map == 0)
		adapter->params.sge.polling = c->use_adaptive_rx_coalesce;

	if (adapter->params.sge.polling) {
		adapter->params.sge.rx_coalesce_usecs = 0;
	} else {
		adapter->params.sge.rx_coalesce_usecs = c->rx_coalesce_usecs;
	}
 	adapter->params.sge.coalesce_enable = c->use_adaptive_rx_coalesce;
	adapter->params.sge.sample_interval_usecs = c->rate_sample_interval;
	t1_sge_set_coalesce_params(adapter->sge, &adapter->params.sge);
	return 0;
}

static int get_coalesce(struct net_device *dev, struct ethtool_coalesce *c)
{
	struct adapter *adapter = dev->priv;

	c->rx_coalesce_usecs = adapter->params.sge.rx_coalesce_usecs;
	c->rate_sample_interval = adapter->params.sge.sample_interval_usecs;
	c->use_adaptive_rx_coalesce = adapter->params.sge.coalesce_enable;
	return 0;
}

static int get_eeprom_len(struct net_device *dev)
{
	struct adapter *adapter = dev->priv;

	return t1_is_asic(adapter) ? EEPROM_SIZE : 0;
}

#define EEPROM_MAGIC(ap) \
	(PCI_VENDOR_ID_CHELSIO | ((ap)->params.chip_version << 16))

static int get_eeprom(struct net_device *dev, struct ethtool_eeprom *e,
		      u8 *data)
{
	int i;
	u8 buf[EEPROM_SIZE] __attribute__((aligned(4)));
	struct adapter *adapter = dev->priv;

	e->magic = EEPROM_MAGIC(adapter);
	for (i = e->offset & ~3; i < e->offset + e->len; i += sizeof(u32))
		t1_seeprom_read(adapter, i, (u32 *)&buf[i]);
	memcpy(data, buf + e->offset, e->len);
	return 0;
}

static struct ethtool_ops t1_ethtool_ops = {
	.get_settings      = get_settings,
	.set_settings      = set_settings,
	.get_drvinfo       = get_drvinfo,
	.get_msglevel      = get_msglevel,
	.set_msglevel      = set_msglevel,
	.get_ringparam     = get_sge_param,
	.set_ringparam     = set_sge_param,
	.get_coalesce      = get_coalesce,
	.set_coalesce      = set_coalesce,
#if LINUX_VERSION_CODE > KERNEL_VERSION(2,4,22)
	.get_eeprom_len    = get_eeprom_len,
#endif
	.get_eeprom        = get_eeprom,
	.get_pauseparam    = get_pauseparam,
	.set_pauseparam    = set_pauseparam,
	.get_rx_csum       = get_rx_csum,
	.set_rx_csum       = set_rx_csum,
	.get_tx_csum       = ethtool_op_get_tx_csum,
#if LINUX_VERSION_CODE > KERNEL_VERSION(2,4,22)
	.set_tx_csum       = ethtool_op_set_tx_csum,
#endif
	.get_sg            = ethtool_op_get_sg,
	.set_sg            = ethtool_op_set_sg,
	.get_link          = ethtool_op_get_link,
	.get_strings       = get_strings,
	.get_stats_count   = get_stats_count,
	.get_ethtool_stats = get_stats,
	.get_regs_len      = get_regs_len,
	.get_regs          = get_regs,
#ifdef NETIF_F_TSO
	.get_tso           = ethtool_op_get_tso,
	.set_tso           = set_tso,
#endif
};
#endif

#ifdef CONFIG_CHELSIO_T1_OFFLOAD
#include <linux/toedev.h>
#include "toe_ctl_defs.h"

#define MTUREG(idx) (A_TP_MTU_REG0 + (idx) * 4)

int best_mtu_index(struct toedev *tdev, unsigned short mtu)
{
	struct adapter *adapter = tdev->priv;
	int i;

	for (i = (NMTUS - 1); i > 0; i--)
		 if (adapter->params.mtus[i] <= mtu)
			break;
	return i;
}

/*
 * Write new values to the MTU table.  Caller must validate that the new MTUs
 * are in ascending order.
 */
static int write_smt_entry(const struct adapter *adapter, int iff, int idx);

static void update_mtu_tab(struct adapter *adapter)
{
	int i;

	for (i = 0; i < NMTUS; ++i) {
		int mtu = (unsigned int)adapter->params.mtus[i];
		t1_write_reg_4(adapter, MTUREG(i), mtu);
	}

	/* T204 workaround: tie TOM's mss to MTU */
	if (is_T2(adapter) && (likely(toe_running(adapter))) &&
            adapter->params.nports > 1) {
		for (i = 0; i < adapter->params.nports; i++) {
			struct sk_buff *skb;
			struct cpl_mss_change *req;
			struct net_device *dev = adapter->port[i].dev;
			unsigned int mss = 
			    t1_sched_update_parms(adapter->sge, i, dev->mtu, 0);
			skb = alloc_skb(sizeof(struct cpl_mss_change), 
					GFP_ATOMIC);
			if (!skb) return;

			req = (struct cpl_mss_change *)skb->data;
			OPCODE_TID(req) = htonl(MK_OPCODE_TID(CPL_MSS_CHANGE, 
						i));
			req->mss = htonl(mss);
			skb_put(skb, sizeof(struct cpl_mss_change));	

                        /* Set protocol to ETH_P_CPL5 for packet filter */
                        skb->protocol = htons(ETH_P_CPL5);
                        skb->mac.raw = skb->data;
			local_bh_disable();
                        toe_receive_skb(adapter->tdev, &skb, 1);
			local_bh_enable();
		}
	}
}

static int write_smt_entry(const struct adapter *adapter, int iff, int idx)
{
	struct net_device *dev = adapter->port[iff].dev;
	struct cpl_smt_write_req *req;
	struct sk_buff *skb = alloc_skb(sizeof(*req), GFP_KERNEL);

	if (!skb) return -ENOMEM;  /* can't really happen */

	req = (struct cpl_smt_write_req *)__skb_put(skb, sizeof(*req));
	OPCODE_TID(req) = htonl(MK_OPCODE_TID(CPL_SMT_WRITE_REQ, iff));
	req->mtu_idx = idx;
	req->iff = iff;
	memset(req->src_mac1, 0, sizeof(req->src_mac1));
	memcpy(req->src_mac0, dev->dev_addr, ETH_ALEN);
	toe_send(adapter->tdev, skb);
	return 0;
}

static int init_smt(struct toedev *tdev)
{
	struct adapter *adapter = tdev->priv;
	int i, idx;

	idx = t1_is_asic(adapter) ? NMTUS - 1 : 5;
	for_each_port(adapter, i)
		write_smt_entry(adapter, i, idx);
	return 0;
}

static int chtoe_offload_open(struct toedev *tdev)
{
	struct adapter *adapter = tdev->priv;
	int i, err = 0;

	if (!adapter->open_device_map && (err = chtoe_up(adapter)) < 0)
		goto out;
	__set_bit(TOEDEV_DEVMAP_BIT, &adapter->open_device_map);
        toe_set_lldev(tdev, adapter->port[0].dev);
	update_mtu_tab(adapter);
        init_smt(tdev);
	for_each_port(adapter, i) {
		struct net_device *dev = adapter->port[i].dev;

		dev->features |= NETIF_F_TCPIP_OFFLOAD;
		TOEDEV(dev) = tdev;
	}
 out:
        return err;
}

static int chtoe_offload_close(struct toedev *tdev)
{
	struct adapter *adapter = tdev->priv;
        int i;

        for_each_port(adapter, i) {
		struct net_device *dev = adapter->port[i].dev;

		dev->features &= ~NETIF_F_TCPIP_OFFLOAD;
		TOEDEV(dev) = NULL;
        }
        toe_set_lldev(tdev, NULL);
	clear_bit(TOEDEV_DEVMAP_BIT, &adapter->open_device_map);
        return 0;
}

static int chtoe_offload_xmit(struct toedev *tdev, struct sk_buff *skb)
{
	return t1_sge_tx(skb, tdev->priv, skb->priority & 1, NULL);
}

static int chtoe_offload_ctl(struct toedev *tdev, unsigned int req, void *data)
{
        struct adapter *adapter = tdev->priv;
        struct toe_page_info *pinfo;
        struct toe_tid_range *tid;
	struct toe_mtu_idx *mtup;
	struct toe_dev_byhwaddr *hwdev;

        switch (req) {
        case TOE_GET_TX_PAGE_INFO:
                pinfo = (struct toe_page_info *)data;
                pinfo->page_size = adapter->params.tp.pm_tx_pg_size;
                /* the last two pages are reserved for HW */
		pinfo->num = adapter->params.tp.pm_tx_num_pgs - 2; 
		pinfo->pm_base = adapter->params.tp.pm_tx_base;
                break;
        case TOE_GET_RX_PAGE_INFO:
                pinfo = (struct toe_page_info *)data;
                pinfo->page_size = adapter->params.tp.pm_rx_pg_size;
                pinfo->num = adapter->params.tp.pm_rx_num_pgs;
		pinfo->pm_base = adapter->params.tp.pm_rx_base;
                break;
        case TOE_GET_TX_MAX_CHUNK:
                // Max of TX page size and 16K.
                *((unsigned int *)data) =
                        min(adapter->params.tp.pm_tx_pg_size, 16384U);
                break;
        case TOE_GET_TID_RANGE:
                tid = (struct toe_tid_range *)data;
                tid->num = t1_mc5_get_tcam_size(adapter->mc5) - 
			adapter->params.mc5.nroutes -
			adapter->params.mc5.nservers;
                tid->base = 0;
                break;
        case TOE_GET_STID_RANGE:
                tid = (struct toe_tid_range *)data;
                tid->num = adapter->params.mc5.nservers;
		tid->base = t1_mc5_get_tcam_size(adapter->mc5) - tid->num -
			adapter->params.mc5.nroutes;
                break;
        case TOE_GET_L2T_CAPACITY:
                *(unsigned int *)data = 256;
                break;
        case TOE_GET_RTBL_RANGE:
		tid = (struct toe_tid_range *)data;
		tid->num = adapter->params.mc5.nroutes;
		tid->base = t1_mc5_get_tcam_size(adapter->mc5) - tid->num;
                break;
	case TOE_GET_BEST_MTU_IDX:
		mtup = (struct toe_mtu_idx *)data;
		mtup->mtu_tab_idx = best_mtu_index(tdev, mtup->mtu);
		break;
	case  TOE_GET_MAX_PORT:
                *(unsigned int *)data = adapter->params.nports;
		break;
	case  TOE_GET_DEV_BY_HWADDR:
		{
		int i;
		struct port_info *pport = &(adapter->port[0]);
		hwdev = (struct toe_dev_byhwaddr *)data;
		hwdev->dev = NULL;
		for (i = 0; i < adapter->params.nports; i++, pport++) {
			if (!memcmp(pport->dev->dev_addr, hwdev->mac, pport->dev->addr_len)) {
				hwdev->dev = (void *)pport->dev;
				break;
			}
		}
		break; 
		}
	}
        return 0;
}

static struct toedev *__devinit chtoe_setup_toedev(struct adapter *adapter)
{
	struct toedev *tdev = alloc_toedev();

	if (!tdev) {
		CH_WARN("%s: unable to allocate TOE device, continuing without"
			"TOE support\n", pci_name(adapter->pdev));
		return NULL;
	}
	tdev->open = chtoe_offload_open;
	tdev->close = chtoe_offload_close;
	tdev->send = chtoe_offload_xmit;
	tdev->ctl = chtoe_offload_ctl;
	tdev->ttid = t1_is_T1B(adapter) ?
		TOE_ID_CHELSIO_T1 : TOE_ID_CHELSIO_T1C;
	tdev->priv = adapter;
	return tdev;
}

static int valid_pm_page_size(unsigned int size)
{
	return size == 2048 || size == 4096 || size == 16384 || size == 65536;
}

static int snmp_read_proc(char *buf, char **start, off_t offset, int count,
			  int *eof, void *data)
{
	struct adapter *adapter = data;
	struct tp_mib_statistics m;
	int len = 0;

	down(&adapter->mib_mutex);
	t1_tp_get_mib_statistics(adapter, &m);
	up(&adapter->mib_mutex);

#define MIB32(s, field) len += sprintf(buf + len, "%-18s %u\n", s, m.field)
#define MIB64(s, hi, lo) \
	len += sprintf(buf + len, "%-18s %" FMT64 "u\n", s, (u64)m.hi + m.lo)

	MIB64("IPInReceives:", ipInReceive_hi, ipInReceive_lo);
	MIB64("IPInHdrErrors:", ipInHdrErrors_hi, ipInHdrErrors_lo);
	MIB64("IPInAddrErrors:", ipInAddrErrors_hi, ipInAddrErrors_lo);
	MIB64("IPInUnknownProtos:", ipInUnknownProtos_hi, ipInUnknownProtos_lo);
	MIB64("IPInDiscards:", ipInDiscards_hi, ipInDiscards_lo);
	MIB64("IPInDelivers:", ipInDelivers_hi, ipInDelivers_lo);
	MIB64("IPOutRequests:", ipOutRequests_hi, ipOutRequests_lo);
	MIB64("IPOutDiscards:", ipOutDiscards_hi, ipOutDiscards_lo);
	MIB64("IPOutNoRoutes:", ipOutNoRoutes_hi, ipOutNoRoutes_lo);
	MIB32("IPReasmTimeout:", ipReasmTimeout);
	MIB32("IPReasmReqds:", ipReasmReqds);
	MIB32("IPReasmOKs:", ipReasmOKs);
	MIB32("IPReasmFails:", ipReasmFails);
	MIB32("TCPActiveOpens:", tcpActiveOpens);
	MIB32("TCPPassiveOpens:", tcpPassiveOpens);
	MIB32("TCPAttemptFails:", tcpAttemptFails);
	MIB32("TCPEstabResets:", tcpEstabResets);
	MIB32("TCPOutRsts:", tcpOutRsts);
	MIB32("TCPCurrEstab:", tcpCurrEstab);
	MIB64("TCPInSegs:", tcpInSegs_hi, tcpInSegs_lo);
	MIB64("TCPOutSegs:", tcpOutSegs_hi, tcpOutSegs_lo);
	MIB64("TCPRetransSeg:", tcpRetransSeg_hi, tcpRetransSeg_lo);
	MIB64("TCPInErrs:", tcpInErrs_hi, tcpInErrs_lo);
	MIB32("TCPRtoMin:", tcpRtoMin);
	MIB32("TCPRtoMax:", tcpRtoMax);

	if (len <= offset + count) *eof = 1;
	*start = buf + offset;
	len -= offset;
	if (len > count) len = count;
	if (len < 0) len = 0;
	return len;
}

#undef MIB32
#undef MIB64
#endif /* CONFIG_CHELSIO_T1_OFFLOAD */

static int intr_read_proc(char *buf, char **start, off_t offset,
			  int count, int *eof, void *data)
{
	struct adapter *adapter = data;
	const struct sge_intr_counts *sc;
#ifdef CONFIG_CHELSIO_T1_OFFLOAD
	const struct pemc3_intr_counts *c3;
	const struct pemc4_intr_counts *c4;
	const struct pemc5_intr_counts *c5;
	const struct ulp_intr_counts *uc;
#endif
	int len = 0;

	sc = t1_sge_get_intr_counts(adapter->sge);
#ifdef CONFIG_CHELSIO_T1_OFFLOAD
	c3 = t1_mc3_get_intr_counts(adapter->mc3);
	c4 = t1_mc4_get_intr_counts(adapter->mc4);
	c5 = t1_mc5_get_intr_counts(adapter->mc5);
	uc = t1_ulp_get_intr_counts(adapter->ulp);
#endif

#define C(cnt, s, v) len += sprintf(buf + len, "%-22s %u\n", s, cnt->v)

	len += sprintf(buf + len, "[SGE]\n");
	C(sc, "ResponseQEmpty:", respQ_empty);
	C(sc, "ResponseQOverflow:", respQ_overflow);
	C(sc, "FreelistQEmpty:", freelistQ_empty);
	C(sc, "PktSizeErr:", pkt_too_big);
	C(sc, "PktMismatch:", pkt_mismatch);
	C(sc, "CmdQ0Full (TOE):", cmdQ_full[0]);
	C(sc, "CmdQ0Full (Eth):", cmdQ_full[2]);
	C(sc, "CmdQ1Full:", cmdQ_full[1]);
	C(sc, "CmdQ0Restart (TOE):", cmdQ_restarted[0]);
	C(sc, "CmdQ0Restart (Eth):", cmdQ_restarted[2]);
	C(sc, "CmdQ1Restart:", cmdQ_restarted[1]);
	C(sc, "EthernetPackets:", ethernet_pkts);
	C(sc, "OffloadPackets:", offload_pkts);
	C(sc, "OffloadPacketBundles:", offload_bundles);
	C(sc, "PureRepsonses:", pure_rsps);
	C(sc, "UnhandledIRQs:", unhandled_irqs);
	C(sc, "tx_ipfrags:", tx_ipfrags);
	C(sc, "tx_reg_pkts:", tx_reg_pkts);
	C(sc, "tx_lso_pkts:", tx_lso_pkts);
	C(sc, "tx_do_cksum:", tx_do_cksum);

	if (adapter->espi) {
		const struct espi_intr_counts *ec;

		ec = t1_espi_get_intr_counts(adapter->espi);
		len += sprintf(buf + len, "\n[ESPI]\n");
		C(ec, "DIP4Err:", DIP4_err);
		C(ec, "RxDrops:", rx_drops);
		C(ec, "TxDrops:", tx_drops);
		C(ec, "RxOvfl:", rx_ovflw);
		C(ec, "ParityErr:", parity_err);
		C(ec, "DIP2ParityErr:", DIP2_parity_err);
	}
#ifdef CONFIG_CHELSIO_T1_OFFLOAD
	len += sprintf(buf + len, "\n[ULP]\n");
	C(uc, "RgnTabParityErr:", region_table_parity_err);
	C(uc, "Etp2ulpDataParityErr:", egress_tp2ulp_data_parity_err);
	C(uc, "Itp2ulpDataParityErr:", ingress_tp2ulp_data_parity_err);
	C(uc, "PMintr:", pm_intr);
	C(uc, "PMe2cCmdSyncErr:", pm_e2c_cmd_payload_sync_err);
	C(uc, "PMc2eCmdSyncErr:", pm_c2e_cmd_payload_sync_err);
	C(uc, "PMe2cFifoRdEmptyErr:", pm_e2c_fifo_read_empty_err);
	C(uc, "PMc2eFifoRdEmptyErr:", pm_c2e_fifo_read_empty_err);
	C(uc, "PMparityErr:", pm_parity_err);
	C(uc, "PMe2cFifoWrFullErr:", pm_e2c_fifo_write_full_err);
	C(uc, "PMc2eFifoWrFullErr:", pm_c2e_fifo_write_full_err);

	len += sprintf(buf + len, "\n[MC3]\n");
	C(c3, "CorrectableErr:", corr_err);
	C(c3, "UncorrectableErr:", uncorr_err);
	C(c3, "ParityErr:", parity_err);
	C(c3, "AddressErr:", addr_err);

	len += sprintf(buf + len, "\n[MC4]\n");
	C(c4, "CorrectableErr:", corr_err);
	C(c4, "UncorrectableErr:", uncorr_err);
	C(c4, "AddressErr:", addr_err);

	len += sprintf(buf + len, "\n[MC5]\n");
	C(c5, "HitOutActiveRegionErr:", hit_out_active_region_err);
	C(c5, "HitInActiveRegionErr:", hit_in_active_region_err);
	C(c5, "HitInRoutingRegionErr:", hit_in_routing_region_err);
	C(c5, "MissErr:", miss_err);
	C(c5, "LipEqualZeroErr:", lip_equal_zero_err);
	C(c5, "LipMissErr:", lip_miss_err);
	C(c5, "ParityErr:", parity_err);
	C(c5, "ActiveRegionFullErr:", active_region_full_err);
	C(c5, "NextFreeAddrSrchErr:", next_free_addr_srch_err);
	C(c5, "SynCookie:", syn_cookie);
	C(c5, "CookieBadMsg:", syn_cookie_bad_message);
	C(c5, "SynCookieOffMsg:", syn_cookie_off_message);
	C(c5, "ReceiveUnknownCmd:", receive_unknown_cmd);
	C(c5, "ParityInReqQErr:", parity_in_request_q_err);
	C(c5, "ParityInDispQErr:", parity_in_dispatch_q_err);
	C(c5, "DelAndActIsEmpty:", del_and_act_is_empty);
#endif
#undef C

	if (len <= offset + count) *eof = 1;
	*start = buf + offset;
	len -= offset;
	if (len > count) len = count;
	if (len < 0) len = 0;
	return len;
}

static int __devinit chtoe_proc_setup(struct adapter *adapter,
				      struct proc_dir_entry *dir)
{
	struct proc_dir_entry *p;
	const char *name;

	if (!dir)
		return -EINVAL;
#ifdef CONFIG_CHELSIO_T1_OFFLOAD
	name = "snmp";
	p = create_proc_read_entry(name, 0, dir, snmp_read_proc, adapter);
	if (!p) return -ENOMEM;
	p->owner = THIS_MODULE;
	name = "irq_details";
#else
	name = adapter->name;
#endif
	/*
	 * Remaining entries are created on a best-effort basis, i.e., we do
	 * not fail for lack of memory.
	 */
	p = create_proc_read_entry(name, 0, dir, intr_read_proc,
				   adapter);
	if (p) p->owner = THIS_MODULE;

	return 0;
}

static void chtoe_proc_cleanup(struct adapter *adapter,
					struct proc_dir_entry *dir)
{
	const char *name;
#ifdef CONFIG_CHELSIO_T1_OFFLOAD
	remove_proc_entry("snmp", dir);
	name = "irq_details";
#else
	name = adapter->name;
#endif
	remove_proc_entry(name, dir);
}
#ifndef CONFIG_CHELSIO_T1_OFFLOAD
#define chtoe_setup_toedev(adapter) NULL
#define update_mtu_tab(adapter)
#define write_smt_entry(adapter, iff, idx)
#endif

static int toetool_ioctl(struct net_device *dev, void *useraddr)
{
	u32 cmd;
	struct adapter *adapter = dev->priv;

	if (copy_from_user(&cmd, useraddr, sizeof(cmd)))
                return -EFAULT;

	switch (cmd) {
	case TOETOOL_SETREG: {
		struct toetool_reg edata;

		if (!capable(CAP_NET_ADMIN))
			return -EPERM;
		if (copy_from_user(&edata, useraddr, sizeof(edata)))
			return -EFAULT;
		
		if (edata.addr == 10000) {
			t1_sched_set_max_avail_bytes(adapter->sge, edata.val);
			break;
		}
		if (edata.addr == 10004) {
			unsigned int port = edata.val >> 28;
			unsigned int drain = edata.val & 0xFFFFFFF;
			if (port > 3)
				break;
			t1_sched_set_drain_bits_per_us(adapter->sge, port, 
						       drain);
			break;
		}
		
		if ((edata.addr & 3) != 0 || edata.addr >= adapter->mmio_len)
			return -EINVAL;
		if (edata.addr == A_ESPI_MISC_CONTROL)
			t1_espi_set_misc_ctrl(adapter, edata.val);
		else 
			writel(edata.val, adapter->regs + edata.addr);
		break;
	}
	case TOETOOL_GETREG: {
		struct toetool_reg edata;

		if (copy_from_user(&edata, useraddr, sizeof(edata)))
			return -EFAULT;
		if ((edata.addr & 3) != 0 || edata.addr >= adapter->mmio_len)
			return -EINVAL;
		if (edata.addr >= 0x900 && edata.addr <= 0x93c)
			edata.val = t1_espi_get_mon(adapter, edata.addr, 1);
		else
			edata.val = readl(adapter->regs + edata.addr);
		if (copy_to_user(useraddr, &edata, sizeof(edata)))
			return -EFAULT;
		break;
	}
	case TOETOOL_SETTPI: {
		struct toetool_reg edata;

		if (!capable(CAP_NET_ADMIN))
			return -EPERM;
		if (copy_from_user(&edata, useraddr, sizeof(edata)))
			return -EFAULT;
		if ((edata.addr & 3) != 0)
			return -EINVAL;
		t1_tpi_write(adapter, edata.addr, edata.val);
		break;
	}
	case TOETOOL_GETTPI: {
		struct toetool_reg edata;

		if (copy_from_user(&edata, useraddr, sizeof(edata)))
			return -EFAULT;
		if ((edata.addr & 3) != 0)
			return -EINVAL;
		t1_tpi_read(adapter, edata.addr, &edata.val);
		if (copy_to_user(useraddr, &edata, sizeof(edata)))
			return -EFAULT;
		break;
	}
	case TOETOOL_T1POWERSAVE: {
		struct toetool_t1powersave edata;
		int err;

		if (!capable(CAP_NET_ADMIN))
			return -EPERM;
		if (copy_from_user(&edata, useraddr, sizeof(edata)))
			return -EFAULT;

		err = t1_clock(adapter, edata.mode);
		if (err != 0)
			return err;

		if (edata.mode & 2) {
			edata.mode = adapter->t1_powersave;
			edata.mode |= 2;
			if (copy_to_user(useraddr, &edata, sizeof(edata)))
				return -EFAULT;
		}
		break;
	}
#ifdef CONFIG_CHELSIO_T1_OFFLOAD
	case TOETOOL_DEVUP:
		if (!capable(CAP_NET_ADMIN))
			return -EPERM;
		return activate_toedev(adapter->tdev);
	case TOETOOL_GETMTUTAB: {
		struct toetool_mtus m = { .cmd = TOETOOL_GETMTUTAB };

		m.nmtus = NMTUS;
		memcpy(m.mtus, adapter->params.mtus, sizeof(m.mtus));
		return copy_to_user(useraddr, &m, sizeof(m)) ? -EFAULT : 0;
	}
	case TOETOOL_SETMTUTAB: {
		struct toetool_mtus m;
		int i;

		if (!capable(CAP_NET_ADMIN))
			return -EPERM;
		if (copy_from_user(&m, useraddr, sizeof(m)))
			return -EFAULT;
		if (m.nmtus != NMTUS)
			return -EINVAL;

		// MTUs must be in ascending order
		for (i = 1; i < NMTUS; ++i)
			if (m.mtus[i] < m.mtus[i - 1]) return -EINVAL;

		memcpy(adapter->params.mtus, m.mtus,
		       sizeof(adapter->params.mtus));
		if (toe_running(adapter))
			update_mtu_tab(adapter);
		break;
	}
	case TOETOOL_GETMTU: {
		struct toetool_reg edata;

		if (copy_from_user(&edata, useraddr, sizeof(edata)))
			return -EFAULT;
		edata.val = t1_read_reg_4(adapter, MTUREG(NMTUS - 1));
		if (copy_to_user(useraddr, &edata, sizeof(edata)))
			return -EFAULT;
		break;
	}
	case TOETOOL_GET_PM: {
		struct toetool_pm m = { .cmd = TOETOOL_GET_PM };

		m.tx_pg_sz  = adapter->params.tp.pm_tx_pg_size;
		m.tx_num_pg = adapter->params.tp.pm_tx_num_pgs;
		m.rx_pg_sz  = adapter->params.tp.pm_rx_pg_size;
		m.rx_num_pg = adapter->params.tp.pm_rx_num_pgs;
		m.pm_total  = adapter->params.tp.pm_size -
			adapter->params.tp.pm_tx_base;
		if (copy_to_user(useraddr, &m, sizeof(m)))
			return -EFAULT;
		break;
	}
	case TOETOOL_SET_PM: {
		struct toetool_pm m;
		unsigned int rx_base;

		if (!adapter->mc4)
			return -EOPNOTSUPP;
		if (!capable(CAP_NET_ADMIN))
			return -EPERM;
		if (adapter->flags & FULL_INIT_DONE)
			return -EBUSY;
		if (copy_from_user(&m, useraddr, sizeof(m)))
			return -EFAULT;
		if (!valid_pm_page_size(m.tx_pg_sz) ||
		    !valid_pm_page_size(m.rx_pg_sz))
			return -EINVAL;
		if (m.tx_num_pg == -1)
			m.tx_num_pg = adapter->params.tp.pm_tx_num_pgs;
		if (m.rx_num_pg == -1)
			m.rx_num_pg = adapter->params.tp.pm_rx_num_pgs;
		if (m.tx_num_pg % 24 || m.rx_num_pg % 24)
			return -EINVAL;
		rx_base = adapter->params.tp.pm_tx_base +
			m.tx_num_pg * m.tx_pg_sz;
		rx_base = (rx_base + m.rx_pg_sz - 1) & ~(m.rx_pg_sz - 1);
		if (rx_base + m.rx_num_pg * m.rx_pg_sz >
		    adapter->params.tp.pm_size)
			return -EINVAL;
		adapter->params.tp.pm_rx_base    = rx_base;
		adapter->params.tp.pm_rx_pg_size = m.rx_pg_sz;
		adapter->params.tp.pm_tx_pg_size = m.tx_pg_sz;
		adapter->params.tp.pm_rx_num_pgs = m.rx_num_pg;
		adapter->params.tp.pm_tx_num_pgs = m.tx_num_pg;
		break;
	}
	case TOETOOL_GET_TCAM: {
		struct toetool_tcam t = { .cmd = TOETOOL_GET_TCAM };

		if (!adapter->mc5)
			return -EOPNOTSUPP;
		t.tcam_size = t1_mc5_get_tcam_size(adapter->mc5);
		t.nservers = adapter->params.mc5.nservers;
		t.nroutes = adapter->params.mc5.nroutes;
		if (copy_to_user(useraddr, &t, sizeof(t)))
			return -EFAULT;
		break;
	}
	case TOETOOL_SET_TCAM: {
		struct toetool_tcam t;
		unsigned int tcam_size;

		if (!adapter->mc5)
			return -EOPNOTSUPP;
		if (!capable(CAP_NET_ADMIN))
			return -EPERM;
		if (adapter->flags & FULL_INIT_DONE)
			return -EBUSY;
		if (copy_from_user(&t, useraddr, sizeof(t)))
			return -EFAULT;
		tcam_size = t1_mc5_get_tcam_size(adapter->mc5);
		if (t.nroutes > 2048 || t.nroutes + t.nservers > tcam_size)
			return -EINVAL;
		adapter->params.mc5.nroutes = t.nroutes;
		adapter->params.mc5.nservers = t.nservers;
		break;
	}
	case TOETOOL_READ_TCAM_WORD: {
		int ret;
		struct toetool_tcam_word t;

		if (!adapter->mc5)
			return -EOPNOTSUPP;
		if (copy_from_user(&t, useraddr, sizeof(t)))
			return -EFAULT;
		ret = t1_read_mc5_range(adapter->mc5, t.addr, 1, t.buf);
		if (ret)
			return ret;
		if (copy_to_user(useraddr, &t, sizeof(t)))
			return -EFAULT;
		break;
	}
	case TOETOOL_GET_TCB: {
		struct toetool_tcb t;
		int ret;

		if (!adapter->mc4)
			return -EOPNOTSUPP;
		if (!(adapter->flags & FULL_INIT_DONE)) /* need MC4 for this */
			return -EIO;
		if (copy_from_user(&t, useraddr, sizeof(t)))
			return -EFAULT;
		ret = t1_mc4_bd_read(adapter->mc4,
				     t.tcb_index * (TCB_SIZE / 32),
				     TCB_SIZE / 32, t.tcb_data);
		if (ret)
			return ret;
		if (copy_to_user(useraddr, &t, sizeof(t)))
			return -EFAULT;
		break;
	}
#endif
	default:
		CH_WARN("%s: unknown chtoetool command %d\n",
		        adapter_name(adapter), cmd);
		return -EOPNOTSUPP;
	}
	return 0;
}

static int t1_ioctl(struct net_device *dev, struct ifreq *req, int cmd)
{
        struct adapter *adapter = dev->priv;
        struct mii_ioctl_data *data = (struct mii_ioctl_data *)&req->ifr_data;

	switch (cmd) {
        case SIOCGMIIPHY:
                data->phy_id = adapter->port[dev->if_port].phy->addr;
                /* FALLTHRU */
        case SIOCGMIIREG: {
		struct cphy *phy = adapter->port[dev->if_port].phy;
		u32 val;

		if (!phy->mdio_read) return -EOPNOTSUPP;
		phy->mdio_read(adapter, data->phy_id, 0, data->reg_num & 0x1f,
			       &val);
                data->val_out = val;
                break;
	}
        case SIOCSMIIREG: {
		struct cphy *phy = adapter->port[dev->if_port].phy;

                if (!capable(CAP_NET_ADMIN)) return -EPERM;
		if (!phy->mdio_write) return -EOPNOTSUPP;
		phy->mdio_write(adapter, data->phy_id, 0, data->reg_num & 0x1f,
			        data->val_in);
                break;
	}

	case SIOCTOETOOL:
		return toetool_ioctl(dev, (void *)req->ifr_data);
	default:
		return -EOPNOTSUPP;
	}
	return 0;
}

#ifdef CONFIG_CHELSIO_T1_OFFLOAD
void scale_connections(struct adapter *adapter, struct net_device *dev, int index)
{
	struct sock *conn_sk;
	struct net_device *conn_dev;
	struct toedev *tdev;
	struct tom_data *tom_d;
	struct tid_info *tid_maps;
	int max_tid;
	int tid;
	struct sk_buff *skb;
	struct cpl_set_tcb_field *req;
	int toe_mtu;
	int old_mtu_idx;

	tdev = adapter->tdev;
	tom_d = TOM_DATA(tdev);
	tid_maps = &tom_d->tid_maps;
			
	max_tid = t1_mc5_get_tcam_size(adapter->mc5) -
			adapter->params.mc5.nroutes -
			adapter->params.mc5.nservers;
	toe_mtu = adapter->params.mtus[index];

	for (tid = 0; tid < max_tid; tid++) {
		conn_sk = tid_maps->tid_tab[tid];
		old_mtu_idx = tid_maps->mss_tab[tid];
		if (conn_sk == NULL)
			break;
		conn_dev = __sk_dst_get(conn_sk)->dev;
		if ((conn_dev == dev) && (index < old_mtu_idx)) {
			/*
			 * Now update that Connection to the new MTU.
			 */
#ifdef __GFP_NOFAIL
			skb = alloc_skb(sizeof(*req), GFP_KERNEL | __GFP_NOFAIL);
#else
			for (;;) {
				skb = alloc_skb(sizeof(*req), GFP_KERNEL);
				if (skb)
					break;
				yield();
			}
#endif
			req = (struct cpl_set_tcb_field *)__skb_put(skb, sizeof(*req));
			OPCODE_TID(req) = htonl(MK_OPCODE_TID(CPL_SET_TCB_FIELD, tid));
			req->offset = W_TCB_MAXSEG;
			req->mask = htonl(M_TCB_MAXSEG << S_TCB_MAXSEG);
			req->val = htonl(index << S_TCB_MAXSEG);
			tid_maps->mss_tab[tid] = index;
			toe_send(tdev, skb);
		}
	}

}
#endif

static int t1_change_mtu(struct net_device *dev, int new_mtu)
{
	int ret;
	struct adapter *adapter = dev->priv;
	struct cmac *mac = adapter->port[dev->if_port].mac;
	int mtu_idx;
	int old_mtu_idx;

	if (!mac->ops->set_mtu) return -EOPNOTSUPP;
	if (new_mtu < 68) return -EINVAL;
	if ((ret = mac->ops->set_mtu(mac, new_mtu)))
		return ret;
#ifdef CONFIG_CHELSIO_T1_OFFLOAD
	if (toe_running(adapter)) {
		/* Get the best MTU */
		for (mtu_idx = (NMTUS - 1); mtu_idx > 0; mtu_idx--)
			if (adapter->params.mtus[mtu_idx] <= new_mtu)
				break;
		for (old_mtu_idx = (NMTUS - 1); old_mtu_idx > 0; old_mtu_idx--)
			if (adapter->params.mtus[old_mtu_idx] <= dev->mtu)
				break;

		if (mtu_idx < old_mtu_idx) {
			if ((adapter->params.nports > 1) &&
					(new_mtu != adapter->params.mtus[mtu_idx])) {
			printk(KERN_INFO
				"TOE connection cannot use %d size MTU, using %d instead.\n",
				new_mtu, adapter->params.mtus[mtu_idx]);
			} else {
				/*
				 * Creates a mismatch between the software MTU
				 * Table verses what's in the hardware. This
				 * needs to be fixed with a second default table
				 * which is part of the adapter structure and
				 * not global.
				 */
				init_mtus(adapter->params.mtus);
				update_mtu_tab(adapter);
				t1_write_reg_4(adapter, MTUREG(mtu_idx), new_mtu);
			}
			scale_connections(adapter, dev, mtu_idx);
		} else if ((mtu_idx >= old_mtu_idx) &&
					(adapter->params.nports == 1)) {
			init_mtus(adapter->params.mtus);
			update_mtu_tab(adapter);
			t1_write_reg_4(adapter, MTUREG(mtu_idx), new_mtu);
			if ((mtu_idx == old_mtu_idx) &&
				(new_mtu > dev->mtu)) {
				if (mtu_idx != 0)
					mtu_idx--;
				scale_connections(adapter, dev, mtu_idx);
			}
		}
	}
#endif
	dev->mtu = new_mtu;
	return 0;
}

static int t1_set_mac_addr(struct net_device *dev, void *p)
{
	struct adapter *adapter = dev->priv;
	struct cmac *mac = adapter->port[dev->if_port].mac;
	struct sockaddr *addr = p;

	if (!mac->ops->macaddress_set)
		return -EOPNOTSUPP;

	memcpy(dev->dev_addr, addr->sa_data, dev->addr_len);
	mac->ops->macaddress_set(mac, dev->dev_addr);
#ifdef CONFIG_CHELSIO_T1_OFFLOAD
	if (toe_running(adapter))
		write_smt_entry(adapter, dev->if_port,
			 (t1_is_asic(adapter) ? NMTUS-1:5));
#endif
	return 0;
}

#if defined(CONFIG_VLAN_8021Q) || defined(CONFIG_VLAN_8021Q_MODULE)
static void vlan_rx_register(struct net_device *dev,
				   struct vlan_group *grp)
{
	struct adapter *adapter = dev->priv;

	spin_lock_irq(&adapter->async_lock);
	adapter->vlan_grp = grp;
	t1_set_vlan_accel(adapter, grp != NULL);
	spin_unlock_irq(&adapter->async_lock);
}

static void vlan_rx_kill_vid(struct net_device *dev, unsigned short vid)
{
	struct adapter *adapter = dev->priv;

	spin_lock_irq(&adapter->async_lock);
	if (adapter->vlan_grp)
		adapter->vlan_grp->vlan_devices[vid] = NULL;
	spin_unlock_irq(&adapter->async_lock);
}
#endif

#ifdef CONFIG_NET_POLL_CONTROLLER
static void t1_netpoll(struct net_device *dev)
{
	unsigned long flags;
	struct adapter *adapter = dev->priv;

	local_irq_save(flags);
        t1_select_intr_handler(adapter)(adapter->pdev->irq, adapter, NULL);
	local_irq_restore(flags);
}
#endif

/*
 * Periodic accumulation of MAC statistics.  This is used only if the MAC
 * does not have any other way to prevent stats counter overflow.
 */
static void mac_stats_task(void *data)
{
	int i;
	struct adapter *adapter = data;

	for_each_port(adapter, i) {
		struct port_info *p = &adapter->port[i];

		if (netif_running(p->dev))
			p->mac->ops->statistics_update(p->mac,
						       MAC_STATS_UPDATE_FAST);
	}

	/* Schedule the next statistics update if any port is active. */
	spin_lock(&adapter->work_lock);
	if (adapter->open_device_map & PORT_MASK)
		schedule_mac_stats_update(adapter,
					  adapter->params.stats_update_period);
	spin_unlock(&adapter->work_lock);
}

/*
 * Processes elmer0 external interrupts in process context.
 */
static void ext_intr_task(void *data)
{
	struct adapter *adapter = data;

	elmer0_ext_intr_handler(adapter);

	// Now reenable external interrupts
	spin_lock_irq(&adapter->async_lock);
	adapter->slow_intr_mask |= F_PL_INTR_EXT;
	t1_write_reg_4(adapter, A_PL_CAUSE, F_PL_INTR_EXT);
	t1_write_reg_4(adapter, A_PL_ENABLE,
		       adapter->slow_intr_mask | F_PL_INTR_SGE_DATA);
	spin_unlock_irq(&adapter->async_lock);
}

/*
 * Interrupt-context handler for elmer0 external interrupts.
 */
void t1_os_elmer0_ext_intr(struct adapter *adapter)
{
	/*
	 * Schedule a task to handle external interrupts as we require
	 * a process context.  We disable EXT interrupts in the interim
	 * and let the task reenable them when it's done.
	 */
	adapter->slow_intr_mask &= ~F_PL_INTR_EXT;
	t1_write_reg_4(adapter, A_PL_ENABLE,
		       adapter->slow_intr_mask | F_PL_INTR_SGE_DATA);
	schedule_work(&adapter->ext_intr_handler_task);
}

void t1_fatal_err(struct adapter *adapter)
{
	if (adapter->flags & FULL_INIT_DONE) {
		t1_sge_stop(adapter->sge);
		t1_interrupts_disable(adapter);
	}
	CH_ALERT("%s: encountered fatal error, operation suspended\n",
		 adapter_name(adapter));
}

static int __devinit init_one(struct pci_dev *pdev,
			      const struct pci_device_id *ent)
{
	static int version_printed;

	int i, err, pci_using_dac = 0;
	unsigned long mmio_start, mmio_len;
	const struct board_info *bi;
	struct adapter *adapter = NULL;
	struct port_info *pi;

	if (!version_printed) {
		printk(KERN_INFO "%s - version %s\n", driver_string,
		       driver_version);
		++version_printed;
	}

	err = pci_enable_device(pdev);
	if (err) return err;

	if (!(pci_resource_flags(pdev, 0) & IORESOURCE_MEM)) {
		CH_ERR("%s: cannot find PCI device memory base address\n",
		       pci_name(pdev));
		err = -ENODEV;
		goto out_disable_pdev;
	}

	if (!pci_set_dma_mask(pdev, PCI_DMA_64BIT)) {
		pci_using_dac = 1;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,5,0)
		if (pci_set_consistent_dma_mask(pdev, PCI_DMA_64BIT)) {
			CH_ERR("%s: unable to obtain 64-bit DMA for"
			       "consistent allocations\n", pci_name(pdev));
			err = -ENODEV;
			goto out_disable_pdev;
		}
#endif
	} else if ((err = pci_set_dma_mask(pdev, PCI_DMA_32BIT)) != 0) {
		CH_ERR("%s: no usable DMA configuration\n", pci_name(pdev));
		goto out_disable_pdev;
	}

	err = pci_request_regions(pdev, driver_name);
	if (err) {
		CH_ERR("%s: cannot obtain PCI resources\n", pci_name(pdev));
		goto out_disable_pdev;
	}

	pci_set_master(pdev);

	// XXX TBD: power management stuff
#ifdef NOTYET
	pm = pci_find_capability(pdev, PCI_CAP_ID_PM);
	if (pm) {
		u16 pwr_command;
		pci_read_config_word(pdev, pm + PCI_PM_CTRL, &pwr_command);
		acpi_idle_state = pwr_command & PCI_PM_CTRL_STATE_MASK;
	}
#endif

	mmio_start = pci_resource_start(pdev, 0);
	mmio_len = pci_resource_len(pdev, 0);
	bi = t1_get_board_info(ent->driver_data);

	for (i = 0; i < bi->port_number; ++i) {
		struct net_device *netdev;

		netdev = alloc_etherdev(adapter ? 0 : sizeof(*adapter));
		if (!netdev) {
			err = -ENOMEM;
			goto out_free_dev;
		}

		SET_MODULE_OWNER(netdev);
#ifdef SET_NETDEV_DEV
		SET_NETDEV_DEV(netdev, &pdev->dev);
#endif
		if (!adapter) {
			adapter = netdev->priv;
			adapter->pdev = pdev;
			adapter->port[0].dev = netdev;  // so we don't leak it

			adapter->regs = ioremap(mmio_start, mmio_len);
			if (!adapter->regs) {
				CH_ERR("%s: cannot map device registers\n",
				       pci_name(pdev));
				err = -ENOMEM;
				goto out_free_dev;
			}

			if (t1_get_board_rev(adapter, bi, &adapter->params)) {
				err = -ENODEV;	  // Can't handle this chip rev
				goto out_free_dev;
			}

			adapter->name = pci_name(pdev);
			adapter->msg_enable = dflt_msg_enable;
			adapter->mmio_len = mmio_len;

			init_MUTEX(&adapter->mib_mutex);
			spin_lock_init(&adapter->tpi_lock);
			spin_lock_init(&adapter->work_lock);
			spin_lock_init(&adapter->async_lock);
			spin_lock_init(&adapter->mac_lock);

			INIT_WORK(&adapter->ext_intr_handler_task,
				  ext_intr_task, adapter);
			INIT_WORK(&adapter->stats_update_task, mac_stats_task,
				  adapter);
#ifdef work_struct
			init_timer(&adapter->stats_update_timer);
			adapter->stats_update_timer.function = mac_stats_timer;
			adapter->stats_update_timer.data =
				(unsigned long)adapter;
#endif

			pci_set_drvdata(pdev, netdev);

			if (bi->clock_mc4)
				adapter->tdev = chtoe_setup_toedev(adapter);
		}

		pi = &adapter->port[i];
		pi->dev = netdev;
		netif_carrier_off(netdev);
		netdev->irq = pdev->irq;
		netdev->if_port = i;
		netdev->mem_start = mmio_start;
		netdev->mem_end = mmio_start + mmio_len - 1;
		netdev->priv = adapter;
#ifdef CONFIG_CHELSIO_T1_OFFLOAD
		TOEDEV(netdev) = adapter->tdev;
#endif
		netdev->features |= NETIF_F_SG | NETIF_F_IP_CSUM;

#ifdef NETIF_F_LLTX
		netdev->features |= NETIF_F_LLTX;
#endif

		adapter->flags |= RX_CSUM_ENABLED | TCP_CSUM_CAPABLE;
		if (pci_using_dac)
			netdev->features |= NETIF_F_HIGHDMA;
		if (vlan_tso_capable(adapter)) {
#if defined(CONFIG_VLAN_8021Q) || defined(CONFIG_VLAN_8021Q_MODULE)
			adapter->flags |= VLAN_ACCEL_CAPABLE;
			netdev->features |=
				NETIF_F_HW_VLAN_TX | NETIF_F_HW_VLAN_RX;
			netdev->vlan_rx_register = vlan_rx_register;
			netdev->vlan_rx_kill_vid = vlan_rx_kill_vid;
#endif
#ifdef NETIF_F_TSO
			/* T204: disable TSO */
			if ( !(is_T2(adapter)) || bi->port_number != 4 ) {
				adapter->flags |= TSO_CAPABLE;
				netdev->features |= NETIF_F_TSO;
			} 
#endif
		}

		netdev->open = chtoe_open;
		netdev->stop = chtoe_close;
		netdev->hard_start_xmit = t1_start_xmit;
		netdev->hard_header_len += (adapter->flags & TSO_CAPABLE) ?
			sizeof(struct cpl_tx_pkt_lso) :
			sizeof(struct cpl_tx_pkt);
		netdev->get_stats = t1_get_stats;
		netdev->set_multicast_list = t1_set_rxmode;
		netdev->do_ioctl = t1_ioctl;
		netdev->change_mtu = t1_change_mtu;
		netdev->set_mac_address = t1_set_mac_addr;
#ifdef CONFIG_NET_POLL_CONTROLLER
		netdev->poll_controller = t1_netpoll;
#endif
		netdev->weight = 64;
#ifdef SET_ETHTOOL_OPS
		SET_ETHTOOL_OPS(netdev, &t1_ethtool_ops);
#endif
	}

	if (t1_init_sw_modules(adapter, bi) < 0) {
		err = -ENODEV;
		goto out_free_dev;
	}

	/*
	 * The card is now ready to go.  If any errors occur during device
	 * registration we do not fail the whole card but rather proceed only
	 * with the ports we manage to register successfully.  However we must
	 * register at least one net device.
	 */
	for (i = 0; i < bi->port_number; ++i) {
		err = register_netdev(adapter->port[i].dev);
		if (err)
			CH_WARN("%s: cannot register net device %s, skipping\n",
				pci_name(pdev), adapter->port[i].dev->name);
		else {
			/*
			 * Change the name we use for messages to the name of
			 * the first successfully registered interface.
			 */
			if (!adapter->registered_device_map)
				adapter->name = adapter->port[i].dev->name;
#ifndef CONFIG_CHELSIO_T1_OFFLOAD
			chtoe_proc_setup(adapter, proc_root_driver);
#endif
			__set_bit(i, &adapter->registered_device_map);
		}
	}
	if (!adapter->registered_device_map) {
		CH_ERR("%s: could not register any net devices\n",
		       pci_name(pdev));
		goto out_release_adapter_res;
	}

#ifdef CONFIG_CHELSIO_T1_OFFLOAD
	if (adapter->tdev) {
		if (register_toedev(adapter->tdev, "toe%d"))
			CH_WARN("%s: unable to register TOE device, continuing"
				"without TOE support\n", pci_name(pdev));
		else {
			__set_bit(TOEDEV_DEVMAP_BIT,
				  &adapter->registered_device_map);
			chtoe_proc_setup(adapter, adapter->tdev->proc_dir);
		}
	}
#endif
	printk(KERN_INFO "%s: %s (rev %d), %s %dMHz/%d-bit\n", adapter->name,
	       bi->desc, adapter->params.chip_revision,
	       adapter->params.pci.is_pcix ? "PCIX" : "PCI",
	       adapter->params.pci.speed, adapter->params.pci.width);
#ifdef CONFIG_CHELSIO_T1_OFFLOAD
	if (adapter->params.tp.pm_size)
		printk(KERN_INFO "%s: %uMB SDRAM, %uMB FCRAM\n", adapter->name,
		       adapter->params.tp.pm_size >> 20,
		       adapter->params.tp.cm_size >> 20);
#endif

	/*
	 * Set the T1B ASIC and memory clocks.
	 */
	adapter->t1_powersave = LCLOCK; // HW default is powersave mode.
	if (t1_is_T1B(adapter))
		t1_clock(adapter, dflt_t1_powersave_mode);

	return 0;

 out_release_adapter_res:
	t1_free_sw_modules(adapter);
 out_free_dev:
	if (adapter) {
		if (adapter->tdev) kfree(adapter->tdev);
		if (adapter->regs) iounmap(adapter->regs);
		for (i = bi->port_number - 1; i >= 0; --i)
			if (adapter->port[i].dev) {
#ifndef CONFIG_CHELSIO_T1_OFFLOAD
				chtoe_proc_cleanup(adapter, proc_root_driver);
#endif
				free_netdev(adapter->port[i].dev);
			}
	}
	pci_release_regions(pdev);
 out_disable_pdev:
	pci_disable_device(pdev);
	pci_set_drvdata(pdev, NULL);
	return err;
}

void bit_bang(struct adapter *adapter, int bitdata, int nbits)
{
	int data;
	int i;
	u32 val;

	enum {
		S_CLOCK	    = 1 << 3,
		S_DATA	    = 1 << 4
	};

	for (i = (nbits - 1); i > -1; i--) {

		udelay(50);

		data = ((bitdata >> i) & 0x1);
		__t1_tpi_read(adapter, A_ELMER0_GPO, &val);

		if (data)
			val |= S_DATA;
		else
			val &= ~S_DATA;

		udelay(50);

		/* Set SCLOCK low */
		val &= ~S_CLOCK;
		__t1_tpi_write(adapter, A_ELMER0_GPO, val);

		udelay(50);

		/* Write SCLOCK high */
		val |= S_CLOCK;
		__t1_tpi_write(adapter, A_ELMER0_GPO, val);

	}
}

static int t1_clock(struct adapter *adapter, int mode)
{
	u32 val;
	int M_CORE_VAL;
	int M_MEM_VAL;

	enum {
		M_CORE_BITS = 9,
		T_CORE_VAL  = 0,
		T_CORE_BITS = 2,
		N_CORE_VAL  = 0,
		N_CORE_BITS = 2,
		M_MEM_BITS  = 9,
		T_MEM_VAL   = 0,
		T_MEM_BITS  = 2,
		N_MEM_VAL   = 0,
		N_MEM_BITS  = 2,
		NP_LOAD     = 1 << 17,
		S_LOAD_MEM  = 1 << 5,
		S_LOAD_CORE = 1 << 6,
		S_CLOCK	    = 1 << 3
	};

	if (!t1_is_T1B(adapter))
		return -ENODEV; /* Can't re-clock this chip. */

	if (mode & 2) {
		return 0; /* show current mode. */
	}

	if ((adapter->t1_powersave & 1) == (mode & 1))
		return -EALREADY; /* ASIC already running in mode. */

	if ((mode & 1) == HCLOCK) {
		M_CORE_VAL  = 0x14;
		M_MEM_VAL   = 0x18;
		adapter->t1_powersave = HCLOCK; /* overclock */
	} else {
		M_CORE_VAL  = 0xe;
		M_MEM_VAL   = 0x10;
		adapter->t1_powersave = LCLOCK; /* underclock */
	}

	/* Don't interrupt this serial stream! */
	TPI_LOCK(adapter);

	/* Initialize for ASIC core */
	__t1_tpi_read(adapter, A_ELMER0_GPO, &val);
	val |= NP_LOAD;
	udelay(50);
	__t1_tpi_write(adapter, A_ELMER0_GPO, val);
	udelay(50);
	__t1_tpi_read(adapter, A_ELMER0_GPO, &val);
	val &= ~S_LOAD_CORE;
	val &= ~S_CLOCK;
	__t1_tpi_write(adapter, A_ELMER0_GPO, val);
	udelay(50);
	
	/* Serial program the ASIC clock synthesizer */
	bit_bang(adapter, T_CORE_VAL, T_CORE_BITS);
	bit_bang(adapter, N_CORE_VAL, N_CORE_BITS);
	bit_bang(adapter, M_CORE_VAL, M_CORE_BITS);
	udelay(50);

	/* Finish ASIC core */
	__t1_tpi_read(adapter, A_ELMER0_GPO, &val);
	val |= S_LOAD_CORE;
	udelay(50);
	__t1_tpi_write(adapter, A_ELMER0_GPO, val);
	udelay(50);
	__t1_tpi_read(adapter, A_ELMER0_GPO, &val);
	val &= ~S_LOAD_CORE;
	udelay(50);
	__t1_tpi_write(adapter, A_ELMER0_GPO, val);
	udelay(50);

	/* Initialize for memory */
	__t1_tpi_read(adapter, A_ELMER0_GPO, &val);
	val |= NP_LOAD;
	udelay(50);
	__t1_tpi_write(adapter, A_ELMER0_GPO, val);
	udelay(50);
	__t1_tpi_read(adapter, A_ELMER0_GPO, &val);
	val &= ~S_LOAD_MEM;
	val &= ~S_CLOCK;
	udelay(50);
	__t1_tpi_write(adapter, A_ELMER0_GPO, val);
	udelay(50);
	
	/* Serial program the memory clock synthesizer */
	bit_bang(adapter, T_MEM_VAL, T_MEM_BITS);
	bit_bang(adapter, N_MEM_VAL, N_MEM_BITS);
	bit_bang(adapter, M_MEM_VAL, M_MEM_BITS);
	udelay(50);

	/* Finish memory */
	__t1_tpi_read(adapter, A_ELMER0_GPO, &val);
	val |= S_LOAD_MEM;
	udelay(50);
	__t1_tpi_write(adapter, A_ELMER0_GPO, val);
	udelay(50);
	__t1_tpi_read(adapter, A_ELMER0_GPO, &val);
	val &= ~S_LOAD_MEM;
	udelay(50);
	__t1_tpi_write(adapter, A_ELMER0_GPO, val);

	TPI_UNLOCK(adapter);
	return 0;
}

/*
 * This function initializes a polling routine, Poll_func
 * which will be polled ever N Microsecond, where N is
 * provided in the cyclic start routine.
 */
void ch_init_cyclic(void *adapter, p_ch_cyclic_t cyclic,
		void (*poll_func)(void *), void *arg)
{
	cyclic->func = poll_func;
	cyclic->arg = arg;
	INIT_WORK(&cyclic->task, (void (*)(void *))ch_cyclic, (void *)cyclic);
#ifdef work_struct
	init_timer(&cyclic->timer);
	cyclic->timer.function = ch_cyclic_timer;
	cyclic->timer.data = (unsigned long)cyclic;
#endif
}

/*
 * Cyclic function which provides a periodic polling
 * capability to Linux. The poll function provided by
 * the 'ch_init_cyclic' function is called from this
 * here, and this routine launce a new one-shot
 * timer to bring it back in some period later.
 */
void ch_cyclic(p_ch_cyclic_t cyclic)
{
	cyclic->func(cyclic->arg);
#ifdef work_struct
	mod_timer(&cyclic->timer, jiffies + cyclic->period);
#else
	schedule_delayed_work(&cyclic->task, cyclic->period);
#endif
}

/*
 * The 'ch_cyclic_timer' is legacy support routine
 * to allow the cyclic routine to run outside the 
 * interrupt context, and allow interrupts to 
 * remain enabled.
 */
#ifdef work_struct
void ch_cyclic_timer(p_ch_cyclic_t cyclic)
{
	schedule_task(&cyclic->task);
}
#endif

/*
 * The 'ch_start_cyclic' starts the polling.
 */
void ch_start_cyclic(p_ch_cyclic_t cyclic, unsigned long period)
{
	cyclic->period = period;
#ifdef work_struct
	mod_timer(&cyclic->timer, jiffies + period);
#else
	schedule_delayed_work(&cyclic->task, period);
#endif

}

/*
 * The 'ch_stop_cyclic' stops the polling.
 */
void ch_stop_cyclic(p_ch_cyclic_t cyclic)
{
#ifdef work_struct
	del_timer_sync(&cyclic->timer);
	flush_scheduled_tasks();
#else
	cancel_delayed_work(&cyclic->task);
#endif
}

static inline void t1_sw_reset(struct pci_dev *pdev)
{
	pci_write_config_dword(pdev, A_PCICFG_PM_CSR, 3);
	pci_write_config_dword(pdev, A_PCICFG_PM_CSR, 0);
}

static void __devexit remove_one(struct pci_dev *pdev)
{
	struct net_device *dev = pci_get_drvdata(pdev);

	if (dev) {
		int i;
		struct adapter *adapter = dev->priv;

#ifdef CONFIG_CHELSIO_T1_OFFLOAD
		if (test_bit(TOEDEV_DEVMAP_BIT,
			     &adapter->registered_device_map)) {
			chtoe_proc_cleanup(adapter, adapter->tdev->proc_dir);
			// XXX unregister toedev
		}
#endif
		for_each_port(adapter, i)
			if (test_bit(i, &adapter->registered_device_map))
				unregister_netdev(adapter->port[i].dev);

		t1_free_sw_modules(adapter);
		iounmap(adapter->regs);
		kfree(adapter->tdev);
		while (--i >= 0)
			if (adapter->port[i].dev) {
#ifndef CONFIG_CHELSIO_T1_OFFLOAD
				chtoe_proc_cleanup(adapter, proc_root_driver);
#endif
				free_netdev(adapter->port[i].dev);
			}
		pci_release_regions(pdev);
		pci_disable_device(pdev);
		pci_set_drvdata(pdev, NULL);
		t1_sw_reset(pdev);
	}
}

static struct pci_driver driver = {
	.name     = driver_name,
	.id_table = t1_pci_tbl,
	.probe    = init_one,
	.remove   = __devexit_p(remove_one),
#ifdef CONFIG_CHTOE_PM
	.suspend  = chtoe_suspend,
	.resume   = chtoe_resume
#endif
};

static int __init t1_init_module(void)
{
	return pci_module_init(&driver);
}

static void __exit t1_cleanup_module(void)
{
	pci_unregister_driver(&driver);
}

module_init(t1_init_module);
module_exit(t1_cleanup_module);
