/*****************************************************************************
 *                                                                           *
 * File:                                                                     *
 *  osdep.h                                                                  *
 *                                                                           *
 * Description:                                                              *
 *  Chelsio OS dependencies.                                                 *
 *                                                                           *
 * This program is free software; you can redistribute it and/or modify      *
 * it under the terms of the GNU General Public License, version 2, as       *
 * published by the Free Software Foundation.                                *
 *                                                                           *
 * You should have received a copy of the GNU General Public License along   *
 * with this program; if not, write to the Free Software Foundation, Inc.,   *
 * 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.                 *
 *                                                                           *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR IMPLIED    *
 * WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF      *
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.                     *
 *                                                                           *
 * http://www.chelsio.com                                                    *
 *                                                                           *
 * Copyright (c) 2003 - 2005 Chelsio Communications, Inc.                    *
 * All rights reserved.                                                      *
 *                                                                           *
 * Maintainers: maintainers@chelsio.com                                      *
 *                                                                           *
 * Authors: Dimitrios Michailidis   <dm@chelsio.com>                         *
 *          Tina Yang               <tainay@chelsio.com>                     *
 *          Felix Marti             <felix@chelsio.com>                      *
 *          Scott Bardone           <sbardone@chelsio.com>                   *
 *          Kurt Ottaway            <kottaway@chelsio.com>                   *
 *                                                                           *
 * History:                                                                  *
 *                                                                           *
 ****************************************************************************/
/* $Date: 2005/10/18 01:11:53 $ $RCSfile: osdep.h,v $ $Revision: 1.54 $ */

#ifndef __CHELSIO_OSDEP_H
#define __CHELSIO_OSDEP_H

#include <linux/version.h>
#if defined(MODULE) && ! defined(MODVERSIONS)
#define MODVERSIONS
#endif
#if defined(MODVERSIONS) && (LINUX_VERSION_CODE < KERNEL_VERSION(2,5,0))
#include <linux/modversions.h>
#endif
#ifdef MODULE
#include <linux/module.h>
#endif

#include <linux/config.h>
#include <linux/types.h>
#include <linux/delay.h>
#include <linux/pci.h>
#include <linux/ethtool.h>
#include <linux/mii.h>
#include <linux/crc32.h>
#include <linux/init.h>
#include <asm/io.h>

#include "chtoe.h"
#include "version.h"

#define PFX      DRV_NAME ": "

#define CH_ERR(fmt, ...)   printk(KERN_ERR PFX fmt, ## __VA_ARGS__)
#define CH_WARN(fmt, ...)  printk(KERN_WARNING PFX fmt, ## __VA_ARGS__)
#define CH_ALERT(fmt, ...) printk(KERN_ALERT PFX fmt, ## __VA_ARGS__)

/* FILE_IDENT Function */
#define FILE_IDENT(a) static char *id __attribute (( unused )) = a

/*
 * Fix for skb_checksum_help.
 * Linux <2.6.7  number of args is 1 and arg is *skb.
 * Linux 2.6.7-9 changed the number of args to 2 and first arg is **pskb.
 * Linux 2.6.10+ changes first arg back to *skb.
 */
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,7)
#define msg_checksum_help(a, b) (skb_checksum_help(a) == NULL)
#elif LINUX_VERSION_CODE < KERNEL_VERSION(2,6,10)
#define msg_checksum_help(a, b) (skb_checksum_help(&a, b) != 0)
#else
#define msg_checksum_help(a, b) (skb_checksum_help(a, b) != 0)
#endif

/* 
 * Use pci_dma_sync_single_for_cpu which is properly defined in 2.6.0+ kernels,
 */
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
#define pci_dma_sync_single_for_cpu pci_dma_sync_single
#endif

/*
 * More powerful macro that selectively prints messages based on msg_enable.
 * For info and debugging messages.
 */
#define CH_MSG(adapter, level, category, fmt, ...) do { \
	if ((adapter)->msg_enable & NETIF_MSG_##category) \
		printk(KERN_##level PFX "%s: " fmt, (adapter)->name, \
		       ## __VA_ARGS__); \
} while (0)

#ifdef DEBUG
# define CH_DBG(adapter, category, fmt, ...) \
	CH_MSG(adapter, DEBUG, category, fmt, ## __VA_ARGS__)
#else
# define CH_DBG(fmt, ...)
#endif

/* Additional NETIF_MSG_* categories */
#define NETIF_MSG_MMIO 0x8000000

#define CH_DEVICE(devid, ssid, idx) \
	{ PCI_VENDOR_ID_CHELSIO, devid, PCI_ANY_ID, ssid, 0, 0, idx }

#define SUPPORTED_PAUSE       (1 << 13)
#define SUPPORTED_LOOPBACK    (1 << 15)

#define ADVERTISED_PAUSE      (1 << 13)
#define ADVERTISED_ASYM_PAUSE (1 << 14)

/*
 * Now that we have included the driver's main data structure,
 * we typedef it to something the rest of the system understands.
 */
typedef struct adapter adapter_t;

#define DELAY_US(x) udelay(x)

#define SPINLOCK		spinlock_t
#define SPIN_LOCK_INIT(lock)	spin_lock_init(&(lock))
#define SPIN_LOCK(lock)		spin_lock(&(lock))
#define SPIN_TRYLOCK(lock)	spin_trylock(&(lock))
#define SPIN_UNLOCK(lock)	spin_unlock(&(lock))
#define SPIN_LOCK_DESTROY(lock)

#define TPI_LOCK(adapter) spin_lock(&(adapter)->tpi_lock)
#define TPI_UNLOCK(adapter) spin_unlock(&(adapter)->tpi_lock)
#define MAC_LOCK(lock)		spin_lock_bh(&(lock))
#define MAC_UNLOCK(lock)	spin_unlock_bh(&(lock))

void t1_os_elmer0_ext_intr(adapter_t *adapter);
void t1_os_link_changed(adapter_t *adapter, int port_id, int link_status,
			int speed, int duplex, int fc);

static inline void DELAY_MS(unsigned long ms)
{
	unsigned long ticks = (ms * HZ + 999) / 1000 + 1;

	while (ticks) {
		set_current_state(TASK_UNINTERRUPTIBLE);
		ticks = schedule_timeout(ticks);
	}
}

static inline u16 t1_read_reg_2(adapter_t *adapter, u32 reg_addr)
{
	u16 val = readw(adapter->regs + reg_addr);

	CH_DBG(adapter, MMIO, "read register 0x%x value 0x%x\n", reg_addr,
	       val);
	return val;
}

static inline void t1_write_reg_2(adapter_t *adapter, u32 reg_addr, u16 val)
{
	CH_DBG(adapter, MMIO, "setting register 0x%x to 0x%x\n", reg_addr,
	       val);
	writew(val, adapter->regs + reg_addr);
}

static inline u32 t1_read_reg_4(adapter_t *adapter, u32 reg_addr)
{
	u32 val = readl(adapter->regs + reg_addr);

	CH_DBG(adapter, MMIO, "read register 0x%x value 0x%x\n", reg_addr,
	       val);
	return val;
}

static inline void t1_write_reg_4(adapter_t *adapter, u32 reg_addr, u32 val)
{
	CH_DBG(adapter, MMIO, "setting register 0x%x to 0x%x\n", reg_addr,
	       val);
	writel(val, adapter->regs + reg_addr);
}

static inline void *t1_os_malloc_wait_zero(size_t len)
{
	void *m = kmalloc(len, GFP_KERNEL);
	if (m)
		memset(m, 0, len);
	return m;
}

static inline void t1_os_free(void *v, size_t len)
{
	kfree(v);
}

static inline void t1_os_pci_write_config_4(adapter_t *adapter, int reg,
					    u32 val)
{
	pci_write_config_dword(adapter->pdev, reg, val);
}

static inline void t1_os_pci_read_config_4(adapter_t *adapter, int reg,
					   u32 *val)
{
	pci_read_config_dword(adapter->pdev, reg, val); 
}

static inline void t1_os_pci_write_config_2(adapter_t *adapter, int reg,
					    u16 val)
{
	pci_write_config_word(adapter->pdev, reg, val);
}

static inline void t1_os_pci_read_config_2(adapter_t *adapter, int reg,
					   u16 *val)
{
	pci_read_config_word(adapter->pdev, reg, val); 
}

static inline const char *adapter_name(adapter_t *adapter)
{
	return adapter->name;
}

static inline const char *port_name(adapter_t *adapter, int port_idx)
{
	return adapter->port[port_idx].dev->name;
}

static inline void t1_os_set_hw_addr(adapter_t *adapter, int port_idx,
				     u8 hw_addr[])
{
	memcpy(adapter->port[port_idx].dev->dev_addr, hw_addr, ETH_ALEN);
}

struct t1_rx_mode {
	struct net_device *dev;
	u32 idx;
	struct dev_mc_list *list;
};

#define t1_rx_mode_promisc(rm)	(rm->dev->flags & IFF_PROMISC) 
#define t1_rx_mode_allmulti(rm)	(rm->dev->flags & IFF_ALLMULTI) 
#define t1_rx_mode_mc_cnt(rm)	(rm->dev->mc_count) 

static inline u8 *t1_get_next_mcaddr(struct t1_rx_mode *rm)
{
	u8 *addr = 0;

	if (rm->idx++ < rm->dev->mc_count) {
		addr = rm->list->dmi_addr;
		rm->list = rm->list->next;
	}
	return addr;
}

typedef struct ch_cyclic_s {
	struct work_struct task;
#ifdef work_struct
        struct timer_list timer;
#endif
	void (*func)(void *);
	void *arg;
	unsigned long period;
} ch_cyclic_t, *p_ch_cyclic_t;

void ch_init_cyclic(void *, p_ch_cyclic_t, void (*)(void *), void *);
void ch_start_cyclic(p_ch_cyclic_t, unsigned long);
void ch_stop_cyclic(p_ch_cyclic_t);
void ch_cyclic(p_ch_cyclic_t);
#ifdef work_struct
void ch_cyclic_timer(p_ch_cyclic_t);
#endif

#endif /* __CHELSIO_OSDEP_H */
