/*****************************************************************************
 *                                                                           *
 * File:                                                                     *
 *  toe_compat.h                                                             *
 *                                                                           *
 * Description:                                                              *
 *  Chelsio 10Gb Ethernet TOE driver                                         *
 *                                                                           *
 * This program is free software; you can redistribute it and/or modify      *
 * it under the terms of the GNU General Public License, version 2, as       *
 * published by the Free Software Foundation.                                *
 *                                                                           *
 * You should have received a copy of the GNU General Public License along   *
 * with this program; if not, write to the Free Software Foundation, Inc.,   *
 * 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.                 *
 *                                                                           *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR IMPLIED    *
 * WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF      *
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.                     *
 *                                                                           *
 * http://www.chelsio.com                                                    *
 *                                                                           *
 * Copyright (c) 2003 - 2005 Chelsio Communications, Inc.                    *
 * All rights reserved.                                                      *
 *                                                                           *
 * Maintainers: maintainers@chelsio.com                                      *
 *                                                                           *
 * Authors: Dimitrios Michailidis   <dm@chelsio.com>                         *
 *          Felix Marti             <felix@chelsio.com>                      *
 *          Karen Xie               <kxie@chelsio.com>                       *
 *                                                                           *
 * History:                                                                  *
 *                                                                           *
 ****************************************************************************/
/* $Date: 2005/07/09 02:47:14 $ $RCSfile: toe_compat.h,v $ $Revision: 1.20 $ */

#ifndef __TOE_COMPAT_H
#define __TOE_COMPAT_H

#include <linux/version.h>

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,5,0)

#define SEQ_START_TOKEN ((void *)1)

#define tcp_sk(sk) (&(sk)->tp_pinfo.af_tcp)
#define inet_sk(sk) (&(sk)->protinfo.af_inet)
#define inet_addr_info(sk) sk
#define sock_owned_by_user(sk) ((sk)->lock.users)

#define TCP_NAGLE_OFF  1
#define TCP_NAGLE_CORK 2

#define sk_user_data user_data
#define sk_route_caps route_caps
#define sk_dst_cache dst_cache
#define sk_state_change state_change
#define sk_err err
#define sk_err_soft err_soft
#define sk_error_report error_report
#define sk_prot prot
#define sk_receive_queue receive_queue
#define sk_backlog backlog
#define sk_data_ready data_ready
#define sk_backlog_rcv backlog_rcv
#define sk_sleep sleep
#define sk_wmem_queued wmem_queued
#define sk_write_space write_space
#define sk_shutdown shutdown
#define sk_socket socket
#define sk_forward_alloc forward_alloc
#define sk_sndbuf sndbuf
#define sk_state state
#define sk_write_queue write_queue
#define sk_allocation allocation
#define sk_family family
#define sk_userlocks userlocks
#define sk_refcnt refcnt
#define sk_lingertime lingertime
#define sk_rcvbuf rcvbuf
#define sk_priority priority
#define sk_pair pair
#define sk_bound_dev_if bound_dev_if

#define work_struct tq_struct
#define INIT_WORK INIT_TQUEUE
#define schedule_work schedule_task
#define __WORK_INITIALIZER(n, f, d) { \
	.list = LIST_HEAD_INIT((n).list), \
	.routine = (f), \
	.data = (d)     \
}
#define DECLARE_WORK(n, f, d) struct tq_struct n = __WORK_INITIALIZER(n, f, d)

#include <net/sock.h>

enum sock_flags {
        SOCK_DEAD,
        SOCK_DONE,
        SOCK_URGINLINE,
        SOCK_KEEPOPEN,
        SOCK_LINGER,
        SOCK_DESTROY,
        SOCK_BROADCAST,
#ifdef CONFIG_CHELSIO_T1_CM
        SOCK_MIGRATE,
#endif
};

extern void unimplemented_sock_flag(void);

static inline void sock_reset_flag(struct sock *sk, int flag)
{
	switch (flag) {
		case SOCK_DEAD: sk->dead = 0; break;
		case SOCK_DONE: sk->done = 0; break;
		case SOCK_URGINLINE: sk->urginline = 0; break;
		case SOCK_KEEPOPEN: sk->keepopen = 0; break;
		case SOCK_LINGER: sk->linger = 0; break;
#ifdef CONFIG_CHELSIO_T1_CM
                /* borrow no_check */
		case SOCK_MIGRATE: sk->no_check = 0; break;
#endif
		default: unimplemented_sock_flag();
	}
}

static inline void sock_set_flag(struct sock *sk, int flag)
{
	switch (flag) {
		case SOCK_DEAD: sk->dead = 1; break;
		case SOCK_DONE: sk->done = 1; break;
		case SOCK_URGINLINE: sk->urginline = 1; break;
		case SOCK_KEEPOPEN: sk->keepopen = 1; break;
		case SOCK_LINGER: sk->linger = 1; break;
#ifdef CONFIG_CHELSIO_T1_CM
                /* borrow no_check */
		case SOCK_MIGRATE: sk->no_check = 1; break;
#endif
		default: unimplemented_sock_flag();
	}
}

static inline int sock_flag(struct sock *sk, int flag)
{
	switch (flag) {
		case SOCK_DEAD: return sk->dead;
		case SOCK_DONE: return sk->done;
		case SOCK_URGINLINE: return sk->urginline;
		case SOCK_KEEPOPEN: return sk->keepopen;
		case SOCK_LINGER: return sk->linger;
#ifdef CONFIG_CHELSIO_T1_CM
                /* borrow no_check */
		case SOCK_MIGRATE: return sk->no_check;
#endif
	}
	unimplemented_sock_flag();
	return 0;
}

static inline void inet_reset_saddr(struct sock *sk)
{
	sk->rcv_saddr = sk->saddr = 0;
}

#define tcp_v4_setup_caps(sk, dst) (sk)->route_caps = (dst)->dev->features
#define ip_fib_main_table main_table

/*
 * This accomodates RHEL which is a cross between 2.4 and 2.6 with additional
 * features.
 */
#ifndef AIO_MAXSEGS
#define ip_route_output_flow(rp, fl, sk, flags) ip_route_output_key(rp, fl)

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,4,23)
#define PDE(inode) ((struct proc_dir_entry *) (inode)->u.generic_ip)
#endif
#endif

#define is_bound(sk) ((sk)->prev != NULL)

#ifndef __user
# define __user
#endif

#else

#if LINUX_VERSION_CODE > KERNEL_VERSION(2,6,10)
# define tcp_opt tcp_sock
# define inet_opt inet_sock
#endif

#define inet_addr_info(sk) inet_sk(sk)
#define is_bound(sk) (tcp_sk(sk)->bind_hash != NULL)
#endif

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,7)
# define sk_stream_wait_connect(sk, timeo) wait_for_tcp_connect(sk, 0, timeo)
# define sk_acceptq_is_full tcp_acceptq_is_full
# define sk_acceptq_removed tcp_acceptq_removed
# define sk_acceptq_added   tcp_acceptq_added
# define sk_stream_write_space tcp_write_space
# define sk_stream_min_wspace tcp_min_write_space

# define recv_wscale rcv_wscale
# define send_wscale snd_wscale
# define winscale_ok wscale_ok

# define WRITE_PENDING(sk, tp) ((tp)->write_pending)

# define TCP_MIB_ATTEMPTFAILS        TcpAttemptFails
# define TCP_MIB_ACTIVEOPENS         TcpActiveOpens
# define LINUX_MIB_LISTENOVERFLOWS   ListenOverflows
# define LINUX_MIB_LISTENDROPS       ListenDrops 
# define LINUX_MIB_TCPABORTONSYN     TCPAbortOnSyn
# define LINUX_MIB_TCPABORTONTIMEOUT TCPAbortOnTimeout
# define LINUX_MIB_TCPABORTONDATA    TCPAbortOnData
# define LINUX_MIB_TCPABORTONCLOSE   TCPAbortOnClose
# define IPSTATS_MIB_OUTNOROUTES     IpOutNoRoutes

extern long tcp_data_wait(struct sock *sk, long timeo);

static inline int sk_wait_data(struct sock *sk, long *timeo)
{
	*timeo = tcp_data_wait(sk, *timeo);
	return 0;
}
#elif LINUX_VERSION_CODE < KERNEL_VERSION(2,6,10)
/* for linux-2.6.9-11.EL RHEL 4 update 1 */
# define recv_wscale rcv_wscale
# define send_wscale snd_wscale
# define winscale_ok wscale_ok
# define sk_pair     sk_security

# define WRITE_PENDING(sk, tp) ((sk)->sk_write_pending)
# define packets_out packets_out.val
#else
# define eff_sacks   rx_opt.eff_sacks
# define dsack       rx_opt.dsack
# define recv_wscale rx_opt.rcv_wscale
# define send_wscale rx_opt.snd_wscale
# define winscale_ok rx_opt.wscale_ok
# define sk_pair     sk_security

# define WRITE_PENDING(sk, tp) ((sk)->sk_write_pending)
#endif

#endif
