/*****************************************************************************
 *                                                                           *
 * File:                                                                     *
 *  toe_ctl_defs.h                                                           *
 *                                                                           *
 * Description:                                                              *
 *  Chelsio 10Gb Ethernet TOE driver                                         *
 *                                                                           *
 * This program is free software; you can redistribute it and/or modify      *
 * it under the terms of the GNU General Public License, version 2, as       *
 * published by the Free Software Foundation.                                *
 *                                                                           *
 * You should have received a copy of the GNU General Public License along   *
 * with this program; if not, write to the Free Software Foundation, Inc.,   *
 * 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.                 *
 *                                                                           *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR IMPLIED    *
 * WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF      *
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.                     *
 *                                                                           *
 * http://www.chelsio.com                                                    *
 *                                                                           *
 * Copyright (c) 2003 - 2005 Chelsio Communications, Inc.                    *
 * All rights reserved.                                                      *
 *                                                                           *
 * Maintainers: maintainers@chelsio.com                                      *
 *                                                                           *
 * Authors: Dimitrios Michailidis   <dm@chelsio.com>                         *
 *          Karen Xie               <kxie@chelsio.com>                       *
 *                                                                           *
 * History:                                                                  *
 *                                                                           *
 ****************************************************************************/
/* $Date: 2005/09/27 07:38:46 $ $RCSfile: toe_ctl_defs.h,v $ $Revision: 1.11 $ */

#ifndef _CHELSIO_TOE_CTL_DEFS_H
#define _CHELSIO_TOE_CTL_DEFS_H

enum {
    TOE_GET_TX_PAGE_INFO,
    TOE_GET_RX_PAGE_INFO,
    TOE_GET_TX_MAX_CHUNK,
    TOE_GET_TID_RANGE,
    TOE_GET_STID_RANGE,
    TOE_GET_L2T_CAPACITY,
    TOE_GET_RTBL_RANGE,
    TOE_GET_BEST_MTU_IDX,
    TOE_GET_MAX_PORT,
    TOE_GET_DEV_BY_HWADDR
};

/*
 * Structure used to describe a TID range.  Valid TIDs are [base, base+num).
 */
struct toe_tid_range {
    unsigned int base;   /* First TID */
    unsigned int num;    /* Number of TIDs in range */
};

/*
 * TX/RX page information.
 */
struct toe_page_info {
    unsigned int page_size;  /* Page size, should be a power of 2 */
    unsigned int num;        /* Number of pages */
    unsigned int pm_base; 
};

/*
 * Structure used to request the MTU table index that best matches
 * a desired MTU.
 */
struct toe_mtu_idx {
    unsigned short mtu;         /* Desired MTU */
    unsigned short mtu_tab_idx; /* Index of closest entry in the MTU table */
};

/*
 * Structure used to request the net_device given a mac address
 */
struct toe_dev_byhwaddr {
    unsigned char *mac;	/* mac address */
    void *dev; 			/* net_device */
};

int best_mtu_index(struct toedev *tdev, unsigned short mtu);
#endif

// Local Variables:
// c-basic-offset: 4
// End:
