/*
 * Copyright (C) 2003 - 2004 Chelsio Communications.  All rights reserved.
 *
 * Written by Karen Xie (kxie@chelsio.com)
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
 * more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59
 * Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */
                                                                                
#include <linux/config.h>
                                                                                
#ifdef CONFIG_CHELSIO_T1_BONDING
                                                                                
#define __CHELSIO_BOND_C__
                                                                                
#include <linux/module.h>
#include <linux/vmalloc.h>
#include <linux/ip.h>
#include <linux/init.h>
#include <linux/netdevice.h>
#include <linux/toedev.h>
#include <linux/spinlock.h>
#include <linux/tcp.h>

#include "../toe_compat.h"
#include "../cpl5_cmd.h"
#include "../toe_ctl_defs.h"
#include "defs.h"
#include "tom.h"
#include "l2t.h"
#include "tcam_rte.h"
#include "cpl5_defs.h"
                                                                                
#include "bond.h"
#include "../../../bonding/bonding.h"
#include "../../../bonding/bond_3ad.h"
                                                                                
#ifdef CONFIG_CHELSIO_T1_CM
#include "../cml/cml_osdep.h"
#endif

#include <linux/proc_fs.h>
#include <linux/seq_file.h>
#include <linux/if_bonding.h>


/****************************************************************/
/*			debug print				*/
/****************************************************************/
//#define OFFLOAD_BOND_DEBUG
//#define STATIC
#define STATIC 	static
                                                                                
#ifdef OFFLOAD_BOND_DEBUG
#define TB_PRINT_INFO(X)  do { printk(KERN_INFO),printk X; } while (0)
/* more detailed debug msg */
#define TB_PRINT_INFO_1(X) do { printk(KERN_INFO),printk X; } while (0)
#else
#define TB_PRINT_INFO(X)
#define TB_PRINT_INFO_1(X)
#endif /* OFFLOAD_BOND_DEBUG */


#define TB_PRINT_WARN(X)  do { printk(KERN_WARNING),printk X; } while (0)
#define TB_PRINT_ERR(X)  do { printk(KERN_ERR),printk X; } while (0)
                                                                                
#define TB_PRINT_SK(str, sk) \
        do { \
        printk(KERN_INFO \
               "%s: sk 0x%08x.%u -- 0x%08x.%u, state 0x%x.\n",  \
               str, inet_addr_info(sk)->rcv_saddr, inet_addr_info(sk)->num, \
               inet_addr_info(sk)->daddr, inet_addr_info(sk)->dport, \
               sk->sk_state); \
        } while (0)

/****************************************************************/
/* 	additional info. saved in net_device 			*/
/****************************************************************/
#define MAX_TOEDEV	1

struct tdev_info {
    struct toedev	*tdev;  
    unsigned int	ref_cnt;
};

struct tb_info {
    unsigned int	flags;
#define TB_SLAVE_LINK_UP	0x1
#define TB_BOND_NEED_FO_RESCAN	0x10
    /* the followings are used only on the virtual dev: bond_dev */
    int			mode;
    unsigned int        tdev_cnt;
    struct tdev_info	tdev_info[MAX_TOEDEV];
    int (*hash)(unsigned int key, struct net_device *bond_dev);
    struct net_device* (*select)(unsigned int key,
                        int (*hash)(unsigned int, struct net_device *),
                        struct net_device *dev);
};

#define TB_INFO(netdev) (*(struct tb_info **)&((netdev)->atalk_ptr) )

/****************************************************************/
/* 			TOEDEV API				*/
/****************************************************************/
/* the offload decision will be made at connect time */
STATIC int t1_bond_can_offload(struct toedev *tdev, struct sock *sk)
{
    return ((tdev && tdev->connect) ? 1 : 0);
}

/* return 0 if offload successful, -1 otherwise */
STATIC int t1_bond_connect(struct toedev *tdev, struct sock *sk)
{
    struct net_device 	*bond_dev = __sk_dst_get(sk)->dev;
    struct tb_info	*mtbp = TB_INFO(bond_dev);
    struct net_device 	*slave_dev;
    struct toedev	*slave_tdev;
    struct l2t_entry 	*e;

    if (!mtbp) {
        TB_PRINT_INFO_1(("chtb: %s, %s error.\n", __FUNCTION__, bond_dev->name));
        return -1;
    }

    slave_dev = mtbp->select(inet_addr_info(sk)->daddr, mtbp->hash, bond_dev);

    /* no suitable slave found */
    if (!slave_dev) {
        TB_PRINT_INFO(("chtb: %s connect, no suitable slave found.\n",
                   bond_dev->name));
        return -1;
    }

    slave_tdev = TOEDEV(slave_dev);
    if (!slave_tdev || !slave_tdev->can_offload(slave_tdev, sk))  {
        TB_PRINT_INFO(("chtb: %s connect, slave %s tdev 0x%p cannot offload.\n",
                   bond_dev->name, slave_dev->name, slave_tdev));
        return -1;
    }

    /* set up the L2T entry */
    e = l2t_get(slave_tdev, __sk_dst_get(sk)->neighbour);
    if (!e) {
        TB_PRINT_INFO(("chtb: %s connect, slave %s l2t_get failed.\n",
                   bond_dev->name, slave_dev->name));
        return -1;
    }

    spin_lock_bh(&e->lock);
    if (e->real_dev == bond_dev) {
        e->real_dev = slave_dev;
    } else if (slave_dev != e->real_dev) {
        slave_dev = e->real_dev;
        slave_tdev = TOEDEV(slave_dev);
    }
    spin_unlock_bh(&e->lock);

    TB_PRINT_INFO(("chtb: %s connect, select slave %s.\n",
                   bond_dev->name, slave_dev->name));

    return (slave_tdev->connect(slave_tdev, sk));
}

STATIC void t1_bond_neigh_update(struct toedev *tdev, struct neighbour *neigh, int fl, struct net_device *bond_dev)
{
    struct tb_info	*mtbp = TB_INFO(bond_dev);
    struct net_device 	*slave_dev;
    struct toedev	*slave_tdev;

    if (!mtbp) return;

    slave_dev = mtbp->select(*(unsigned int *)neigh->primary_key, mtbp->hash, bond_dev);
    TB_PRINT_INFO_1(("chtb: %s neigh_update key 0x%x select slave %s.\n",
                   bond_dev->name, *(unsigned int *)neigh->primary_key, 
                   slave_dev ? slave_dev->name:"?"));
    if (!slave_dev) return;
    slave_tdev = TOEDEV(slave_dev);
    if (!slave_tdev || !slave_tdev->neigh_update) return;

    /* in 802.3AD, we are hashing through slaves, every update could */
    /* result a different slave, so check if the one being used is still up */
    /* if it is, ignore the update */
    if (mtbp->mode == BOND_MODE_8023AD) {
        struct tdev_info	*tp = &(mtbp->tdev_info[0]);
        struct net_device 	*new_dev;
        struct toedev	        *new_tdev;
        int			tdev_cnt = mtbp->tdev_cnt;
        int			i;

        /* go through the l2t table, see if failover is needed */
        for (i = 0; i < tdev_cnt; i++, tp++) {
            struct l2t_entry    *e;
            e = l2t_get(tp->tdev, neigh);
            if (!e) continue;
            new_dev = slave_dev;
            new_tdev = slave_tdev;
            if (e->real_dev != bond_dev) {
    	        struct tb_info	*stbp = TB_INFO(e->real_dev);
                if (stbp && (stbp->flags & TB_SLAVE_LINK_UP)) {
                    new_dev = e->real_dev;           
                    new_tdev = TOEDEV(new_dev);
                }
            }
            TB_PRINT_INFO_1(("chtb: %s neigh_update, key 0x%x, using slave %s.\n",
                   bond_dev->name, *(unsigned int *)neigh->primary_key, 
                   new_dev->name));
             new_tdev->neigh_update(new_tdev, neigh, fl, new_dev);
        }
    } else {
        TB_PRINT_INFO_1(("chtb: %s neigh_update, key 0x%x, using slave %s.\n",
                   bond_dev->name, *(unsigned int *)neigh->primary_key, 
                   slave_dev->name));
        slave_tdev->neigh_update(slave_tdev, neigh, fl, slave_dev);
    } 
}

/****************************************************************/
/* 			Bonding API				*/
/****************************************************************/
#ifdef CONFIG_PROC_FS
void t1_bond_info_show_offload(struct seq_file *seq)
{
    struct bonding 	*bond = seq->private;
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,4,26)
    struct net_device	*bond_dev = bond->device;
#else
    struct net_device	*bond_dev = bond->dev;
#endif
    struct tb_info	*tbp = TB_INFO(bond_dev);
    struct tdev_info	*tp;
    struct slave	*slave;
    int 		i, max;

    if (!tbp) return;

    max = tbp->tdev_cnt;
    seq_printf(seq, "\noffload: flags 0x%x, mode %d\n\ttdevs(%u): ",
               tbp->flags, tbp->mode, tbp->tdev_cnt);
    for (i = 0, tp = &(tbp->tdev_info[0]); i < max; i++, tp++) {
        seq_printf(seq, "%s(%u), ", tp->tdev->name, tp->ref_cnt);
    }
    seq_printf(seq, "\n\tslaves:");
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,4,26)
    for (slave = bond->prev; slave != (struct slave *)bond; slave = slave->prev)
#else
    bond_for_each_slave(bond, slave, i)
#endif
    {
        tbp = TB_INFO(slave->dev);
        if (tbp) {
            seq_printf(seq, "%s(0x%x), ", slave->dev->name, tbp->flags);
        }
    }
    seq_printf(seq, "\n");
}
#endif

/****************************************************************/
/*			Hash Function				*/
/****************************************************************/
static int tb_hash_3ad(unsigned int key, struct net_device *bond_dev)
{
    return ( (key >> 24) ^ bond_dev->dev_addr[5] );
}

/****************************************************************/
/*		Failover -- port switching			*/
/****************************************************************/
/* abort all tcp connections associated with this bonding device */
STATIC int tb_abort_conn_on_tdev(struct net_device *bond_dev, struct toedev *tdev)
{
    struct tom_data 	*d = TOM_DATA(tdev);
    struct tid_info 	*t = &d->tid_maps;
    struct sock 	*sk;
    int			i, cnt = 0;

    /* go through from's tid table and release any conn. on from */
    for (i = 0; i < t->ntids; i++) {
        if ( (sk = t->tid_tab[i]) ) {
            struct tcp_opt *tp = tcp_sk(sk);
            struct l2t_entry *e = L2T_ENTRY(tp);
            if ( (e->dev == bond_dev) && (sk->sk_state != TCP_CLOSE) ) {
                struct sk_buff *skb;
                skb = alloc_skb(sizeof(struct cpl_abort_req), GFP_ATOMIC);
                if (!skb) {
                    TB_PRINT_INFO(("chtb: abort sk, temp out of memory.\n"));
                    continue;
                }
                TB_PRINT_SK("chtb: abort", sk);
                sk->sk_err = ECONNRESET;
                t1_send_reset(sk, CPL_ABORT_SEND_RST, skb);
                cnt++;
            }
        }
    }
    return cnt;
}

/* send l2t cpl to change port */
STATIC int t1_bond_change_l2t(struct net_device *dev, struct l2t_entry *e)
{
    struct toedev               *tdev = TOEDEV(dev);
    struct sk_buff		*skb;
    struct cpl_l2t_write_req 	*req;
    int 			idx = l2e_idx(L2DATA(tdev), e);

    if (e->dev == dev) return 0; 

    skb = alloc_skb(sizeof(*req), GFP_ATOMIC);
    if (!skb) {
        TB_PRINT_INFO(("chtb: change l2t to %s, out of memory.\n", dev->name));
        return -ENOMEM;
    }

    req = (struct cpl_l2t_write_req *)__skb_put(skb, sizeof(*req));
    OPCODE_TID(req) = htonl(MK_OPCODE_TID(CPL_L2T_WRITE_REQ, idx));
    req->params = htonl(V_L2TOP_IDX(idx) | V_L2TOP_VLAN(e->vlan) |
                        V_L2TOP_IFF(dev->if_port));
    memcpy(req->dst_mac, e->neigh->ha, sizeof(req->dst_mac));
    memcpy(e->dmac, req->dst_mac, sizeof(e->dmac));
    skb->priority = CPL_PRIORITY_CONTROL;
    skb->real_dev = dev;
    toe_send(tdev, skb);
    TB_PRINT_INFO(("chtb: send L2T_WRITE_REQ %s/%s, entry %d.\n",
                   e->dev->name, dev->name, idx));

    l2t_update(dev, tdev, e->neigh, 0);
                                                                                
    return 0;
}

/* invalidate l2t entries associated with the device */
STATIC int t1_bond_invalidate_slave_l2t(struct net_device *dev)
{
    struct net_device	*bond_dev = dev->master;
    struct toedev       *tdev = TOEDEV(dev);
    struct l2t_data	*d = L2DATA(tdev);
    struct l2t_entry	*e, *rover;
    int              	i, max, cnt = 0;

    TB_PRINT_INFO_1(("chtb: %s invalidate slave %s, %s l2t.\n", 
                   bond_dev->name, dev->name, tdev->name));

    write_lock_bh(&d->lock);
    e = &d->l2tab[1];
    max = d->nentries;
    rover = d->rover;
    write_unlock_bh(&d->lock);

    for (i = 1; i < max && e != rover; e++, i++) {
        if (e->dev && e->dev == bond_dev &&
            e->real_dev && e->real_dev == dev) {
            spin_lock_bh(&e->lock);
            e->real_dev = bond_dev;
            spin_unlock_bh(&e->lock);
            cnt++;
        }
    }
                                                                                
    return cnt;
}

/* delete all l2t entries associated with the bonding device */
STATIC int tb_delete_l2t_on_tdev(struct net_device *bond_dev, struct toedev *tdev)
{
    struct l2t_data	*d = L2DATA(tdev);
    struct l2t_entry	*e, *rover;
    int 		i, max, cnt = 0;

    TB_PRINT_INFO_1(("chtb: %s, delete l2t on tdev %s.\n", 
                   bond_dev->name, tdev->name));

    write_lock_bh(&d->lock);
    e = &d->l2tab[1];
    max = d->nentries;
    rover = d->rover;
    write_unlock_bh(&d->lock);

    for (i = 1; i < max && e != rover; e++, i++) {
        if (e->dev && e->dev == bond_dev) {
            TB_PRINT_INFO_1(("chtb: %s del l2t idx %d.\n",
                           tdev->name, i));
            spin_lock_bh(&e->lock);
            e->dev = e->real_dev = NULL;
            spin_unlock_bh(&e->lock);
            l2e_free(d, e);
            cnt++;
        }
    }
    return cnt;
}

STATIC void tb_remove_slave_tdev(struct toedev *, struct net_device *, int);
STATIC int t1_bond_failover_scan_l2t(struct net_device *bond_dev)
{
    struct tb_info      *mtbp = TB_INFO(bond_dev);
    struct tdev_info	*tp;
    int			tdev_cnt;
    int			i;

    if (!mtbp) {
        TB_PRINT_INFO_1(("chtb: %s, %s mtbp NULL.\n", __FUNCTION__, bond_dev->name));
        return -EINVAL;
    }
    tp = &(mtbp->tdev_info[0]);
    tdev_cnt = mtbp->tdev_cnt;

    local_bh_disable();

    for (i = 0; i < tdev_cnt; i++, tp++) {
        struct tom_data         *d = TOM_DATA(tp->tdev);
        struct tid_info         *t = &d->tid_maps;
        struct l2t_data         *l = L2DATA(tp->tdev);
        struct l2t_entry        *e, *rover;
        unsigned int            j, max;

        /* go through tid tables, halt tx if there is outstanding data */
        max = t->ntids;
        for (j = 0; j < max; j++) {
            struct sock         *sk = t->tid_tab[j];
            struct tcp_opt      *tp;
            if (!sk) continue;
            tp = tcp_sk(sk);
            if (!tp) continue;
            e = L2T_ENTRY(tp);
            if (e && e->dev && e->dev == bond_dev) {
                struct net_device   *old_dev = e->real_dev;
                struct tb_info      *stbp = TB_INFO(old_dev);

                /* wait for tx to drain, if there are un-acked data */
                if ( (old_dev == bond_dev ||
                     (stbp && ((stbp->flags & TB_SLAVE_LINK_UP) == 0)) ) && 
                     (tp->snd_una != tp->snd_nxt) ) {
                      TB_PRINT_INFO(("chtb: sk 0x%p, tid %u, 0x%x != 0x%x, need to flush TX.\n", sk, j, tp->snd_una, tp->snd_nxt));
	            CLOSE_FLAGS(tp) |= TX_HALT;
                }
            }
        }

        /* go through the l2t table, see if failover is needed */
        write_lock_bh(&l->lock);
        e = &l->l2tab[1];
        max = l->nentries;
        rover = l->rover;
        write_unlock_bh(&l->lock);

        for (j = 1; j < max && e != rover; e++, j++) {
            if (e->dev && e->dev == bond_dev) {
                struct net_device   *old_dev = e->real_dev;
                struct net_device   *new_dev;
                struct tb_info      *stbp = TB_INFO(old_dev);

                if (old_dev == bond_dev || 
                    (stbp && ((stbp->flags & TB_SLAVE_LINK_UP) == 0))) {
                    new_dev = mtbp->select(e->addr, mtbp->hash, bond_dev);
                    if (!new_dev) { goto err; }
                    /* on same toe, port switching */
                    TB_PRINT_INFO(("chtb: %s, tdev %s, idx %d, %s -> %s.\n",
                                    bond_dev->name, tp->tdev->name, j, 
                                    old_dev->name, new_dev->name));

                    t1_bond_change_l2t(new_dev, e);
                }
            }
        }

        /* this slave is removed, delete tdev */
        if (tp->ref_cnt == 0)
            tb_remove_slave_tdev(tp->tdev, bond_dev, 1);
    }
    local_bh_enable();

    mtbp->flags &= ~TB_BOND_NEED_FO_RESCAN;
    return 0;
err:
    local_bh_enable();
    mtbp->flags |= TB_BOND_NEED_FO_RESCAN;
    TB_PRINT_INFO_1(("chtb: %s, %s error.\n", __FUNCTION__, bond_dev->name));
    return -1;
}

/****************************************************************/
/* add/remove toedev to bond_dev:
 * 	toedev is allocated for bond_dev when the 1st 
 * 	offloadable slave is added; 
 *	and is deleted, when last slave is released		*/	
/****************************************************************/
STATIC void tb_insert_slave_tdev(struct toedev *tdev, struct net_device *bond_dev)
{
    struct tb_info	*tbp = TB_INFO(bond_dev);
    struct tdev_info	*tp;
    int			max;
    int 		i;

    if (!tbp) return;
    tp = &(tbp->tdev_info[0]);
    max = tbp->tdev_cnt;

    for (i = 0; i < max; i++, tp++) {
        if (tp->tdev == tdev) {
            tp->ref_cnt++;
            TB_PRINT_INFO_1(("chtb: %s, tdev %s, incr ref_cnt %u.\n", 
                           bond_dev->name, tdev->name, tp->ref_cnt));
            return;
        }
    }

    if (i >= MAX_TOEDEV) {
        TB_PRINT_ERR(("chtb ERR: %s, tdev max reached %u.\n", 
                       bond_dev->name, i));
        return;
    }

    tp = &(tbp->tdev_info[max]);
    tp->tdev = tdev;
    tp->ref_cnt = 1;
    (tbp->tdev_cnt)++;
    TB_PRINT_INFO_1(("chtb: %s, add tdev %s, total_tdev %u.\n", 
                   bond_dev->name, tdev->name, tbp->tdev_cnt));
}

STATIC void tb_remove_slave_tdev(struct toedev *tdev, struct net_device *bond_dev, int del)
{
    struct tb_info	*tbp = TB_INFO(bond_dev);
    struct tdev_info	*tp;
    int			max;
    int 		i;

    if (!tbp) return;
    tp = &(tbp->tdev_info[0]);
    max = tbp->tdev_cnt;

    for (i = 0; i < max; i++, tp++) {
        if (tp->tdev == tdev) {
            if (tp->ref_cnt) tp->ref_cnt--;
            TB_PRINT_INFO_1(("chtb: %s, tdev %s, decr ref_cnt %u.\n", 
                           bond_dev->name, tdev->name, tp->ref_cnt));
            break;
        }
    }

    if (i == max) {
        TB_PRINT_ERR(("chtb ERR: %s remove tdev %s, not found!\n",
                        bond_dev->name, tdev->name));
        return;
    }

    if (del && tp->ref_cnt == 0)  {
        TB_PRINT_INFO_1(("chtb: %s, remove tdev %s, total_tdev %u.\n", 
                   bond_dev->name, tdev->name, tbp->tdev_cnt));
        /* abort all connections */
        tb_abort_conn_on_tdev(bond_dev, tdev);
        /* delete all l2t entries associated with this bond */
        tb_delete_l2t_on_tdev(bond_dev, tdev);

        i++;
        if (i < max) {
            struct tdev_info *tp_next = tp + 1;
            for (; i < max; i++, tp++) {
                memcpy(tp, tp_next, sizeof(struct tdev_info));
            }
        }
        tbp->tdev_cnt--;
    }
}

STATIC void tb_alloc_toedev(struct net_device *bond_dev) 
{
    struct toedev	*tdev;

    /* set up toedev */
    tdev = alloc_toedev();
    if (!tdev) {
        TB_PRINT_ERR(("chtb ERR: bond %s, alloc toedev failed.\n", 
                      bond_dev->name));
            return;
    }
    TOEDEV(bond_dev) = tdev;
    tdev->can_offload = t1_bond_can_offload;

    TB_PRINT_INFO_1(("chtb: %s alloc tdev 0x%p.\n",
                   bond_dev->name, tdev));
}

STATIC void tb_setup_toedev(struct net_device *bond_dev) 
{
    struct toedev	*tdev = TOEDEV(bond_dev);

    /* set up toedev */
    tdev->connect = t1_bond_connect;
    tdev->neigh_update = t1_bond_neigh_update;

    bond_dev->features |= NETIF_F_TCPIP_OFFLOAD;

    TB_PRINT_INFO_1(("chtb: %s setup tdev 0x%p.\n",
                   bond_dev->name, tdev));
}

STATIC void tb_reset_toedev(struct net_device *bond_dev) 
{
    struct toedev	*tdev = TOEDEV(bond_dev);

    bond_dev->features &= ~NETIF_F_TCPIP_OFFLOAD;
    tdev->connect = NULL;
    tdev->neigh_update = NULL;
    TB_PRINT_INFO_1(("chtb: %s reset tdev 0x%p.\n",
                   bond_dev->name, tdev));
}

STATIC void tb_delete_toedev(struct net_device *bond_dev) 
{
    struct toedev	*tdev = TOEDEV(bond_dev);
    struct tb_info	*tbp = TB_INFO(bond_dev);
    struct tdev_info	*tp;
    int			max;
    int			i;

    if (!tbp) return;
    tp = &(tbp->tdev_info[0]);
    max = tbp->tdev_cnt;

    for (i = 0; i < max; i++, tp++) {
        tb_remove_slave_tdev(tp->tdev, bond_dev, 1);
    }

    TOEDEV(bond_dev) = NULL;
    bond_dev->features &= ~NETIF_F_TCPIP_OFFLOAD;

    if (tdev) {
        TB_PRINT_INFO_1(("chtb: %s delete toedev 0x%p.\n",
                   bond_dev->name, tdev));
        kfree(tdev);
    }
}

/****************************************************************/
/*			slave add/remove			*/
/****************************************************************/
static inline void tb_alloc_tb_info(struct net_device *dev)
{
    struct tb_info *tbp;
                                                                                
    tbp = kmalloc(sizeof(struct tb_info), GFP_ATOMIC);
    if (tbp) {
        TB_PRINT_INFO_1(("chtb: %s alloc tbp 0x%p.\n", dev->name, tbp));
        memset(tbp, 0, sizeof(struct tb_info));
    } else {
        TB_PRINT_ERR(("chtb ERR: %s alloc tbp failed.\n", dev->name));
    }
    TB_INFO(dev) = tbp;
}
                                                                                
static inline void tb_del_tb_info(struct net_device *dev)
{
    struct tb_info	*tbp = TB_INFO(dev);

    TB_PRINT_INFO_1(("chtb: %s delete tbp 0x%p.\n", dev->name, tbp));
    TB_INFO(dev) = NULL;
    kfree(tbp);
}

/* temp check to restrict all slaves on the same TOE card until CM is ready */
/* return 1 if same toe, 0 otherwise */
STATIC int tb_slave_same_toe(struct net_device *bond_dev, struct net_device *slave_dev)
{
    struct tb_info	*tbp = TB_INFO(bond_dev);
    struct toedev	*tdev = TOEDEV(slave_dev);
    struct tdev_info	*tp;
    int			max;
    int 		i;

    if (!tdev) {
        TB_PRINT_INFO_1(("chtb: slave %s no toe.\n", slave_dev->name));
        return 1;
    }
    if (!tbp) {
        TB_PRINT_INFO_1(("chtb: %s, %s no TB.\n", __FUNCTION__, bond_dev->name));
        return 1;
    }
    tp = &(tbp->tdev_info[0]);
    max = tbp->tdev_cnt;
    if (max == 0) {
        TB_PRINT_INFO_1(("chtb: slave %s on 1st toe %s.\n", 
                       slave_dev->name, tdev->name));
        return 1;
    }
    
    for (i = 0; i < max; i++, tp++) {
        if (tp->tdev == tdev) {
            TB_PRINT_INFO_1(("chtb: slave %s on the same toe %s.\n", 
                       slave_dev->name, tdev->name));
            return 1;
        }
    }

    TB_PRINT_INFO_1(("chtb: slave %s on different toe %s.\n", 
                   slave_dev->name, tdev->name));
    return 0;
}

int t1_bond_new_slave_check(struct net_device *bond_dev, struct net_device *slave_dev)
{
    struct bonding      *bond = (struct bonding *)bond_dev->priv;

    if (!bond) {
        printk(KERN_ERR "%s: %s, bond NULL.\n", __FUNCTION__, bond_dev->name); 
        return 0;
    }

    /* we don't want mix of offload/non-offload slaves! */
    if (bond->slave_cnt == 0) {
        bond_dev->features |= slave_dev->features & NETIF_F_TCPIP_OFFLOAD;
    } else {
        if ( (bond_dev->features ^ slave_dev->features) & 
             NETIF_F_TCPIP_OFFLOAD) {
            TB_PRINT_ERR(("chtb ERR: %s slave %s mixing TOE/NON_TOE.\n",
                      bond_dev->name, slave_dev->name));
            return -EOPNOTSUPP; 
        }
        /* limit all slaves on the same TOE until CM is ready */
        if ( (bond_dev->features & NETIF_F_TCPIP_OFFLOAD) &&
             !(tb_slave_same_toe(bond_dev, slave_dev)) ) {
            TB_PRINT_ERR(("chtb ERR: %s, slave %s not on the same TOE.\n",
                      bond_dev->name, slave_dev->name));
            return -EOPNOTSUPP; 
        }
    }

    return 0;
}

STATIC void t1_bond_new_slave(struct slave *slave)
{
    struct net_device   *slave_dev = slave->dev;
    struct net_device   *bond_dev = slave_dev->master;
    struct bonding	*bond = (struct bonding *)bond_dev->priv;
    struct tb_info	*tbp;

    if (TOEDEV(slave_dev) == NULL) return;

    tb_alloc_tb_info(slave_dev);
    tbp = TB_INFO(slave_dev);
    if (!tbp) return;

    if ((bond->slave_cnt == 0)) {
        /* set up toedev */
        tb_setup_toedev(bond_dev);
    }

    tb_insert_slave_tdev(TOEDEV(slave_dev), bond_dev);

    /* Take the max of the slaves' header lengths. With Chelsio
     * slaves, we need to add room for the CPL messages to the TOE
     * (The slave devices have already done this).
     */
    if (bond_dev->hard_header_len < slave_dev->hard_header_len) {
        bond_dev->hard_header_len = slave_dev->hard_header_len;
    }
}

STATIC void t1_bond_del_slave(struct slave *slave, int flag)
{
    struct net_device   *slave_dev = slave->dev;
    struct net_device   *bond_dev = slave_dev->master;
    struct bonding	*bond = (struct bonding *)bond_dev->priv;
    int			rc;

    /* invalidate L2T entries */
    rc = t1_bond_invalidate_slave_l2t(slave_dev);

    /* remove slave's tdev */
    tb_remove_slave_tdev(TOEDEV(slave_dev), bond_dev, flag);

    /* delete offload bond specific info */
    tb_del_tb_info(slave_dev);

    if (bond->slave_cnt == 0) {
        /* reset offload flag & toedev */
        tb_reset_toedev(bond_dev);
    }
}

/****************************************************************/
/*			bond add/remove				*/
/****************************************************************/

/* bond is being added */
STATIC void t1_bond_new_bond(struct bond_init_data *init_data)
{
    struct net_device   *bond_dev = init_data->dev;
    struct toedev	*tdev;
    struct tb_info	*tbp;

    tb_alloc_toedev(bond_dev);
    tdev = TOEDEV(bond_dev);
    if (!tdev) return;

    tb_alloc_tb_info(bond_dev);
    tbp = TB_INFO(bond_dev);
    if (!tbp) return;

    tbp->mode = init_data->mode;
    tbp->select = init_data->select;
    if (init_data->mode == BOND_MODE_8023AD) {
        tbp->hash = tb_hash_3ad;
    }
}

STATIC void t1_bond_empty_bond(struct net_device *bond_dev)
{
    struct bonding      *bond = (struct bonding *)bond_dev->priv;
    struct slave        *slave;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,26)
    int                 i;
#endif

    if (!bond) {
        TB_PRINT_WARN(("chtb: master %s bond NULL.\n", bond_dev->name));
        return;
    }

    /* delete all slaves */
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,4,26)
    for (slave = bond->prev; slave != (struct slave *)bond; slave = slave->prev)
#else
    bond_for_each_slave(bond, slave, i)
#endif
    {
        t1_bond_del_slave(slave, 1);
    }

    tb_reset_toedev(bond_dev);
}

/* bond is being removed */
STATIC void t1_bond_del_bond(struct net_device *bond_dev)
{
    struct bonding	*bond = (struct bonding *)bond_dev->priv;
    if ( bond->slave_cnt) {
        t1_bond_empty_bond(bond_dev);
    }
    tb_delete_toedev(bond_dev);
    tb_del_tb_info(bond_dev);
}

/****************************************************************/
/*                                                              */
/*              	bonding notification			*/
/*                                                              */
/****************************************************************/

int t1_bond_notify(unsigned long event, void *data)
{
    TB_PRINT_INFO_1(("chtb: rcv bonding notification event %lu, \"%s\"\n",
                  event, bond_notify_str[event]));
                                                                                
    if (!data) {
        TB_PRINT_ERR(("chtb ERR: bonding event %lu, \"%s\", data NULL.\n",
                      event, bond_notify_str[event]));
        return -EINVAL;
    }
                                                                                
    switch (event) {
    case OFFLOAD_BOND_INIT:
        {
            struct bond_init_data *init_data = (struct bond_init_data *)data;
            TOEDEV(init_data->dev) = NULL;
            switch(init_data->mode) {
                case BOND_MODE_ACTIVEBACKUP:
                case BOND_MODE_8023AD:
                    t1_bond_new_bond(init_data);
                    break;
                default:
                    TB_PRINT_INFO_1(("chtb: mode %d unsupported.\n",
                                     init_data->mode));
                    break;
            }
        }
        break;
    case OFFLOAD_BOND_DONE:
        {
            struct net_device	*bond_dev = (struct net_device *)data;
            t1_bond_del_bond(bond_dev);
        }
        break;
    case OFFLOAD_BOND_RELEASE_ALL:
        {
            struct net_device	*bond_dev = (struct net_device *)data;
            if (bond_dev->features & NETIF_F_TCPIP_OFFLOAD)
                t1_bond_empty_bond((struct net_device *)data);
            else
                TB_PRINT_INFO_1(("chtb: bond %s no toe.\n", bond_dev->name));
        }
        break;
    case OFFLOAD_BOND_ENSLAVE:
        {
            struct slave        *slave = (struct slave *)data;
            struct net_device   *bond_dev = slave->dev->master;
            struct bonding	*bond = (struct bonding *)bond_dev->priv;
            if ( bond->slave_cnt == 0 || 
                 (bond_dev->features & NETIF_F_TCPIP_OFFLOAD) ) 
                t1_bond_new_slave(slave);
            else
                TB_PRINT_INFO_1(("chtb: bond %s no toe or slave cnt %d.\n",
                                  bond_dev->name, bond->slave_cnt));
        }
        break;
    case OFFLOAD_BOND_RELEASE:
        {
            struct slave        *slave = (struct slave *)data;
            struct tb_info	*stbp = TB_INFO(slave->dev);
            struct net_device   *bond_dev = slave->dev->master;
            struct bonding	*bond = (struct bonding *)bond_dev->priv;

            if (bond_dev->features & NETIF_F_TCPIP_OFFLOAD) {
                /* bond is empty, no need to do any failover */
                if (bond->slave_cnt == 0) {
                    t1_bond_empty_bond(bond_dev);
                } else {
                /* bond is not empty, do failover */
                    stbp->flags &= ~TB_SLAVE_LINK_UP;
                    /* failover unavailable, keep the slave tdev */
                    if (t1_bond_failover_scan_l2t(bond_dev) != 0) {
                        t1_bond_del_slave(slave, 0);
                    } else {
                        t1_bond_del_slave(slave, 1);
                    }
                }
            } else
                TB_PRINT_INFO_1(("chtb: bond %s no toe.\n", bond_dev->name));
        }
        break;
    case OFFLOAD_BOND_SLAVE_LINKUP:
        {
            struct slave        *slave = (struct slave *)data;
            struct tb_info	*tbp = TB_INFO(slave->dev);
            if (tbp) {
                tbp->flags |= TB_SLAVE_LINK_UP;
            } else
                TB_PRINT_INFO_1(("chtb: slave %s tbp NULL.\n", slave->dev->name));
        }
        break;
    case OFFLOAD_BOND_SLAVE_LINKDOWN:
        {
            struct slave	*slave = (struct slave *)data; 
            struct tb_info	*tbp = TB_INFO(slave->dev);
            if (tbp) {
                struct tb_info	*mtbp = TB_INFO(slave->dev->master);
                if (mtbp) {
                    tbp->flags &= ~TB_SLAVE_LINK_UP;
                    mtbp->flags |= TB_BOND_NEED_FO_RESCAN;
                } else
                    TB_PRINT_INFO_1(("chtb: master %s tbp NULL.\n",
                                     slave->dev->master->name));
            } else
                TB_PRINT_INFO_1(("chtb: slave %s tbp NULL.\n",
                                  slave->dev->name));
        }
        break;
    case OFFLOAD_BOND_SLAVE_LINKCHNG:
    case OFFLOAD_BOND_ACTIVE_LAG:
        {
            struct net_device	*bond_dev = (struct net_device *)data;
            struct bonding	*bond = (struct bonding *)bond_dev->priv;
            struct tb_info	*tbp = TB_INFO(bond_dev);

            if ( (bond_dev->features & NETIF_F_TCPIP_OFFLOAD) &&
                 bond->slave_cnt && tbp &&
                 (tbp->flags & TB_BOND_NEED_FO_RESCAN) ) {
                t1_bond_failover_scan_l2t(bond_dev);
            } else
                TB_PRINT_INFO_1(("chtb: %s feature 0x%lx, slave %d, mtbp 0x%p, flags 0x%x.\n",
                                  bond_dev->name, bond_dev->features,
                                  bond->slave_cnt, tbp, tbp->flags));
        }
        break;
    }
    return 0; 
}

EXPORT_SYMBOL(t1_bond_notify);
EXPORT_SYMBOL(t1_bond_new_slave_check);
#ifdef CONFIG_PROC_FS
EXPORT_SYMBOL(t1_bond_info_show_offload);
#endif
#endif
