/*
 * Copyright (C) 2003 - 2004 Chelsio Communications.  All rights reserved.
 *
 * Written by Karen Xie (kxie@chelsio.com)
 *
 * This program is free software; you can redistribute it and/or modify it 
 * under the terms of the GNU General Public License as published by the Free 
 * Software Foundation; either version 2 of the License, or (at your option) 
 * any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or 
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for 
 * more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 
 * Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#ifndef _CHELSIO_BOND_H
#define _CHELSIO_BOND_H

#include <linux/config.h>
                                                                                
#ifdef CONFIG_CHELSIO_T1_BONDING

#include <linux/if_bonding.h>
#include <linux/netdevice.h>
#include <linux/toedev.h>

/* interested in slave info: enslave/release, link up/down */
#define t1_bond_need_slave_info(mode,dev) \
	( (((mode) == BOND_MODE_ACTIVEBACKUP) || \
	  ((mode) == BOND_MODE_8023AD)) && \
	( (dev)->features & NETIF_F_TCPIP_OFFLOAD) )

/* interested in bond info: add/remove */
#define t1_bond_need_bond_info(mode,master) \
	(((mode) == BOND_MODE_ACTIVEBACKUP) || 	\
	 ((mode) == BOND_MODE_8023AD))

/* bonding driver notification events */
enum {
   OFFLOAD_BOND_INIT,
   OFFLOAD_BOND_DONE,
   OFFLOAD_BOND_ENSLAVE,
   OFFLOAD_BOND_RELEASE,
   OFFLOAD_BOND_RELEASE_ALL,
   OFFLOAD_BOND_SLAVE_LINKUP,
   OFFLOAD_BOND_SLAVE_LINKDOWN,
   OFFLOAD_BOND_SLAVE_LINKCHNG,
   OFFLOAD_BOND_ACTIVE_LAG,
   OFFLOAD_BOND_END
};

struct bond_init_data {
    int 		mode;
    struct net_device	*dev;
    struct net_device* (*select)(unsigned int key, 
			int (*hash)(unsigned int, struct net_device *),
			struct net_device *dev);
};

#ifdef __CHELSIO_BOND_C__
static char bond_notify_str[OFFLOAD_BOND_END][10] = {
	"init",
	"done",
	"enslave",
	"release",
	"rel all",
	"linkup",
	"linkdown",
	"linkchng",
	"act LAG"
};
#else
extern struct notifier_block t1_bond_notifier;

extern int t1_bond_register_notifier(struct notifier_block *nb);
extern int t1_bond_unregister_notifier(struct notifier_block *nb);
extern int t1_bond_notify(unsigned long event, void *data);
extern int t1_bond_new_slave_check(struct net_device *, struct net_device *);
#ifdef CONFIG_PROC_FS
void t1_bond_info_show_offload(struct seq_file *seq);
#endif

#endif

#endif

#endif

// Local Variables:
// c-basic-offset: 4
// End:
