/*
 * Copyright (C) 2003 - 2004 Chelsio Communications.  All rights reserved.
 *
 * Written by Divy Le Ray
 *
 * This program is free software; you can redistribute it and/or modify it 
 * under the terms of the GNU General Public License as published by the Free 
 * Software Foundation; either version 2 of the License, or (at your option) 
 * any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or 
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for 
 * more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 
 * Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#ifndef _CHELSIO_TCAM_RTE_H
#define _CHELSIO_TCAM_RTE_H

#include <linux/types.h>

/* info for a TCAM entry  */
struct tcam_rte_entry {
	int	rte_index;
	u8	rte_status;	/* Validity of the TCAM entry */
	u32	rte_faddr;	/* Stored foreign address */
	u32	rte_netmask;	/* Stored netmask address */
	int	rte_nm_len;	/* netmask length */
	u8	rte_l2t_ix;	/* Stored L2T index */
	u_int	rte_flags;

	u32	gw_opt_addr;
	u32	gw_opt_nm_len;
	struct list_head gw_dep;
/*
	int	gw_next;
	int	gw_prev;
*/
};

#define RTE_IN_TCAM 0x1		/* The route is in TCAM */
#define RTE_GW	    0X2		/* The entry is a gateway */
#define RTE_IND     0x4	 	/* The entry is indirect */

/* info for a L2T entry  */
struct term_l2t_entry {
	int 	l2t_index;
	u32	l2t_status;	/* Validity of the l2t entry */
	u8	l2t_dmac[6];	/* MAC address */
	u16	l2t_phy_port;	/* Physical port used as source */
	u16	l2t_vlan_tag;	/* VLAN tag */
};

/* PLO_OPT constants */
#define REG_FIRST 0
#define REG_GAP0 16
#define REG_LAST 32
#define REG_NUM (REG_LAST + 1)

/*
 * PLO_OPT algorithm regions.
 * In the high regions (netmask > 16) rg_lo points to the region's first entry,
 * rg_hi points to the next entry to be used for insertion.
 * In the low regions, rg_hi points to the region's last entry, rg_lo points
 * to the next entry to be used for insertion.
 */

struct tcam_opt_region {
	int	rg_lo;
	int	rg_hi;
};

struct tcam_rte_unit {
	u32			tcam_rte_base;	/* TCAM rte base address */
	int			rte_n_entries;	/* TCAM rte number of entries */
	int			l2t_n_entries;	/* TCAM L2T number of entries */
	atomic_t		tcam_rte_cnt;
	struct tcam_opt_region	opt_reg[REG_NUM]; /* PLO_OPT regions */
	struct tcam_rte_entry	*cache_rte;	/* Host cache of the TCAM rte */
	struct semaphore	rte_mutex;

#ifdef CONFIG_PROC_FS
	struct semaphore	read_rte_mutex;
	wait_queue_head_t	cpl_rte_wait;	/* Dumping rep/rpl sync */
	atomic_t		pending_reqs;	/* dumping reqs in process */
	struct tcam_rte_entry 	*rte_entry;	/* TCAM entries */
	struct term_l2t_entry	*l2t_entry;	/* L2T entries */
#endif
};


/* Getting the route info from the device */
#define TOM_RTE_DATA(dev)	((dev)->l3opt)
#define TCAM_NM32(len) (((len) == 32) ? (unsigned)(-1)				\
				      : (((1 << (len)) - 1) << (32 - len)))

/* TCAM routing notifier events */
enum {
	TCAM_RTE_INSERT,
	TCAM_RTE_DELETE
};

/* Base for routing related CPL messages'atids. */
#define RTE_BASE_ATID 10000

/* Control for TCAM insertion */
enum {
	TCAM_RTE_NEW = 0,
	TCAM_RTE_SHIFT = 1,
	TCAM_RTE_UPDATE = 2
};

#define TCAM_RTE_CTRL_INSERT 0x3
#define TCAM_RTE_CTRL_UPDATE 0x2

struct toe_tid_range;
struct toedev;
struct notifier_block;

extern struct tcam_rte_unit *init_tcam_rte(struct toe_tid_range *t_info,
					   int l2t_capacity);
extern void free_tcam_rte(struct tcam_rte_unit *t);
extern int tcam_rte_insert(struct toedev *dev, struct tcam_rte_entry *r);
extern int tcam_rte_delete(struct toedev *dev, struct tcam_rte_entry *r);
extern int tcam_rte_start(struct toedev *dev);
extern int register_tcam_rte_notifier(struct notifier_block *nb);
extern int unregister_tcam_rte_notifier(struct notifier_block *nb);

#ifdef CONFIG_PROC_FS
struct proc_dir_entry;

extern int tcam_rte_proc_setup(struct proc_dir_entry *dir, struct toedev *dev);
extern void tcam_rte_proc_free(struct proc_dir_entry *dir);


#else
#define tcam_rte_proc_setup(dir, d) 0
#define tcam_rte_proc_free(dir)
#endif


#endif /* _CHELSIO_TCAM_RTE_H */

// Local Variables:
// c-basic-offset: 8
// End:

