/*
 * Copyright (C) 2003 - 2004 Chelsio Communications.  All rights reserved.
 *
 * Written by Dimitris Michailidis (dm@chelsio.com)
 *
 * This program is free software; you can redistribute it and/or modify it 
 * under the terms of the GNU General Public License as published by the Free 
 * Software Foundation; either version 2 of the License, or (at your option) 
 * any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or 
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for 
 * more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 
 * Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#ifndef _CHELSIO_TOM_SYSCTL_H
#define _CHELSIO_TOM_SYSCTL_H

/*
 * TOM tunable parameters.  They can be manipulated through sysctl(2) or /proc.
 */
struct tom_tunables {
    int max_host_sndbuf;      // max host RAM consumed by a sndbuf
    int tx_hold_thres;        // push/pull threshold for non-full TX sk_buffs
    int max_tx_pages;         // max # of TX pages per connection
    int rx_credit_thres;      // min # of RX credits needed for RX_DATA_ACK
    int min_newconn_tx_pages; // min # of pages needed to admit new connection
    int toe_automode_on;      // Glabal switch activating automode for services
    int mss;                  // TX_DATA payload MSS
    int delack;               // delayed ACK control
    int max_conn;             // maximum number of offloaded connections
    int soft_backlog_limit;   // whether the listen backlog limit is soft
    unsigned long tx_lease_time;  // lease time on TX pages
    int tx_wscale;            // inflated TX window scale
};

struct toedev;
struct tom_sysctl_table;
struct tom_data;

extern void t1_init_tunables(struct tom_data *t, struct tom_tunables *p);
extern void t1_sysctl_unregister(struct tom_sysctl_table *t);
extern struct tom_sysctl_table *t1_sysctl_register(struct toedev *dev,
						 const struct tom_tunables *p);
#endif

// Local Variables:
// c-basic-offset: 4
// End:
