/*****************************************************************************
 *                                                                           *
 * File:                                                                     *
 *  trace.h                                                                  *
 *                                                                           *
 * Description:                                                              *
 *  Chelsio definitions and inline functions for trace buffers.              *
 *                                                                           *
 * This program is free software; you can redistribute it and/or modify      *
 * it under the terms of the GNU General Public License, version 2, as       *
 * published by the Free Software Foundation.                                *
 *                                                                           *
 * You should have received a copy of the GNU General Public License along   *
 * with this program; if not, write to the Free Software Foundation, Inc.,   *
 * 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.                 *
 *                                                                           *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR IMPLIED    *
 * WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF      *
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.                     *
 *                                                                           *
 * http://www.chelsio.com                                                    *
 *                                                                           *
 * Copyright (c) 2003 - 2005 Chelsio Communications, Inc.                    *
 * All rights reserved.                                                      *
 *                                                                           *
 * Maintainers: maintainers@chelsio.com                                      *
 *                                                                           *
 * Authors: Dimitrios Michailidis   <dm@chelsio.com>                         *
 *          Tina Yang               <tainay@chelsio.com>                     *
 *          Felix Marti             <felix@chelsio.com>                      *
 *                                                                           *
 * History:                                                                  *
 *                                                                           *
 ****************************************************************************/
/* $Date: 2005/07/09 02:47:14 $ $RCSfile: trace.h,v $ $Revision: 1.11 $ */

#ifndef __T1_TRACE_H__
#define __T1_TRACE_H__

#include <linux/config.h>

#if defined(CONFIG_PROC_FS) && defined(CONFIG_X86_TSC)
# define T1_TRACE
#endif

#ifdef T1_TRACE

#include <linux/time.h>
#include <linux/timex.h>

#define T1_TRACE_NUM_PARAM       6
#define T1_TRACE_MAX_TB_LEN      128
#define T1_TRACE_MAX_TB_NAME_LEN 32

#define T1_TRACE0(b, s) \
        if ((b) != NULL) \
                (void *) t1_trace ((b), (s));
#define	T1_TRACE1(b, s, p0) \
        if ((b) != NULL) { \
                unsigned long *_p; \
                _p = t1_trace ((b), (s)); \
                _p[0] = (t1_tracearg_t) (p0); \
        }
#define T1_TRACE2(b, s, p0, p1) \
        if ((b) != NULL) { \
                unsigned long *_p; \
                _p = t1_trace ((b), (s)); \
                _p[0] = (t1_tracearg_t) (p0); \
                _p[1] = (t1_tracearg_t) (p1); \
        }
#define T1_TRACE3(b, s, p0, p1, p2) \
        if ((b) != NULL) { \
                unsigned long *_p; \
                _p = t1_trace ((b), (s)); \
                _p[0] = (t1_tracearg_t) (p0); \
                _p[1] = (t1_tracearg_t) (p1); \
                _p[2] = (t1_tracearg_t) (p2); \
        }
#define T1_TRACE4(b, s, p0, p1, p2, p3) \
        if ((b) != NULL) { \
                unsigned long *_p; \
                _p = t1_trace ((b), (s)); \
                _p[0] = (t1_tracearg_t) (p0); \
                _p[1] = (t1_tracearg_t) (p1); \
                _p[2] = (t1_tracearg_t) (p2); \
                _p[3] = (t1_tracearg_t) (p3); \
        }
#define T1_TRACE5(b, s, p0, p1, p2, p3, p4) \
        if ((b) != NULL) { \
                unsigned long *_p; \
                _p = t1_trace ((b), (s)); \
                _p[0] = (t1_tracearg_t) (p0); \
                _p[1] = (t1_tracearg_t) (p1); \
                _p[2] = (t1_tracearg_t) (p2); \
                _p[3] = (t1_tracearg_t) (p3); \
                _p[4] = (t1_tracearg_t) (p4); \
        }
#define T1_TRACE6(b, s, p0, p1, p2, p3, p4, p5) \
        if ((b) != NULL) { \
                unsigned long *_p; \
                _p = t1_trace ((b), (s)); \
                _p[0] = (t1_tracearg_t) (p0); \
                _p[1] = (t1_tracearg_t) (p1); \
                _p[2] = (t1_tracearg_t) (p2); \
                _p[3] = (t1_tracearg_t) (p3); \
                _p[4] = (t1_tracearg_t) (p4); \
                _p[5] = (t1_tracearg_t) (p5); \
        }


typedef unsigned long t1_tracearg_t;

struct t1_traceb_e {
        cycles_t      tsc;
        char *        fmt;
        t1_tracearg_t param[T1_TRACE_NUM_PARAM];
};

struct t1_traceb {
        struct t1_traceb_e   ep[T1_TRACE_MAX_TB_LEN];
        unsigned int    nume;
        unsigned int    count;
        char            name[T1_TRACE_MAX_TB_NAME_LEN];
};

static __inline__ unsigned long *t1_trace (struct t1_traceb *tb, char *fmt)
{
        struct t1_traceb_e *ep = &tb->ep[tb->count++ & (tb->nume - 1)];

        ep->fmt = fmt;
        ep->tsc = get_cycles();

        return (unsigned long *) &ep->param[0];
}

struct t1_traceb *t1_trace_alloc (char *);
void   t1_trace_free (struct t1_traceb *);
#endif

#endif /* __T1_TRACE_H__ */
