//
// ddp.h
//
// Direct Data Placement (DDP) API for Chelsio TOE
//
#ifndef __CHTOE_DDP_H__
#define __CHTOE_DDP_H__

#include "ulp.h"

#define PAGE_POD_PAGE_SIZE_4K           0
#define PAGE_POD_PAGE_SIZE_8K           1
#define PAGE_POD_PAGE_SIZE_16K          2
#define PAGE_POD_PAGE_SIZE_16M_ULP1     3
#define PAGE_POD_PAGE_SIZE_16M_ULP0     12

/* opaque structure maintaining ddp map table state */
#define MAX_MEMORY_REGIONS      256

typedef struct chtoe_ddp_map_table chtoe_ddp_map_table;
struct chtoe_ddp_map_table
{
    spinlock_t dmt_lock;
    __u8  pp_size;
    __u8  pp_shift;
    __u8  tag_id_bits;
    __u8  tag_color_bits;
    __u32 dmt_tag_alias_bits; // # ddp tag alias bits
    __u32 dmt_tag_id_mask;
    __u32 dmt_tag_color_mask;
    __u32 dmt_tag_mask;     // ddp tag ID+COLOR mask
    __u32 dmt_off_map_tag;
    __u32 dmt_map_tag_invalid;
    __u32 dmt_size;         // # elements in ddp table
    __u32 dmt_memregion[MAX_MEMORY_REGIONS];// array of memory regions
    __u8 *dmt_color;        // last color for each ddp map table entry
    __u8 *dmt_table;        // ddp map table
    int (*setup_pagepod)(struct toedev *tdev, chtoe_ddp_map_table *dmt,
        int sgcnt, struct scatterlist *sg, __u32 offset, __u32 tid,
        int maplen, __u8 psize, __u32 *tag);
};

void* t1_init_ddp_table(struct ulp_info *ulp);
void t1_free_ddp(void *dmt);

#endif // __CHTOE_DDP_H__
