/* $Date: 2005/03/08 05:17:37 $ $RCSfile: xpak.c,v $ $Revision: 1.31 $ */
#include "cphy.h"
#include "elmer0.h"

FILE_IDENT("@(#) $Id: xpak.c,v 1.31 2005/03/08 05:17:37 sbardone Exp $");

static int xpak_reset(struct cphy *cphy, int wait)
{
	return 0;
}

static int xpak_interrupt_enable(struct cphy *cphy)
{
	return 0;
}

static int xpak_interrupt_disable(struct cphy *cphy)
{
	return 0;
}

static int xpak_interrupt_clear(struct cphy *cphy)
{
	return 0;
}

static int xpak_set_loopback(struct cphy *cphy, int on)
{
	return 0;
}

static int xpak_get_link_status(struct cphy *cphy, int *link_ok, int *speed,
				int *duplex, int *fc)
{
	if (link_ok)
		*link_ok = 1;
	if (speed)
		*speed = SPEED_10000;
	if (duplex)
		*duplex = DUPLEX_FULL;
	if (fc)
		*fc = PAUSE_RX | PAUSE_TX;
	return 0;
}

static void xpak_destroy(struct cphy *cphy)
{
	t1_os_free((void *)cphy, sizeof(*cphy));
}

#ifdef C99_NOT_SUPPORTED
static struct cphy_ops xpak_ops = {
	 xpak_destroy,
         xpak_reset,
         xpak_interrupt_enable,
         xpak_interrupt_disable,
         xpak_interrupt_clear,
         NULL,
         NULL,
         NULL,
         NULL,
         NULL,
	 xpak_set_loopback,
         NULL,
         xpak_get_link_status,
};
#else
static struct cphy_ops xpak_ops = {
	.destroy           = xpak_destroy,
	.reset             = xpak_reset,
	.interrupt_enable  = xpak_interrupt_enable,
	.interrupt_disable = xpak_interrupt_disable,
	.interrupt_clear   = xpak_interrupt_clear,
	.get_link_status   = xpak_get_link_status,
	.set_loopback      = xpak_set_loopback,
};
#endif

static struct cphy *xpak_phy_create(adapter_t * adapter, int phy_addr,
				    struct mdio_ops *mdio_ops)
{
	struct cphy *cphy = t1_os_malloc_wait_zero(sizeof(*cphy));

	if (!cphy)
		return NULL;

	cphy->ops        = &xpak_ops;
	cphy->adapter    = adapter;
	cphy->mdio_read  = mdio_ops->read;
	cphy->mdio_write = mdio_ops->write;
	return cphy;
}

static int xpak_phy_reset(adapter_t *adapter)
{
	u32 val;

	t1_tpi_read(adapter, A_ELMER0_GPO, &val);
	val &= ~4;
	t1_tpi_write(adapter, A_ELMER0_GPO, val);
	DELAY_MS(100);

	/*
	 * Errata #26 states to wait 5 seconds after reset before transceiver
	 * becomes operational.
	 */
	t1_tpi_write(adapter, A_ELMER0_GPO, val | 4);
	DELAY_MS(5000);

	/* Now lets enable the Laser. Delay 100us
	 * as defined in XPAK errata.
	 */
	t1_tpi_read(adapter, A_ELMER0_GPO, &val);
	val |= 0x8000;
	t1_tpi_write(adapter, A_ELMER0_GPO, val);
	DELAY_US(100);
	return 0;
}

struct gphy t1_xpak_ops = {
	xpak_phy_create,
	xpak_phy_reset
};
