/*
 * icom.h
 *
 * Copyright (C) 2001 Michael Anderson, IBM Corporation
 *
 * Serial device driver include file.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 */

#include<linux/serial_core.h>

#define BAUD_TABLE_LIMIT	((sizeof(icom_acfg_baud)/sizeof(int)) - 1)
static int icom_acfg_baud[] = {
	300,
	600,
	900,
	1200,
	1800,
	2400,
	3600,
	4800,
	7200,
	9600,
	14400,
	19200,
	28800,
	38400,
	57600,
	76800,
	115200,
	153600,
	230400,
	307200,
	460800,
};

struct icom_regs {
	u32 control;		/* Adapter Control Register     */
	u32 interrupt;		/* Adapter Interrupt Register   */
	u32 int_mask;		/* Adapter Interrupt Mask Reg   */
	u32 int_pri;		/* Adapter Interrupt Priority r */
	u32 int_reg_b;		/* Adapter non-masked Interrupt */
	u32 resvd01;
	u32 resvd02;
	u32 resvd03;
	u32 control_2;		/* Adapter Control Register 2   */
	u32 interrupt_2;	/* Adapter Interrupt Register 2 */
	u32 int_mask_2;		/* Adapter Interrupt Mask 2     */
	u32 int_pri_2;		/* Adapter Interrupt Prior 2    */
	u32 int_reg_2b;		/* Adapter non-masked 2         */
};

struct func_dram {
	u32 reserved[108];	/* 0-1B0   reserved by personality code */
	u32 RcvStatusAddr;	/* 1B0-1B3 Status Address for Next rcv */
	u8 RcvStnAddr;		/* 1B4     Receive Station Addr */
	u8 IdleState;		/* 1B5     Idle State */
	u8 IdleMonitor;		/* 1B6     Idle Monitor */
	u8 FlagFillIdleTimer;	/* 1B7     Flag Fill Idle Timer */
	u32 XmitStatusAddr;	/* 1B8-1BB Transmit Status Address */
	u8 StartXmitCmd;	/* 1BC     Start Xmit Command */
	u8 HDLCConfigReg;	/* 1BD     Reserved */
	u8 CauseCode;		/* 1BE     Cause code for fatal error */
	u8 xchar;		/* 1BF     High priority send */
	u32 reserved3;		/* 1C0-1C3 Reserved */
	u8 PrevCmdReg;		/* 1C4     Reserved */
	u8 CmdReg;		/* 1C5     Command Register */
	u8 async_config2;	/* 1C6     Async Config Byte 2 */
	u8 async_config3;	/* 1C7     Async Config Byte 3 */
	u8 dce_resvd[20];	/* 1C8-1DB DCE Rsvd           */
	u8 dce_resvd21;		/* 1DC     DCE Rsvd (21st byte */
	u8 misc_flags;		/* 1DD     misc flags         */
#define V2_HARDWARE     0x40
#define ICOM_HDW_ACTIVE 0x01
	u8 call_length;		/* 1DE     Phone #/CFI buff ln */
	u8 call_length2;	/* 1DF     Upper byte (unused) */
	u32 call_addr;		/* 1E0-1E3 Phn #/CFI buff addr */
	u16 timer_value;	/* 1E4-1E5 general timer value */
	u8 timer_command;	/* 1E6     general timer cmd  */
	u8 dce_command;		/* 1E7     dce command reg    */
	u8 dce_cmd_status;	/* 1E8     dce command stat   */
	u8 x21_r1_ioff;		/* 1E9     dce ready counter  */
	u8 x21_r0_ioff;		/* 1EA     dce not ready ctr  */
	u8 x21_ralt_ioff;	/* 1EB     dce CNR counter    */
	u8 x21_r1_ion;		/* 1EC     dce ready I on ctr */
	u8 rsvd_ier;		/* 1ED     Rsvd for IER (if ne */
	u8 ier;			/* 1EE     Interrupt Enable   */
	u8 isr;			/* 1EF     Input Signal Reg   */
	u8 osr;			/* 1F0     Output Signal Reg  */
	u8 reset;		/* 1F1     Reset/Reload Reg   */
	u8 disable;		/* 1F2     Disable Reg        */
	u8 sync;		/* 1F3     Sync Reg           */
	u8 error_stat;		/* 1F4     Error Status       */
	u8 cable_id;		/* 1F5     Cable ID           */
	u8 cs_length;		/* 1F6     CS Load Length     */
	u8 mac_length;		/* 1F7     Mac Load Length    */
	u32 cs_load_addr;	/* 1F8-1FB Call Load PCI Addr */
	u32 mac_load_addr;	/* 1FC-1FF Mac Load PCI Addr  */
};

/*
 * adapter defines and structures
 */
#define ICOM_CONTROL_START_A         0x00000008
#define ICOM_CONTROL_STOP_A          0x00000004
#define ICOM_CONTROL_START_B         0x00000002
#define ICOM_CONTROL_STOP_B          0x00000001
#define ICOM_CONTROL_START_C         0x00000008
#define ICOM_CONTROL_STOP_C          0x00000004
#define ICOM_CONTROL_START_D         0x00000002
#define ICOM_CONTROL_STOP_D          0x00000001
#define ICOM_IRAM_OFFSET             0x1000
#define ICOM_IRAM_SIZE               0x0C00
#define ICOM_DCE_IRAM_OFFSET         0x0A00
#define ICOM_CABLE_ID_VALID          0x01
#define ICOM_CABLE_ID_MASK           0xF0
#define ICOM_DISABLE                 0x80
#define CMD_XMIT_RCV_ENABLE          0xC0
#define CMD_XMIT_ENABLE              0x40
#define CMD_RCV_DISABLE              0x00
#define CMD_RCV_ENABLE               0x80
#define CMD_RESTART                  0x01
#define CMD_HOLD_XMIT                0x02
#define CMD_SND_BREAK                0x04
#define RS232_CABLE                  0x06
#define V24_CABLE                    0x0E
#define V35_CABLE                    0x0C
#define V36_CABLE                    0x02
#define NO_CABLE                     0x00
#define START_DOWNLOAD               0x80
#define ICOM_INT_MASK_PRC_A          0x00003FFF
#define ICOM_INT_MASK_PRC_B          0x3FFF0000
#define ICOM_INT_MASK_PRC_C          0x00003FFF
#define ICOM_INT_MASK_PRC_D          0x3FFF0000
#define INT_RCV_COMPLETED            0x1000
#define INT_XMIT_COMPLETED           0x2000
#define INT_IDLE_DETECT              0x0800
#define INT_RCV_DISABLED             0x0400
#define INT_XMIT_DISABLED            0x0200
#define INT_RCV_XMIT_SHUTDOWN        0x0100
#define INT_FATAL_ERROR              0x0080
#define INT_CABLE_PULL               0x0020
#define INT_SIGNAL_CHANGE            0x0010
#define HDLC_PPP_PURE_ASYNC          0x02
#define HDLC_FF_FILL                 0x00
#define HDLC_HDW_FLOW                0x01
#define START_XMIT                   0x80
#define ICOM_ACFG_DRIVE1             0x20
#define ICOM_ACFG_NO_PARITY          0x00
#define ICOM_ACFG_PARITY_ENAB        0x02
#define ICOM_ACFG_PARITY_ODD         0x01
#define ICOM_ACFG_8BPC               0x00
#define ICOM_ACFG_7BPC               0x04
#define ICOM_ACFG_6BPC               0x08
#define ICOM_ACFG_5BPC               0x0C
#define ICOM_ACFG_1STOP_BIT          0x00
#define ICOM_ACFG_2STOP_BIT          0x10
#define ICOM_DTR                     0x80
#define ICOM_RTS                     0x40
#define ICOM_RI                      0x08
#define ICOM_DSR                     0x80
#define ICOM_DCD                     0x20
#define ICOM_CTS                     0x40

#define NUM_XBUFFS 1
#define NUM_RBUFFS 2
#define RCV_BUFF_SZ 0x0200
#define XMIT_BUFF_SZ 0x1000
struct statusArea {
    /**********************************************/
	/* Transmit Status Area                       */
    /**********************************************/
	struct {
		u32 leNext;	/* Next entry in Little Endian on Adapter */
		u32 leNextASD;
		u32 leBuffer;	/* Buffer for entry in LE for Adapter */
		u16 leLengthASD;
		u16 leOffsetASD;
		u16 leLength;	/* Length of data in segment */
		u16 flags;
#define SA_FLAGS_DONE           0x0080	/* Done with Segment */
#define SA_FLAGS_CONTINUED      0x8000	/* More Segments */
#define SA_FLAGS_IDLE           0x4000	/* Mark IDLE after frm */
#define SA_FLAGS_READY_TO_XMIT  0x0800
#define SA_FLAGS_STAT_MASK      0x007F
	} xmit[NUM_XBUFFS];

    /**********************************************/
	/* Receive Status Area                        */
    /**********************************************/
	struct {
		u32 leNext;	/* Next entry in Little Endian on Adapter */
		u32 leNextASD;
		u32 leBuffer;	/* Buffer for entry in LE for Adapter */
		u16 WorkingLength;	/* size of segment */
		u16 reserv01;
		u16 leLength;	/* Length of data in segment */
		u16 flags;
#define SA_FL_RCV_DONE           0x0010	/* Data ready */
#define SA_FLAGS_OVERRUN         0x0040
#define SA_FLAGS_PARITY_ERROR    0x0080
#define SA_FLAGS_FRAME_ERROR     0x0001
#define SA_FLAGS_FRAME_TRUNC     0x0002
#define SA_FLAGS_BREAK_DET       0x0004	/* set conditionally by device driver, not hardware */
#define SA_FLAGS_RCV_MASK        0xFFE6
	} rcv[NUM_RBUFFS];
};

struct icom_adapter;


#define ICOM_MAJOR       243
#define ICOM_MINOR_START 0

struct icom_port {
	struct uart_port uart_port;
	u8 imbed_modem;
#define ICOM_UNKNOWN		1
#define ICOM_RVX		2
#define ICOM_IMBED_MODEM	3
	unsigned char cable_id;
	unsigned char read_status_mask;
	unsigned char ignore_status_mask;
	unsigned long int_reg;
	struct icom_regs *global_reg;
	struct func_dram *dram;
	int port;
	struct statusArea *statStg;
	dma_addr_t statStg_pci;
	u32 *xmitRestart;
	dma_addr_t xmitRestart_pci;
	unsigned char *xmit_buf;
	dma_addr_t xmit_buf_pci;
	unsigned char *recv_buf;
	dma_addr_t recv_buf_pci;
	int next_rcv;
	int put_length;
	int status;
#define ICOM_PORT_ACTIVE	1	/* Port exists. */
#define ICOM_PORT_OFF		0	/* Port does not exist. */
	int load_in_progress;
	struct icom_adapter *adapter;
};

struct icom_adapter {
	unsigned long base_addr;
	unsigned long base_addr_pci;
	unsigned char irq_number;
	struct pci_dev *pci_dev;
	struct icom_port port_info[4];
	int index;
	int version;
#define ADAPTER_V1	0x0001
#define ADAPTER_V2	0x0002
	u32 subsystem_id;
#define FOUR_PORT_MODEL				0x0252
#define V2_TWO_PORTS_RVX			0x021A
#define V2_ONE_PORT_RVX_ONE_PORT_IMBED_MDM	0x0251
	int numb_ports;
	struct list_head icom_adapter_entry;
	struct kobject kobj;
};

/* prototype */
extern void iCom_sercons_init(void);

static unsigned char callSetup[1968] = {
	0xBD, 0xD9, 0x23, 0x00, 0xDD, 0xDD, 0x18, 0x05, 0x23, 0x80, 0x3E,
	0x7F, 0x23, 0x00, 0x3E, 0x7C,
	0x23, 0x10, 0x3E, 0xB7, 0x3E, 0xB5, 0x23, 0x20, 0x3E, 0xB6, 0x3E,
	0xB4, 0xA2, 0x0A, 0x86, 0x0A,
	0xAE, 0x0A, 0xBE, 0x0A, 0x23, 0x80, 0x3E, 0x01, 0x23, 0x2A, 0x3E,
	0x02, 0x3D, 0xDA, 0x23, 0xFF,
	0x3E, 0x03, 0x22, 0x86, 0xD9, 0xDD, 0x10, 0x1D, 0x22, 0x08, 0x2F,
	0xF0, 0xD9, 0xF1, 0x18, 0x56,
	0xD9, 0xDD, 0x18, 0x33, 0x23, 0x00, 0x3E, 0x06, 0x3E, 0x46, 0x3E,
	0x86, 0x3E, 0xC6, 0x21, 0xF0,
	0x37, 0xFA, 0x2F, 0xFA, 0x3E, 0x08, 0x23, 0x00, 0x3E, 0x07, 0x3E,
	0x87, 0x3E, 0xC7, 0x23, 0xFF,
	0x3E, 0x09, 0x3E, 0x47, 0x00, 0x3A, 0x23, 0x00, 0x3E, 0x06, 0x3E,
	0x08, 0x23, 0x13, 0x3E, 0x07,
	0x23, 0x01, 0x3E, 0x09, 0xA2, 0xFC, 0xA2, 0xF6, 0xCD, 0xF0, 0x10,
	0x3F, 0x82, 0xFC, 0xD9, 0xDD,
	0x18, 0x4F, 0x23, 0xFF, 0x3E, 0x87, 0x68, 0x70, 0x23, 0x14, 0x68,
	0x62, 0x3F, 0x20, 0x23, 0x00,
	0x3E, 0x87, 0x68, 0x70, 0x23, 0x14, 0x68, 0x62, 0x3F, 0x20, 0x22,
	0x86, 0x00, 0x54, 0x68, 0x70,
	0x23, 0x14, 0x68, 0x62, 0x3F, 0x20, 0x22, 0x08, 0x2F, 0xF0, 0x3D,
	0xF5, 0x3B, 0xB0, 0x10, 0x6E,
	0x23, 0x00, 0x3E, 0x02, 0x23, 0x01, 0x3E, 0x7C, 0x23, 0x14, 0x3E,
	0x03, 0x23, 0x00, 0x3E, 0x7C,
	0x23, 0x80, 0xDD, 0xF2, 0x1B, 0xA3, 0x68, 0x69, 0x3B, 0x00, 0x10,
	0x61, 0x23, 0x09, 0x3E, 0x02,
	0x3D, 0xDA, 0x23, 0x08, 0xDD, 0xDD, 0x18, 0x6D, 0x23, 0x88, 0x3E,
	0x7F, 0x21, 0xDA, 0xC9, 0xF0,
	0x10, 0x72, 0x33, 0x04, 0x3E, 0x02, 0x3D, 0xDA, 0xD9, 0xF1, 0x10,
	0x79, 0x23, 0x00, 0x3D, 0xF1,
	0x00, 0xA1, 0x21, 0xF5, 0x2F, 0xF0, 0x3B, 0xE0, 0x10, 0x7F, 0x23,
	0x22, 0x00, 0x98, 0x3B, 0x60,
	0x10, 0x83, 0x23, 0x22, 0x00, 0x98, 0x3B, 0xC0, 0x10, 0x87, 0x23,
	0xEE, 0x00, 0x98, 0x3B, 0xB0,
	0x10, 0x8B, 0x23, 0x44, 0x00, 0x98, 0x3B, 0x20, 0x10, 0x8F, 0x23,
	0xC4, 0x00, 0x98, 0x3B, 0x30,
	0x10, 0x93, 0x23, 0x22, 0x00, 0x98, 0x3B, 0xF0, 0x1B, 0xA7, 0x00,
	0xBE, 0xD9, 0xDD, 0x18, 0x99,
	0x3E, 0x46, 0x23, 0x11, 0x68, 0x69, 0x3B, 0x00, 0x10, 0x9A, 0x22,
	0x06, 0x37, 0xFF, 0x3D, 0xEF,
	0x81, 0xF5, 0x0A, 0xE0, 0x08, 0xC4, 0x08, 0xE3, 0x0A, 0x5B, 0x0B,
	0x5D, 0xDD, 0xF3, 0x10, 0xA1,
	0xBD, 0xF3, 0x9D, 0xD9, 0x21, 0xDE, 0x3B, 0x00, 0x18, 0xB5, 0x5D,
	0x00, 0x4B, 0xC8, 0x41, 0xE0,
	0x22, 0x7F, 0x33, 0x40, 0x3E, 0x7F, 0x21, 0xDE, 0x4D, 0xC8, 0x58,
	0x00, 0x4B, 0xC8, 0x41, 0xFC,
	0x22, 0x7F, 0x2F, 0xBF, 0x3E, 0x7F, 0x21, 0xF7, 0x4D, 0xC8, 0x00,
	0xBD, 0x23, 0xDD, 0x3D, 0xF4,
	0x0B, 0xC6, 0x96, 0xF4, 0x9E, 0xF6, 0x00, 0xC3, 0xDD, 0xE6, 0x68,
	0x6E, 0xC5, 0xE6, 0x68, 0x6F,
	0xC1, 0xE6, 0x18, 0xD4, 0x81, 0xE6, 0x68, 0x70, 0x43, 0xC8, 0x45,
	0xCA, 0x21, 0xE4, 0x68, 0x62,
	0x21, 0xE5, 0x68, 0x74, 0x47, 0xD4, 0x68, 0x6D, 0x41, 0xD2, 0x43,
	0xC8, 0x50, 0x00, 0x47, 0xC8,
	0x41, 0xC8, 0x68, 0x71, 0x68, 0x6E, 0xC1, 0xE6, 0x68, 0x6E, 0x85,
	0xE6, 0x8A, 0xF4, 0xC1, 0xE6,
	0x68, 0x6F, 0xA5, 0xE6, 0x68, 0x6D, 0x21, 0xE7, 0x2F, 0x0F, 0x3D,
	0xCA, 0x2F, 0x0C, 0x3B, 0x0C,
	0x68, 0x6E, 0x21, 0xE7, 0xDE, 0xF0, 0x1A, 0x4F, 0xDA, 0xF0, 0x19,
	0x0C, 0xA2, 0x0A, 0x99, 0xE7,
	0x23, 0x00, 0x3D, 0xD6, 0x3D, 0xE8, 0x21, 0xCA, 0x3B, 0x0F, 0x10,
	0xFE, 0x23, 0x81, 0x3E, 0x01,
	0x21, 0xDA, 0x3E, 0x02, 0x23, 0x7E, 0x3E, 0x03, 0x92, 0xF1, 0x01,
	0x0A, 0x23, 0x40, 0x3E, 0x01,
	0x21, 0xC6, 0x3E, 0x02, 0x21, 0xC7, 0x3E, 0x03, 0x21, 0xCA, 0x3B,
	0x0C, 0x23, 0x00, 0x11, 0x09,
	0x23, 0x01, 0x3D, 0xD7, 0x82, 0x0A, 0x68, 0x6D, 0x21, 0xCA, 0x3B,
	0x0F, 0x19, 0x4A, 0x21, 0xE7,
	0xDE, 0xF0, 0x1A, 0x4F, 0xD6, 0xF0, 0x11, 0x30, 0x68, 0x70, 0x43,
	0xC8, 0x21, 0xCA, 0xC2, 0xF0,
	0x21, 0xC9, 0x68, 0x64, 0xDD, 0xD7, 0x11, 0x26, 0x3D, 0xC8, 0x21,
	0xCA, 0xC6, 0xF0, 0x21, 0xC8,
	0x68, 0x64, 0x3D, 0xC8, 0x21, 0xCA, 0xCA, 0xF0, 0x21, 0xC8, 0x68,
	0x64, 0x68, 0x6B, 0x29, 0xD8,
	0x3B, 0x7F, 0xDA, 0xF1, 0x21, 0xE7, 0x2F, 0x0F, 0x3D, 0xCA, 0x11,
	0x30, 0x91, 0xE7, 0x02, 0x14,
	0xDE, 0x04, 0x68, 0x6E, 0x68, 0x70, 0x43, 0xC8, 0x21, 0xCA, 0xC2,
	0xF0, 0x21, 0xC9, 0x68, 0x64,
	0x3D, 0xD8, 0x21, 0xCA, 0x3D, 0xCB, 0xDD, 0xD7, 0x11, 0x47, 0x21,
	0xCA, 0xC6, 0xF0, 0x21, 0xD8,
	0x68, 0x64, 0x3D, 0xD8, 0x21, 0xCA, 0xCA, 0xF0, 0x21, 0xD8, 0x68,
	0x64, 0x3D, 0xD8, 0x21, 0xE7,
	0x2F, 0x0F, 0x3D, 0xCA, 0xDE, 0x04, 0x68, 0x6E, 0x22, 0x00, 0xCA,
	0x0A, 0x11, 0x56, 0xA2, 0x0A,
	0x99, 0xE8, 0x21, 0xCA, 0x3B, 0x0F, 0x1A, 0x56, 0x82, 0x0A, 0x68,
	0x6D, 0x3D, 0xC8, 0x22, 0x04,
	0x3D, 0xC9, 0x21, 0xCA, 0x3B, 0x0F, 0x11, 0x66, 0x21, 0xC9, 0x2F,
	0x38, 0x3B, 0x00, 0x11, 0x63,
	0x21, 0xC8, 0x3E, 0xB0, 0x01, 0xFA, 0xD5, 0xE7, 0x68, 0x6E, 0x02,
	0x14, 0x21, 0xC9, 0xD6, 0xF0,
	0x11, 0x6A, 0x95, 0xE8, 0xCE, 0xF0, 0x11, 0x6D, 0x9D, 0xE8, 0x21,
	0xD7, 0x3B, 0x00, 0x19, 0xF9,
	0xDE, 0xF0, 0x11, 0x7A, 0xDA, 0xF0, 0x19, 0xBC, 0x21, 0xC8, 0x3B,
	0x54, 0x11, 0xFA, 0x23, 0xC2,
	0x3D, 0xD7, 0x01, 0xFA, 0x3B, 0x01, 0x11, 0x80, 0x21, 0xC8, 0x3B,
	0x52, 0x11, 0xF7, 0x01, 0xF3,
	0x3B, 0x02, 0x11, 0x86, 0x21, 0xC8, 0x3B, 0x4F, 0x11, 0xF7, 0x01,
	0xF3, 0x3B, 0x03, 0x11, 0x8C,
	0x21, 0xC8, 0x3B, 0x42, 0x11, 0xF7, 0x01, 0xF3, 0x3B, 0x04, 0x11,
	0x92, 0x21, 0xC8, 0x3B, 0x47,
	0x11, 0xF7, 0x01, 0xF3, 0x3B, 0x05, 0x11, 0x98, 0x21, 0xC8, 0x3B,
	0x20, 0x11, 0xF7, 0x01, 0xF3,
	0x3B, 0x06, 0x11, 0x9E, 0x21, 0xC8, 0x3B, 0x54, 0x11, 0xF7, 0x01,
	0xF3, 0x3B, 0x07, 0x11, 0xA4,
	0x21, 0xC8, 0x3B, 0x4F, 0x11, 0xF7, 0x01, 0xF3, 0x3B, 0x08, 0x11,
	0xAA, 0x21, 0xC8, 0x3B, 0x4E,
	0x11, 0xF7, 0x01, 0xF3, 0x3B, 0x09, 0x11, 0xF7, 0x21, 0xC8, 0x3B,
	0x59, 0x11, 0xF7, 0x23, 0x80,
	0x3D, 0xD7, 0x21, 0xCB, 0xC6, 0xF0, 0x21, 0xD8, 0x68, 0x64, 0x3D,
	0xD8, 0x21, 0xCB, 0xCA, 0xF0,
	0x21, 0xD8, 0x68, 0x64, 0x3D, 0xD8, 0x01, 0xFA, 0x2F, 0x3F, 0x3B,
	0x02, 0x11, 0xC3, 0x21, 0xC8,
	0x3B, 0x4F, 0x11, 0xF0, 0x01, 0xF3, 0x3B, 0x03, 0x11, 0xC9, 0x21,
	0xC8, 0x3B, 0x4E, 0x11, 0xF0,
	0x01, 0xF3, 0x3B, 0x04, 0x11, 0xCF, 0x21, 0xC8, 0x3B, 0x59, 0x11,
	0xF0, 0x01, 0xF3, 0x3B, 0x05,
	0x11, 0xD5, 0x21, 0xC8, 0x3B, 0x20, 0x11, 0xF0, 0x01, 0xF3, 0x3B,
	0x06, 0x11, 0xDB, 0x21, 0xC8,
	0x3B, 0x52, 0x11, 0xF0, 0x01, 0xF3, 0x3B, 0x07, 0x11, 0xE1, 0x21,
	0xC8, 0x3B, 0x4F, 0x11, 0xF0,
	0x01, 0xF3, 0x3B, 0x08, 0x11, 0xE7, 0x21, 0xC8, 0x3B, 0x42, 0x11,
	0xF0, 0x01, 0xF3, 0x3B, 0x09,
	0x11, 0xF0, 0x21, 0xC8, 0x3B, 0x47, 0x11, 0xF0, 0x21, 0xE8, 0x33,
	0x01, 0x3D, 0xE8, 0x01, 0xFA,
	0x23, 0x80, 0x3D, 0xD7, 0x01, 0x74, 0x21, 0xD7, 0x68, 0x68, 0x3D,
	0xD7, 0x01, 0xF9, 0x23, 0x00,
	0x3D, 0xD7, 0x95, 0xE7, 0xD1, 0xE8, 0x1A, 0x23, 0x45, 0xDE, 0xDE,
	0xF1, 0x12, 0x01, 0x91, 0xE8,
	0x02, 0x16, 0x51, 0xCC, 0x21, 0xD6, 0x68, 0x62, 0x21, 0xC8, 0x68,
	0x67, 0x45, 0xDE, 0x23, 0x01,
	0x68, 0x63, 0x47, 0xDE, 0x21, 0xD6, 0x68, 0x68, 0x3D, 0xD6, 0x21,
	0xE8, 0x2F, 0x07, 0x3B, 0x01,
	0x1A, 0x14, 0x21, 0xD6, 0x3B, 0x08, 0x12, 0x23, 0xD1, 0xE8, 0x1A,
	0x23, 0x21, 0xD6, 0x3B, 0x00,
	0x1A, 0x23, 0x41, 0xE0, 0x57, 0xCC, 0xDD, 0xF2, 0x1B, 0xA3, 0xC2,
	0xF1, 0x12, 0x1B, 0x68, 0x62,
	0x43, 0xE0, 0x23, 0x00, 0x3D, 0xD6, 0x21, 0xCA, 0x3B, 0x0F, 0x12,
	0x48, 0x21, 0xC9, 0x2F, 0x38,
	0x3B, 0x00, 0x12, 0x2C, 0x95, 0xE7, 0x68, 0x6D, 0x3B, 0x10, 0x12,
	0x43, 0x21, 0xC8, 0x3B, 0x7E,
	0x12, 0x40, 0x22, 0x74, 0x3B, 0x47, 0x12, 0x3D, 0x22, 0x75, 0x3B,
	0x0F, 0x12, 0x3D, 0x21, 0xE7,
	0xDE, 0xF0, 0x1A, 0x4F, 0xD6, 0xF0, 0x68, 0x6E, 0x02, 0x56, 0x23,
	0x04, 0x3D, 0xE8, 0x02, 0x56,
	0x23, 0x03, 0x3D, 0xE8, 0x02, 0x56, 0xCE, 0xF0, 0x68, 0x6F, 0x23,
	0x02, 0x3D, 0xE8, 0x02, 0x56,
	0x21, 0xE8, 0x2F, 0x07, 0x3B, 0x01, 0x1A, 0x56, 0xD1, 0xE7, 0x68,
	0x6E, 0x02, 0x56, 0xA2, 0x0A,
	0xCE, 0xF4, 0x68, 0x6F, 0xDD, 0xE7, 0x68, 0x6E, 0x23, 0x07, 0x3D,
	0xE8, 0xA2, 0x0A, 0x23, 0x00,
	0x3D, 0xE7, 0x8E, 0xF4, 0x68, 0x6D, 0x21, 0xE7, 0x2F, 0x0F, 0x3D,
	0xCA, 0x2F, 0x0C, 0x3D, 0xCB,
	0x3B, 0x08, 0x68, 0x6E, 0x21, 0xE7, 0xDE, 0xF0, 0x1A, 0xD4, 0xDA,
	0xF0, 0x1A, 0x80, 0x99, 0xE7,
	0x23, 0x00, 0x3D, 0xD6, 0x3D, 0xE8, 0x21, 0xCA, 0x3B, 0x0B, 0x12,
	0x77, 0x23, 0x81, 0x3E, 0x01,
	0x21, 0xDA, 0x3E, 0x02, 0x23, 0x7E, 0x3E, 0x03, 0x86, 0x0A, 0xAE,
	0x0A, 0x68, 0x6D, 0x23, 0x40,
	0x3E, 0x01, 0x21, 0xC6, 0x3E, 0x02, 0x21, 0xC7, 0x3E, 0x03, 0x86,
	0x0A, 0xAE, 0x0A, 0x68, 0x6D,
	0x21, 0xD6, 0x3B, 0x00, 0x12, 0xA4, 0x45, 0xDE, 0xDE, 0xF1, 0x12,
	0x8E, 0x21, 0xCA, 0x3B, 0x0B,
	0x12, 0xDC, 0xD1, 0xE7, 0x12, 0xC0, 0xC6, 0x04, 0x68, 0x6E, 0x02,
	0xDC, 0x21, 0xDF, 0x3B, 0x00,
	0x12, 0x95, 0x21, 0xDE, 0x3B, 0x08, 0xDA, 0xF1, 0x12, 0x96, 0x23,
	0x08, 0x41, 0xE0, 0x55, 0xCC,
	0xDD, 0xF2, 0x1B, 0xA3, 0xC2, 0xF1, 0x12, 0x98, 0x68, 0x62, 0x43,
	0xE0, 0x45, 0xDE, 0x68, 0x63,
	0x47, 0xDE, 0x3D, 0xD6, 0x23, 0x00, 0x3D, 0xD8, 0xC2, 0x04, 0x68,
	0x6E, 0xD5, 0xE7, 0x1A, 0xB1,
	0x21, 0xCA, 0x3B, 0x0B, 0x12, 0xB1, 0x23, 0x7E, 0x68, 0x6A, 0x3E,
	0x00, 0x95, 0xE7, 0x96, 0xF1,
	0xAE, 0x0A, 0x51, 0xCC, 0x21, 0xD8, 0x68, 0x62, 0x68, 0x68, 0x3D,
	0xD8, 0x68, 0x66, 0x68, 0x6B,
	0x3E, 0x00, 0x3E, 0xB1, 0xCE, 0x0A, 0x1A, 0xD0, 0x21, 0xD6, 0x68,
	0x69, 0x3D, 0xD6, 0x02, 0x80,
	0x22, 0x05, 0x2F, 0x0F, 0x3B, 0x0C, 0xDA, 0xF1, 0x68, 0x6F, 0x22,
	0x78, 0x68, 0x6B, 0x3E, 0x00,
	0x22, 0x79, 0x3E, 0x00, 0x23, 0x7E, 0x68, 0x6A, 0x3E, 0x00, 0x91,
	0xE7, 0xCE, 0x0A, 0x68, 0x6E,
	0x86, 0x0A, 0xAE, 0x0A, 0x99, 0xE8, 0x02, 0xDC, 0x86, 0x0A, 0xAE,
	0x0A, 0xCE, 0xF4, 0x68, 0x6F,
	0xDD, 0xE7, 0x68, 0x6E, 0x23, 0x07, 0x3D, 0xE8, 0x23, 0x00, 0x3D,
	0xE7, 0x8E, 0xF4, 0x68, 0x6D,
	0x22, 0x86, 0xD9, 0xDD, 0x12, 0xE4, 0x22, 0x08, 0x2F, 0xF0, 0x3B,
	0xB0, 0x1B, 0x13, 0x22, 0x86,
	0xD9, 0xDD, 0x12, 0xEB, 0x22, 0x08, 0x2F, 0xF0, 0x3B, 0xF0, 0x1B,
	0xA7, 0xD9, 0xDD, 0x1A, 0xF5,
	0x21, 0xF0, 0x37, 0xFA, 0x2F, 0xFA, 0x3E, 0x08, 0x02, 0xFE, 0xDD,
	0xF0, 0x68, 0x6C, 0xF2, 0x06,
	0xD9, 0xF0, 0x68, 0x6C, 0xE6, 0x06, 0xC5, 0xF0, 0x68, 0x6C, 0xE2,
	0x06, 0xDD, 0xF2, 0x1B, 0xA3,
	0xDD, 0xF1, 0x13, 0x05, 0xDD, 0xD9, 0x1B, 0xAD, 0xBD, 0xF1, 0x22,
	0x06, 0x37, 0xFF, 0x39, 0xEF,
	0x13, 0x0A, 0x68, 0x6D, 0x3D, 0xC8, 0x35, 0xEF, 0x2D, 0xEE, 0x3B,
	0x00, 0x21, 0xC8, 0x3D, 0xEF,
	0x68, 0x6F, 0x92, 0xF4, 0x68, 0x6D, 0x22, 0x86, 0xD9, 0xDD, 0x13,
	0x17, 0x22, 0x08, 0x2F, 0xF0,
	0x3B, 0xF0, 0x1B, 0xA7, 0xD9, 0xDD, 0x1B, 0x20, 0xDD, 0xF0, 0x68,
	0x6C, 0xFE, 0x08, 0x03, 0x23,
	0xDD, 0xF0, 0x68, 0x6C, 0xF2, 0x06, 0xDD, 0xE9, 0x13, 0x2C, 0xD9,
	0xE9, 0x1B, 0x2C, 0x22, 0xBC,
	0x3B, 0x10, 0x13, 0x2D, 0x99, 0xE9, 0x82, 0xF4, 0x3E, 0xBC, 0xDD,
	0xEA, 0x13, 0x36, 0xD9, 0xEA,
	0x1B, 0x36, 0x22, 0xBD, 0x3B, 0x10, 0x13, 0x37, 0x99, 0xEA, 0x82,
	0xF4, 0x3E, 0xBD, 0xDD, 0xEB,
	0x13, 0x40, 0xD9, 0xEB, 0x1B, 0x40, 0x22, 0xBE, 0x3B, 0x10, 0x13,
	0x41, 0x99, 0xEB, 0x82, 0xF4,
	0x3E, 0xBE, 0xDD, 0xEC, 0x13, 0x4A, 0xD9, 0xEC, 0x1B, 0x4A, 0x22,
	0xBF, 0x3B, 0x10, 0x13, 0x4B,
	0x99, 0xEC, 0x82, 0xF4, 0x3E, 0xBF, 0xDD, 0xF2, 0x1B, 0xA3, 0xDD,
	0xF1, 0x13, 0x52, 0xDD, 0xD9,
	0x1B, 0xAD, 0xBD, 0xF1, 0x22, 0x06, 0x37, 0xFF, 0x2F, 0x80, 0x39,
	0xEF, 0x13, 0x58, 0x68, 0x6D,
	0x3D, 0xEF, 0xDD, 0xEE, 0x68, 0x6E, 0x92, 0xF4, 0x68, 0x6D, 0x21,
	0xE7, 0x2F, 0x0F, 0x3B, 0x04,
	0x68, 0x6E, 0x21, 0xE7, 0xDE, 0xF0, 0x1B, 0x83, 0xDA, 0xF0, 0x1B,
	0x78, 0x68, 0x70, 0x43, 0xC8,
	0x21, 0xCA, 0xC2, 0xF0, 0x21, 0xC9, 0x68, 0x64, 0x3D, 0xD8, 0x23,
	0x81, 0x3E, 0x01, 0x21, 0xDA,
	0x33, 0x20, 0x3E, 0x02, 0x23, 0x80, 0x3E, 0x01, 0x23, 0x00, 0x3E,
	0xF5, 0x99, 0xE7, 0x68, 0x6D,
	0x68, 0x70, 0x43, 0xC8, 0x21, 0xCA, 0xC2, 0xF0, 0x21, 0xC9, 0x68,
	0x64, 0x68, 0x6B, 0x29, 0xD8,
	0x3B, 0x3B, 0xDA, 0xF1, 0x13, 0x8E, 0xCE, 0xF4, 0x68, 0x6F, 0xDD,
	0xE7, 0x68, 0x6E, 0x21, 0xDA,
	0x2F, 0xDF, 0x3E, 0x02, 0x68, 0x6A, 0xFE, 0x08, 0xBD, 0xF0, 0x03,
	0x9F, 0xD6, 0xF0, 0x1B, 0x9B,
	0x22, 0xF5, 0x3B, 0x08, 0xDA, 0xF1, 0x68, 0x6E, 0x21, 0xDA, 0x2F,
	0xDF, 0x3E, 0x02, 0x68, 0x6A,
	0xFE, 0x08, 0xBD, 0xF0, 0x95, 0xE7, 0x22, 0xF5, 0x3B, 0x22, 0xDA,
	0xF1, 0x68, 0x6E, 0x23, 0x00,
	0x3D, 0xE7, 0x8E, 0xF4, 0x68, 0x6D, 0x0B, 0xC6, 0xBD, 0xF2, 0x9E,
	0xF6, 0x03, 0xA4, 0x23, 0xCC,
	0x3D, 0xF4, 0x0B, 0xC6, 0x96, 0xF4, 0x9E, 0xF6, 0x03, 0xAC, 0xBD,
	0xD9, 0xA2, 0x0A, 0x86, 0x0A,
	0xAE, 0x0A, 0xBD, 0xF1, 0x0B, 0x5D, 0xD9, 0xF1, 0x1B, 0xBD, 0x0A,
	0xE0, 0xDD, 0xF3, 0x13, 0xB1,
	0x23, 0x00, 0x3D, 0xF1, 0xBD, 0xF3, 0x9D, 0xD9, 0x00, 0x00, 0x58,
	0x00, 0x4B, 0xC8, 0x41, 0xF8,
	0x22, 0x7F, 0x2F, 0xBF, 0x3E, 0x7F, 0x21, 0xF6, 0x4D, 0xC8, 0x03,
	0xC5, 0xA2, 0x0A, 0x86, 0x0A,
	0xAE, 0x0A, 0xD9, 0xDD, 0x1B, 0xCD, 0x23, 0x00, 0x3E, 0x46, 0x23,
	0x00, 0x3D, 0xF5, 0x3D, 0xF0,
	0x3D, 0xEF, 0x23, 0x12, 0x3E, 0x06, 0x23, 0xC0, 0x3E, 0x08, 0x68,
	0x6D, 0x00, 0x00, 0x00, 0x00
};

static unsigned char resRVdce[192] = {
	0x22, 0x86, 0xD9, 0xDD, 0x15, 0x04, 0x22, 0x08, 0x2F, 0xF0, 0x3B,
	0xF0, 0x1D, 0x30, 0xD9, 0xDD,
	0x1D, 0x0E, 0x21, 0xF0, 0x37, 0xFA, 0x2F, 0xFA, 0x3E, 0x08, 0x05,
	0x17, 0xDD, 0xF0, 0x68, 0x6C,
	0xF2, 0x06, 0xD9, 0xF0, 0x68, 0x6C, 0xE6, 0x06, 0xC5, 0xF0, 0x68,
	0x6C, 0xE2, 0x06, 0xDD, 0xF2,
	0x1D, 0x2C, 0xDD, 0xF1, 0x15, 0x1E, 0xDD, 0xD9, 0x1D, 0x36, 0xBD,
	0xF1, 0x22, 0x06, 0x37, 0xFF,
	0x39, 0xEF, 0x15, 0x23, 0x68, 0x6D, 0x3D, 0xC8, 0x35, 0xEF, 0x2D,
	0xEE, 0x3B, 0x00, 0x21, 0xC8,
	0x3D, 0xEF, 0x68, 0x6F, 0x92, 0xF4, 0x68, 0x6D, 0x0D, 0x4F, 0xBD,
	0xF2, 0x9E, 0xF6, 0x05, 0x2D,
	0x23, 0xCC, 0x3D, 0xF4, 0x0D, 0x4F, 0x96, 0xF4, 0x9E, 0xF6, 0x05,
	0x35, 0xBD, 0xD9, 0xA2, 0x0A,
	0x86, 0x0A, 0xAE, 0x0A, 0xBD, 0xF1, 0x0D, 0x5F, 0xD9, 0xF1, 0x1D,
	0x46, 0x0D, 0x00, 0xDD, 0xF3,
	0x15, 0x3A, 0x23, 0x00, 0x3D, 0xF1, 0xBD, 0xF3, 0x9D, 0xD9, 0x00,
	0x00, 0x58, 0x00, 0x4B, 0xC8,
	0x41, 0xF8, 0x22, 0x7F, 0x2F, 0xBF, 0x3E, 0x7F, 0x21, 0xF6, 0x4D,
	0xC8, 0x05, 0x4E, 0xA2, 0x0A,
	0x86, 0x0A, 0xAE, 0x0A, 0xD9, 0xDD, 0x1D, 0x56, 0x23, 0x00, 0x3E,
	0x46, 0x23, 0x00, 0x3D, 0xF5,
	0x3D, 0xF0, 0x3D, 0xEF, 0x23, 0x12, 0x3E, 0x06, 0x23, 0xC0, 0x3E,
	0x08, 0x68, 0x6D, 0x68, 0x6D
};

static unsigned char funcLoad[1760] = {
	0x81, 0xDD, 0xA1, 0xC5, 0x23, 0x40, 0x3E, 0x01, 0x21, 0xC6, 0x3E,
	0x02, 0x21, 0xC7, 0x3E, 0x03,
	0x23, 0x10, 0x3E, 0xB7, 0x3E, 0xB5, 0x23, 0x20, 0x3E, 0xB6, 0x3E,
	0xB4, 0xCD, 0xBD, 0x18, 0x13,
	0x23, 0x7E, 0x3D, 0x9F, 0x00, 0x15, 0x23, 0xC0, 0x3D, 0x9F, 0x50,
	0x00, 0x43, 0x84, 0x23, 0x00,
	0x3D, 0x98, 0x0B, 0x20, 0x09, 0xB0, 0xC1, 0xC5, 0x18, 0x01, 0x49,
	0x8C, 0x68, 0x6D, 0x21, 0xBF,
	0x3B, 0x00, 0x18, 0x27, 0xC2, 0x04, 0x10, 0x34, 0x3E, 0x00, 0x23,
	0x00, 0x3D, 0xBF, 0xC5, 0xC5,
	0x18, 0x34, 0xC9, 0xC5, 0x10, 0x32, 0xC2, 0x04, 0x10, 0x34, 0x23,
	0x00, 0x68, 0x6A, 0x3E, 0x00,
	0x3E, 0x00, 0x00, 0x34, 0x49, 0x8E, 0x68, 0x6D, 0x21, 0xC5, 0x39,
	0xC4, 0x13, 0x3C, 0xC2, 0xF1,
	0x10, 0x5B, 0xD5, 0x98, 0x18, 0x46, 0x95, 0x98, 0x45, 0x8A, 0xDE,
	0xF1, 0x10, 0x51, 0x21, 0x42,
	0x39, 0x40, 0x10, 0x56, 0x45, 0x88, 0xDE, 0xF1, 0x10, 0x57, 0x00,
	0x5B, 0xB5, 0x98, 0x45, 0x88,
	0xDE, 0xF1, 0x10, 0x57, 0x45, 0x8A, 0xDE, 0xF1, 0x10, 0x51, 0x21,
	0x42, 0x39, 0x40, 0x10, 0x56,
	0x00, 0x5B, 0x49, 0x8A, 0x45, 0x84, 0x47, 0x8A, 0x68, 0x6D, 0x00,
	0x5B, 0x00, 0xDB, 0x49, 0x88,
	0x45, 0x84, 0x47, 0x88, 0x68, 0x6D, 0x0D, 0x00, 0x00, 0x1B, 0x00,
	0x1F, 0xDE, 0x04, 0x10, 0x1F,
	0x78, 0x62, 0x67, 0x8C, 0xD9, 0x97, 0x10, 0x66, 0xB9, 0x97, 0x00,
	0x99, 0xDE, 0x04, 0x10, 0xB8,
	0x68, 0x70, 0x43, 0x80, 0x45, 0x82, 0x21, 0xB7, 0x68, 0x62, 0x47,
	0x82, 0x22, 0x00, 0x3D, 0x90,
	0xCA, 0x0A, 0x18, 0x99, 0x22, 0x04, 0x2F, 0x28, 0x3D, 0x91, 0x3B,
	0x00, 0x18, 0x80, 0xD6, 0xF0,
	0x10, 0x7C, 0x95, 0xA4, 0x8D, 0xA4, 0x00, 0x80, 0xCE, 0xF0, 0x10,
	0x80, 0x9D, 0xA4, 0x8D, 0xA4,
	0x45, 0x70, 0x21, 0x90, 0x68, 0x67, 0x68, 0x60, 0x47, 0x70, 0x21,
	0x99, 0x68, 0x68, 0x3D, 0x99,
	0x53, 0x74, 0x18, 0xD9, 0xCD, 0xA4, 0xAD, 0xA4, 0x18, 0x99, 0x3B,
	0x20, 0x68, 0x6E, 0x21, 0x70,
	0x3B, 0x60, 0xDA, 0xF1, 0x10, 0x97, 0x23, 0x00, 0x3D, 0x70, 0x39,
	0x74, 0x18, 0xD9, 0x23, 0x00,
	0x00, 0xBC, 0x78, 0x5E, 0x67, 0x8C, 0x21, 0x70, 0x3B, 0x60, 0xDA,
	0xF1, 0x10, 0xA3, 0x23, 0x00,
	0x3D, 0x70, 0x39, 0x74, 0x18, 0xA7, 0xCA, 0x0A, 0x18, 0xA7, 0xC1,
	0x97, 0x10, 0xAC, 0xAA, 0x0A,
	0xA1, 0x97, 0xA2, 0x0A, 0x23, 0x50, 0x00, 0xB5, 0xDD, 0xA4, 0x10,
	0xB0, 0x23, 0x90, 0x00, 0xB5,
	0xD5, 0xA4, 0x10, 0xB4, 0x23, 0x11, 0x00, 0xB5, 0x23, 0x10, 0xBD,
	0xA4, 0xB5, 0xA4, 0x00, 0xBC,
	0x41, 0x80, 0x68, 0x71, 0x68, 0x6E, 0x00, 0x99, 0x45, 0x40, 0x68,
	0x67, 0x68, 0x60, 0x21, 0x99,
	0x68, 0x67, 0x23, 0x00, 0x3D, 0x99, 0x21, 0x40, 0x68, 0x6B, 0x27,
	0x02, 0x3B, 0x40, 0x10, 0xC9,
	0x23, 0x00, 0x39, 0x42, 0x18, 0xCD, 0x3D, 0x40, 0x68, 0x6D, 0x45,
	0x40, 0x3D, 0x40, 0x23, 0x50,
	0x68, 0x67, 0x68, 0x60, 0x68, 0x66, 0x33, 0x80, 0x68, 0x67, 0xA2,
	0x0A, 0x78, 0x5E, 0x67, 0x8C,
	0x00, 0x1F, 0x81, 0x97, 0x00, 0x99, 0x45, 0x42, 0x68, 0x66, 0x3D,
	0x94, 0x68, 0x60, 0x68, 0x66,
	0x2F, 0x3F, 0x3D, 0x95, 0x3B, 0x00, 0x19, 0x20, 0xDD, 0x97, 0x19,
	0x00, 0xD9, 0x97, 0x10, 0xEA,
	0x9D, 0x97, 0x01, 0x00, 0x45, 0x50, 0x43, 0x90, 0x50, 0x00, 0x21,
	0x95, 0x68, 0x62, 0x53, 0x90,
	0xDA, 0xF1, 0x19, 0x5B, 0x45, 0x50, 0x68, 0x63, 0x47, 0x50, 0x41,
	0x4C, 0x21, 0x95, 0x4F, 0x74,
	0x68, 0x62, 0x43, 0x4C, 0x45, 0x54, 0x68, 0x62, 0x47, 0x54, 0x79,
	0x04, 0x67, 0x8A, 0x00, 0x5B,
	0x21, 0x95, 0x45, 0x54, 0x68, 0x62, 0x47, 0x54, 0x21, 0x74, 0x68,
	0x6B, 0x25, 0x95, 0x3B, 0x60,
	0xDA, 0xF1, 0x11, 0x0B, 0x23, 0x00, 0x3D, 0x74, 0x21, 0x94, 0xDA,
	0xF0, 0x11, 0x10, 0x82, 0x0A,
	0x21, 0x94, 0xD2, 0xF0, 0x19, 0x20, 0x45, 0x50, 0xDE, 0xF1, 0x11,
	0x88, 0xDD, 0x97, 0x19, 0x88,
	0x41, 0x48, 0xDE, 0xF1, 0x19, 0x1E, 0x23, 0x0C, 0x55, 0x48, 0x01,
	0x88, 0x99, 0x97, 0x01, 0x88,
	0xCD, 0x94, 0x11, 0x24, 0x8E, 0xF2, 0x01, 0x88, 0x41, 0x44, 0xDE,
	0xF1, 0x19, 0x45, 0x45, 0x42,
	0x68, 0x60, 0x68, 0x66, 0xDE, 0xF0, 0x11, 0x2D, 0x82, 0x0A, 0x21,
	0x94, 0xDD, 0x97, 0x11, 0x32,
	0x33, 0x02, 0xBD, 0x97, 0xB9, 0x97, 0x3D, 0x56, 0x41, 0xB0, 0x23,
	0x10, 0x68, 0x62, 0x23, 0x04,
	0x57, 0x54, 0x79, 0x3C, 0x67, 0x8A, 0x00, 0x5B, 0x92, 0xF2, 0x23,
	0x00, 0x3D, 0x54, 0x3D, 0x55,
	0x41, 0x44, 0x43, 0xB0, 0x23, 0x10, 0x55, 0x44, 0x01, 0x88, 0x41,
	0xB0, 0x23, 0x10, 0x55, 0x44,
	0x79, 0x4B, 0x67, 0x8A, 0x00, 0x5B, 0x41, 0x44, 0xDE, 0xF1, 0x11,
	0x27, 0x68, 0x70, 0x23, 0x0E,
	0x68, 0x62, 0x43, 0x78, 0x79, 0x55, 0x67, 0x8A, 0x00, 0x5B, 0x41,
	0x78, 0x68, 0x71, 0x19, 0x45,
	0x79, 0x55, 0x67, 0x8A, 0x00, 0x5B, 0x21, 0x95, 0x68, 0x6B, 0x29,
	0x50, 0x3D, 0x95, 0x41, 0x4C,
	0x21, 0x50, 0x4F, 0x74, 0x45, 0x54, 0x68, 0x62, 0x47, 0x54, 0x79,
	0x68, 0x67, 0x8A, 0x00, 0x5B,
	0x21, 0x74, 0x68, 0x6B, 0x25, 0x50, 0x3D, 0x74, 0x41, 0x48, 0xDE,
	0xF1, 0x11, 0x77, 0x9D, 0x97,
	0x21, 0x95, 0x45, 0x54, 0x68, 0x62, 0x47, 0x54, 0x50, 0x00, 0x47,
	0x50, 0x01, 0x04, 0x23, 0x0C,
	0x55, 0x48, 0x79, 0x7C, 0x67, 0x8A, 0x00, 0x5B, 0x45, 0x50, 0x43,
	0x90, 0x50, 0x00, 0x21, 0x95,
	0x68, 0x62, 0x53, 0x90, 0xDA, 0xF1, 0x19, 0x5B, 0x45, 0x50, 0x68,
	0x63, 0x47, 0x50, 0x00, 0xF5,
	0x21, 0x42, 0x68, 0x6B, 0x27, 0x02, 0x3B, 0x40, 0x11, 0x8E, 0x23,
	0x00, 0x3D, 0x42, 0x00, 0x5B,
	0x4B, 0x92, 0x09, 0xB0, 0x41, 0xB0, 0xDE, 0xF1, 0x11, 0x9B, 0x79,
	0x99, 0x67, 0x8A, 0x49, 0x92,
	0x68, 0x6D, 0x4B, 0x92, 0x01, 0x92, 0x23, 0x10, 0x55, 0x44, 0x79,
	0xA1, 0x67, 0x8A, 0x49, 0x92,
	0x68, 0x6D, 0x82, 0x0A, 0x78, 0x5E, 0x67, 0x8C, 0x50, 0x00, 0x43,
	0x70, 0x43, 0x74, 0x23, 0x00,
	0x3D, 0xB5, 0x3D, 0x99, 0x3D, 0x54, 0x3D, 0x55, 0x3D, 0x97, 0x00,
	0x55, 0x81, 0x98, 0x01, 0xB1,
	0xA1, 0x98, 0xA2, 0x0A, 0x51, 0x00, 0x47, 0x40, 0x47, 0x42, 0x45,
	0x84, 0x47, 0x8A, 0x78, 0x5D,
	0x67, 0x8C, 0xC1, 0x98, 0x68, 0x6E, 0x8A, 0xF2, 0x68, 0x6D, 0x00,
	0x34, 0x96, 0xF2, 0xB1, 0x98,
	0xDD, 0xBC, 0x10, 0x34, 0xBD, 0xBC, 0x79, 0xC8, 0x67, 0x8E, 0x7A,
	0x5F, 0x67, 0x88, 0x00, 0x34,
	0x00, 0x34, 0xC2, 0x04, 0x10, 0x34, 0x4B, 0x92, 0x0A, 0x4F, 0x49,
	0x92, 0xC1, 0x9D, 0x11, 0xD1,
	0x00, 0x34, 0x79, 0xD4, 0x67, 0x8E, 0x68, 0x6D, 0xC2, 0x04, 0x10,
	0x34, 0x4B, 0x92, 0x0A, 0x4F,
	0x49, 0x92, 0xC1, 0x9D, 0x11, 0xDC, 0x00, 0x34, 0x45, 0x6E, 0x68,
	0x66, 0x68, 0x6B, 0x3E, 0x00,
	0x21, 0x6E, 0x68, 0x68, 0x3B, 0x00, 0x11, 0xE5, 0x23, 0x80, 0x3D,
	0x6E, 0x68, 0x6B, 0x29, 0x6C,
	0x11, 0xEC, 0xC9, 0x98, 0x19, 0xF9, 0x01, 0xEE, 0xC9, 0x98, 0x19,
	0xF3, 0x3B, 0x00, 0x68, 0x6E,
	0x7A, 0x10, 0x67, 0x8E, 0x68, 0x6D, 0x3B, 0x1F, 0xDA, 0xF1, 0x19,
	0xF9, 0x3B, 0x00, 0x19, 0xFF,
	0x68, 0x6D, 0x45, 0x88, 0xDE, 0xF1, 0x68, 0x6E, 0x7A, 0x82, 0x67,
	0x88, 0x68, 0x6D, 0x7A, 0x02,
	0x67, 0x8E, 0x68, 0x6D, 0xC2, 0x04, 0x10, 0x34, 0x4B, 0x92, 0x0A,
	0x4F, 0x49, 0x92, 0xC1, 0x9D,
	0x12, 0x0A, 0x00, 0x34, 0x21, 0x6C, 0x39, 0x6E, 0x18, 0x34, 0x79,
	0xD4, 0x67, 0x8E, 0x01, 0xDC,
	0xC2, 0x04, 0x10, 0x34, 0x4B, 0x92, 0x0A, 0x4F, 0x49, 0x92, 0xC1,
	0x9D, 0x12, 0x18, 0x00, 0x34,
	0xBD, 0x98, 0x91, 0x98, 0xAD, 0x98, 0x7A, 0x3F, 0x67, 0x8E, 0x7A,
	0x20, 0x67, 0x88, 0x68, 0x6D,
	0x41, 0x58, 0xDE, 0xF1, 0x12, 0x29, 0x41, 0x7C, 0x23, 0x14, 0x55,
	0x58, 0x7A, 0x29, 0x67, 0x88,
	0x00, 0x5B, 0x23, 0x80, 0x3D, 0x6A, 0x41, 0x7C, 0x23, 0x12, 0x68,
	0x62, 0x23, 0x01, 0x57, 0x6A,
	0x7A, 0x33, 0x67, 0x88, 0x00, 0x5B, 0x8D, 0x98, 0x41, 0x58, 0xDE,
	0xF1, 0x12, 0x39, 0x9D, 0x98,
	0x68, 0x6D, 0x96, 0xF2, 0xBD, 0xBC, 0xB1, 0x98, 0x79, 0xC8, 0x67,
	0x8E, 0x02, 0x6D, 0xDD, 0x98,
	0x10, 0x34, 0xDD, 0xBC, 0x19, 0xBE, 0xD5, 0x6B, 0x1A, 0x4A, 0xD1,
	0x6B, 0x1A, 0x48, 0x91, 0x6B,
	0xC6, 0x04, 0x10, 0x34, 0x96, 0xF2, 0xB1, 0x98, 0x79, 0xC0, 0x67,
	0x8E, 0x00, 0x34, 0xC1, 0xBD,
	0x12, 0x56, 0x22, 0x06, 0xDA, 0xF0, 0x1A, 0x57, 0xBE, 0x0A, 0xA1,
	0x9D, 0x68, 0x6D, 0x81, 0x9D,
	0x22, 0x05, 0x2F, 0x0F, 0x3B, 0x01, 0xDA, 0xF1, 0x68, 0x6E, 0x9E,
	0x0A, 0x68, 0x6D, 0x41, 0xB8,
	0x23, 0x04, 0x55, 0x7C, 0x7A, 0x65, 0x67, 0x88, 0x00, 0x5B, 0x41,
	0x7C, 0xDE, 0xF1, 0x12, 0x6D,
	0x23, 0x88, 0x3D, 0xBE, 0x79, 0xC0, 0x67, 0x8E, 0x00, 0x5B, 0x23,
	0x14, 0x55, 0x58, 0x43, 0x7C,
	0x7A, 0x74, 0x67, 0x88, 0x99, 0xB6, 0x00, 0x5B, 0x41, 0x58, 0xDE,
	0xF1, 0x12, 0x7D, 0x41, 0xB8,
	0x23, 0x04, 0x57, 0x84, 0x7A, 0x7D, 0x67, 0x88, 0x00, 0x5B, 0x99,
	0x98, 0x89, 0x98, 0x50, 0x80,
	0x47, 0x6C, 0x47, 0x6E, 0x45, 0x68, 0xDE, 0xF1, 0x12, 0xA7, 0x23,
	0x80, 0xDD, 0x6B, 0x13, 0x19,
	0x23, 0x10, 0x3D, 0x6A, 0x41, 0x7C, 0x23, 0x12, 0x68, 0x62, 0x23,
	0x01, 0x57, 0x6A, 0x7A, 0x92,
	0x67, 0x88, 0x00, 0x5B, 0x41, 0x58, 0x23, 0x82, 0xDE, 0xF1, 0x1B,
	0x19, 0x23, 0x14, 0x55, 0x58,
	0x43, 0x7C, 0x7A, 0x9C, 0x67, 0x88, 0x00, 0x5B, 0xDD, 0x6B, 0x1A,
	0xA7, 0x41, 0x58, 0xDE, 0xF1,
	0x12, 0xA7, 0x41, 0xB8, 0x23, 0x04, 0x57, 0x84, 0x7A, 0xA7, 0x67,
	0x88, 0x00, 0x5B, 0x45, 0x66,
	0xDE, 0xF1, 0x1A, 0xC2, 0x65, 0x66, 0x63, 0x90, 0x45, 0x64, 0x53,
	0x90, 0xDA, 0xF1, 0x1A, 0xBB,
	0x68, 0x5D, 0x67, 0x66, 0x41, 0x5C, 0x23, 0x84, 0xDE, 0xF1, 0x1B,
	0x19, 0x23, 0x0A, 0x55, 0x5C,
	0x7A, 0xAA, 0x67, 0x88, 0x00, 0x5B, 0x68, 0x7D, 0x47, 0x64, 0x41,
	0x60, 0x68, 0x7C, 0x43, 0x60,
	0x70, 0x00, 0x67, 0x66, 0x45, 0x64, 0xDE, 0xF1, 0x12, 0xCE, 0x41,
	0x5C, 0x23, 0x86, 0xDE, 0xF1,
	0x1B, 0x19, 0x23, 0x0A, 0x55, 0x5C, 0x7A, 0xCE, 0x67, 0x88, 0x00,
	0x5B, 0x21, 0x6E, 0x68, 0x6B,
	0x29, 0x6C, 0x1A, 0xD5, 0x3B, 0x00, 0x1A, 0xD5, 0x02, 0xD8, 0x23,
	0x00, 0x68, 0x6B, 0x29, 0x6C,
	0x3D, 0x96, 0x45, 0x68, 0x43, 0x90, 0x70, 0x00, 0x68, 0x42, 0x73,
	0x90, 0xDA, 0xF1, 0x12, 0xE7,
	0x45, 0x64, 0x53, 0x90, 0xDA, 0xF1, 0x21, 0x64, 0x12, 0xED, 0x21,
	0x68, 0x02, 0xED, 0x63, 0x90,
	0x45, 0x64, 0x53, 0x90, 0xDA, 0xF1, 0x1A, 0xEE, 0x21, 0x64, 0x3D,
	0x96, 0x21, 0x96, 0x41, 0x60,
	0x4D, 0x6C, 0x68, 0x62, 0x43, 0x60, 0x45, 0x64, 0x68, 0x63, 0x47,
	0x64, 0x45, 0x68, 0x68, 0x63,
	0x47, 0x68, 0x7A, 0xFC, 0x67, 0x88, 0x00, 0x5B, 0x21, 0x6C, 0x68,
	0x6B, 0x25, 0x96, 0x3B, 0x00,
	0x13, 0x02, 0x23, 0x80, 0x3D, 0x6C, 0x45, 0x68, 0xDE, 0xF1, 0x13,
	0x0C, 0xDD, 0x6B, 0x1B, 0x0C,
	0xA9, 0x98, 0xD9, 0x98, 0x10, 0x5B, 0x03, 0x11, 0xD9, 0x98, 0x10,
	0x5B, 0x21, 0x6C, 0x3B, 0x80,
	0x12, 0x82, 0xB9, 0x98, 0x79, 0xC9, 0x67, 0x8E, 0xC2, 0x04, 0x10,
	0x5B, 0x49, 0x8E, 0x68, 0x6D,
	0x03, 0x14, 0x3D, 0xBE, 0x0B, 0x20, 0x09, 0xB0, 0x9E, 0xF4, 0x00,
	0x1B, 0x81, 0x98, 0x03, 0x21,
	0xA1, 0x98, 0x86, 0x0A, 0xAE, 0x0A, 0xC5, 0x98, 0x13, 0x28, 0x22,
	0x01, 0x2F, 0xFC, 0x3E, 0x01,
	0x45, 0x84, 0x47, 0x88, 0x79, 0xBD, 0x67, 0x8E, 0xC1, 0x98, 0x68,
	0x6E, 0x86, 0xF2, 0x68, 0x6D,
	0x79, 0xC0, 0x67, 0x8E, 0xB1, 0x98, 0xC5, 0x98, 0x13, 0x5F, 0x22,
	0x01, 0x2F, 0xFD, 0xD5, 0xBD,
	0x13, 0x3A, 0x33, 0x03, 0x3E, 0x01, 0x03, 0x5F, 0x3D, 0x90, 0x35,
	0xC4, 0x3D, 0x91, 0xD1, 0x90,
	0x1B, 0x62, 0xD5, 0x90, 0x13, 0x4A, 0xD5, 0x91, 0x10, 0x37, 0x09,
	0xB0, 0x0B, 0x20, 0x82, 0xF2,
	0x95, 0xC4, 0x00, 0x37, 0xDD, 0x91, 0x13, 0x51, 0xDD, 0x90, 0x13,
	0x50, 0x09, 0x90, 0x03, 0x51,
	0x09, 0xAE, 0xD9, 0x91, 0x13, 0x5F, 0xD9, 0x90, 0x1B, 0x30, 0xD1,
	0x98, 0x13, 0x5E, 0xCD, 0x98,
	0x1B, 0x5C, 0x21, 0x90, 0x33, 0x40, 0x03, 0x60, 0x96, 0xF2, 0xB1,
	0x98, 0x0B, 0x1E, 0x21, 0x90,
	0x3D, 0xC4, 0x00, 0x37, 0x91, 0xC4, 0x58, 0x00, 0x4B, 0xC8, 0x41,
	0xFC, 0x22, 0x7F, 0x2F, 0xBF,
	0x3E, 0x7F, 0x21, 0xF7, 0x4D, 0xC8, 0x03, 0x6B, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00
};
