/*
 * Copyright 2004 Digi International (www.digi.com)
 *      Scott H Kilau <Scott_Kilau at digi dot com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY, EXPRESS OR IMPLIED; without even the
 * implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
 * PURPOSE.  See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *	NOTE: THIS IS A SHARED HEADER. DO NOT CHANGE CODING STYLE!!!
 *
 *************************************************************************
 *
 * This file is intended to contain all the kernel "differences" between the
 * various kernels that we support.
 *
 *************************************************************************/

#ifndef __JSM_KCOMPAT_H
#define __JSM_KCOMPAT_H

# ifndef KERNEL_VERSION
#  define KERNEL_VERSION(a,b,c)  (((a) << 16) + ((b) << 8) + (c))
# endif



#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)

/* Sparse stuff */
# ifndef __user
#  define __user
#  define __kernel
#  define __safe
#  define __force
#  define __chk_user_ptr(x) (void)0
# endif

# define JSM_MAJOR(x)			(imajor(x))
# define JSM_MINOR(x)			(iminor(x))
# define JSM_TTY_MAJOR(tty)		(MAJOR(tty_devnum(tty)))
# define JSM_TTY_MINOR(tty)		(MINOR(tty_devnum(tty)))

# define JSM_MOD_INC_USE_COUNT(rtn)	(rtn = 1)
# define JSM_MOD_DEC_USE_COUNT		

# define JSM_IRQRETURN_TYPE		irqreturn_t
# define JSM_IRQ_RETURN(x)		return x;

# define PARM_GEN_IP			(PDE(file->f_dentry->d_inode))

# define JSM_GET_TTY_COUNT(x)		(x->count)



# elif LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,0)

/* No Sparse stuff in 2.4.x */
# define __user
# define __kernel
# define __safe
# define __force
# define __chk_user_ptr(x) (void)0

# define JSM_MAJOR(x)			(MAJOR((x)->i_rdev))
# define JSM_MINOR(x)			(MINOR((x)->i_rdev))
# define JSM_TTY_MAJOR(tty)		(MAJOR(tty->device))
# define JSM_TTY_MINOR(tty)		(MINOR(tty->device))

# define JSM_MOD_INC_USE_COUNT(rtn) \
	{ \
		MOD_INC_USE_COUNT; \
		rtn = 1; \
	}

# define JSM_MOD_DEC_USE_COUNT 	MOD_DEC_USE_COUNT

# define JSM_IRQRETURN_TYPE		void
# define JSM_IRQ_RETURN(x)		return;

# define PARM_GEN_IP			(file->f_dentry->d_inode->u.generic_ip)

/*
 * Not all 2.4.x kernels have min/max macro's in the kernel.h header.
 * I call your attention to the Red Hat AS 2.4.9-e.27 kernel.
 */
# ifndef min
#  define min(a,b) (((a)<(b))?(a):(b))
# endif
# ifndef max
#  define max(a,b) (((a)>(b))?(a):(b))
# endif

/*
 * Yes, this is ugly, but its needed because Red Hat added the
 * "atomic read for tty count" patch to their 2.4 kernels, but
 * kernel.org (and other vendors) did not.
 */
# if defined(REDHAT_AS_30)  || defined(REDHAT_ES_30) || defined(REDHAT_WS_30) || \
    defined(REDHAT_AS_3)   || defined(REDHAT_ES_3)  || defined(REDHAT_WS_3)  || \
    defined(REDHATAS30)    || defined(REDHATES30)   || defined(REDHATWS30)   || \
    defined(REDHATAS3)     || defined(REDHATES3)    || defined(REDHATWS3)    || \
    defined(REDHAT_FEDORA) || defined(FEDORA)

#  define JSM_GET_TTY_COUNT(x) (atomic_read(&(x->count)))

# else

#  define JSM_GET_TTY_COUNT(x) (x->count)

# endif


# else



# error "this driver does not support anything below the 2.4 kernel series."



# endif

#endif /* ! __JSM_KCOMPAT_H */
